/*
 *  linux/arch/nds32/include/asm/io.h
 *
 *  Copyright (C) 1996-2000 Russell King
 *  Copyright (C) 2008 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Modifications:
 *  16-Sep-1996	RMK	Inlined the inx/outx functions & optimised for both
 *			constant addresses and variable addresses.
 *  04-Dec-1997	RMK	Moved a lot of this stuff to the new architecture
 *			specific IO header files.
 *  27-Mar-1999	PJB	Second parameter of memcpy_toio is const..
 *  04-Apr-1999	PJB	Added check_signature.
 *  12-Dec-1999	RMK	More cleanups
 *  18-Jun-2000 RMK	Removed virt_to_* and friends definitions
 *  05-Oct-2004 BJD     Moved memory string functions to use void __iomem
 */
#ifndef __ASM_NDS32_IO_H
#define __ASM_NDS32_IO_H

#ifdef __KERNEL__

#include <linux/types.h>
#include <asm/byteorder.h>
#include <asm/memory.h>
#include <asm-generic/iomap.h>


/*
 * ISA I/O bus memory addresses are 1:1 with the physical address.
 */
#define isa_virt_to_bus virt_to_phys
#define isa_page_to_bus page_to_phys
#define isa_bus_to_virt phys_to_virt

/*
 * Generic IO read/write.  These perform native-endian accesses.  Note
 * that some architectures will want to re-define __raw_{read,write}w.
 */

#define __raw_writeb(v,a)	(*(volatile unsigned char __force  *)(a) = (v))
#define __raw_writew(v,a)	(*(volatile unsigned short __force *)(a) = (v))
#define __raw_writel(v,a)	(*(volatile unsigned int __force   *)(a) = (v))

#define __raw_readb(a)		(*(volatile unsigned char __force  *)(a))
#define __raw_readw(a)		(*(volatile unsigned short __force *)(a))
#define __raw_readl(a)		(*(volatile unsigned int __force   *)(a))

/*
 * Bad read/write accesses...
 */
extern void __readwrite_bug(const char *fn);

/*
 * Now, pick up the machine-defined IO definitions
 */
#ifndef __FARADAY_PLATFORM_IO_HEADER__
#define __FARADAY_PLATFORM_IO_HEADER__

#include <linux/compiler.h>

#ifndef IO_SPACE_LIMIT
#define IO_SPACE_LIMIT      0xffffffff
#endif

#ifndef __io
#define __io(a)             ((void __iomem *)(a))
#endif
#define IO_ADDRESS(a)	__io(a)
#ifndef __mem_pci
#define __mem_pci(a)        (a)
#endif

#endif

#ifdef __io_pci
#warning machine class uses buggy __io_pci
#endif
#if defined(__arch_putb) || defined(__arch_putw) || defined(__arch_putl) || \
    defined(__arch_getb) || defined(__arch_getw) || defined(__arch_getl)
#endif

/*
 *  IO port access primitives
 *  -------------------------
 *
 * The ARM doesn't have special IO access instructions; all IO is memory
 * mapped.  Note that these are defined to perform little endian accesses
 * only.  Their primary purpose is to access PCI and ISA peripherals.
 *
 * Note that for a big endian machine, this implies that the following
 * big endian mode connectivity is in place, as described by numerious
 * ARM documents:
 *
 *    PCI:  D0-D7   D8-D15 D16-D23 D24-D31
 *    ARM: D24-D31 D16-D23  D8-D15  D0-D7
 *
 * The machine specific io.h include defines __io to translate an "IO"
 * address to a memory address.
 *
 * Note that we prevent GCC re-ordering or caching values in expressions
 * by introducing sequence points into the in*() definitions.  Note that
 * __raw_* do not guarantee this behaviour.
 *
 * The {in,out}[bwl] macros are for emulating x86-style PCI/ISA IO space.
 */
#ifdef __io
#ifdef __NDS32_EB__
#define inw(p)	({ unsigned int __v = be16_to_cpu(__raw_readw(__io(p))); __v; })
#define inl(p)	({ unsigned int __v = be32_to_cpu(__raw_readl(__io(p))); __v; })
#define outw(v,p)		__raw_writew(cpu_to_be16(v),__io(p))
#define outl(v,p)		__raw_writel(cpu_to_be32(v),__io(p))
#else
#define inw(p)	({ unsigned int __v = le16_to_cpu(__raw_readw(__io(p))); __v; })
#define inl(p)	({ unsigned int __v = le32_to_cpu(__raw_readl(__io(p))); __v; })
#define outw(v,p)		__raw_writew(cpu_to_le16(v),__io(p))
#define outl(v,p)		__raw_writel(cpu_to_le32(v),__io(p))
#endif

#define inb(p)	({ unsigned int __v = __raw_readb(__io(p)); __v; })
#define outb(v,p)		__raw_writeb(v,__io(p))

#endif

#define outb_p(val,port)	outb((val),(port))
#define outw_p(val,port)	outw((val),(port))
#define outl_p(val,port)	outl((val),(port))
#define inb_p(port)		inb((port))
#define inw_p(port)		inw((port))
#define inl_p(port)		inl((port))

/*
 * String version of IO memory access ops:
 */
extern void _memcpy_fromio(void *, const volatile void __iomem *, size_t);
extern void _memcpy_toio(volatile void __iomem *, const void *, size_t);
extern void _memset_io(volatile void __iomem *, int, size_t);

#define mmiowb()

/*
 *  Memory access primitives
 *  ------------------------
 *
 * These perform PCI memory accesses via an ioremap region.  They don't
 * take an address as such, but a cookie.
 *
 * Again, this are defined to perform little endian accesses.  See the
 * IO port primitives for more information.
 */
#ifdef __mem_pci
#ifdef __NDS32_EB__
#define readw(c) ({ unsigned int __v = be16_to_cpu(__raw_readw(__mem_pci(c))); __v; })
#define readl(c) ({ unsigned int __v = be32_to_cpu(__raw_readl(__mem_pci(c))); __v; })
#define writew(v,c)		__raw_writew(cpu_to_be16(v),__mem_pci(c))
#define writel(v,c)		__raw_writel(cpu_to_be32(v),__mem_pci(c))
#else
#define readw(c) ({ unsigned int __v = le16_to_cpu(__raw_readw(__mem_pci(c))); __v; })
#define readl(c) ({ unsigned int __v = le32_to_cpu(__raw_readl(__mem_pci(c))); __v; })
#define writew(v,c)		__raw_writew(cpu_to_le16(v),__mem_pci(c))
#define writel(v,c)		__raw_writel(cpu_to_le32(v),__mem_pci(c))
#endif

#define readb(c) ({ unsigned int __v = __raw_readb(__mem_pci(c)); __v; })
#define writeb(v,c)		__raw_writeb(v,__mem_pci(c))

#define readb_relaxed(addr) readb(addr)
#define readw_relaxed(addr) readw(addr)
#define readl_relaxed(addr) readl(addr)

#define memset_io(c,v,l)	_memset_io(__mem_pci(c),(v),(l))
#define memcpy_fromio(a,c,l)	_memcpy_fromio((a),__mem_pci(c),(l))
#define memcpy_toio(c,a,l)	_memcpy_toio(__mem_pci(c),(a),(l))

#define eth_io_copy_and_sum(s,c,l,b) \
				eth_copy_and_sum((s),__mem_pci(c),(l),(b))

#elif !defined(readb)

#define readb(c)			(__readwrite_bug("readb"),0)
#define readw(c)			(__readwrite_bug("readw"),0)
#define readl(c)			(__readwrite_bug("readl"),0)
#define writeb(v,c)			__readwrite_bug("writeb")
#define writew(v,c)			__readwrite_bug("writew")
#define writel(v,c)			__readwrite_bug("writel")

#define eth_io_copy_and_sum(s,c,l,b)	__readwrite_bug("eth_io_copy_and_sum")

#endif	/* __mem_pci */

/*
 * If this architecture has ISA IO, then define the isa_read/isa_write
 * macros.
 */
#ifdef __mem_isa

#define isa_readb(addr)			__raw_readb(__mem_isa(addr))
#define isa_readw(addr)			__raw_readw(__mem_isa(addr))
#define isa_readl(addr)			__raw_readl(__mem_isa(addr))
#define isa_writeb(val,addr)		__raw_writeb(val,__mem_isa(addr))
#define isa_writew(val,addr)		__raw_writew(val,__mem_isa(addr))
#define isa_writel(val,addr)		__raw_writel(val,__mem_isa(addr))
#define isa_memset_io(a,b,c)		_memset_io(__mem_isa(a),(b),(c))
#define isa_memcpy_fromio(a,b,c)	_memcpy_fromio((a),__mem_isa(b),(c))
#define isa_memcpy_toio(a,b,c)		_memcpy_toio(__mem_isa((a)),(b),(c))

#define isa_eth_io_copy_and_sum(a,b,c,d) \
				eth_copy_and_sum((a),__mem_isa(b),(c),(d))

#else	/* __mem_isa */

#define isa_readb(addr)			(__readwrite_bug("isa_readb"),0)
#define isa_readw(addr)			(__readwrite_bug("isa_readw"),0)
#define isa_readl(addr)			(__readwrite_bug("isa_readl"),0)
#define isa_writeb(val,addr)		__readwrite_bug("isa_writeb")
#define isa_writew(val,addr)		__readwrite_bug("isa_writew")
#define isa_writel(val,addr)		__readwrite_bug("isa_writel")
#define isa_memset_io(a,b,c)		__readwrite_bug("isa_memset_io")
#define isa_memcpy_fromio(a,b,c)	__readwrite_bug("isa_memcpy_fromio")
#define isa_memcpy_toio(a,b,c)		__readwrite_bug("isa_memcpy_toio")

#define isa_eth_io_copy_and_sum(a,b,c,d) \
				__readwrite_bug("isa_eth_io_copy_and_sum")

#endif	/* __mem_isa */

/*
 * ioremap and friends.
 *
 * ioremap takes a PCI memory address, as specified in
 * Documentation/IO-mapping.txt.
 */
extern void __iomem * __ioremap(unsigned long, size_t, unsigned long, unsigned long);
extern void __iounmap(void __iomem *addr);

#ifndef __arch_ioremap
#define ioremap(cookie,size)		__ioremap(cookie,size,0,1)
#define ioremap_nocache(cookie,size)	__ioremap(cookie,size,0,1)
#define iounmap(cookie)			__iounmap(cookie)
#else
#define ioremap(cookie,size)		__arch_ioremap((cookie),(size),0,1)
#define ioremap_nocache(cookie,size)	__arch_ioremap((cookie),(size),0,1)
#define iounmap(cookie)			__arch_iounmap(cookie)
#endif

/*
 * can the hardware map this into one segment or not, given no other
 * constraints.
 */
#define BIOVEC_MERGEABLE(vec1, vec2)	\
	((bvec_to_phys((vec1)) + (vec1)->bv_len) == bvec_to_phys((vec2)))

/*
 * Convert a physical pointer to a virtual kernel pointer for /dev/mem
 * access
 */
#define xlate_dev_mem_ptr(p)	__va(p)

/*
 * Convert a virtual cached pointer to an uncached pointer
 */
#define xlate_dev_kmem_ptr(p)	p
static inline void readsb(const void __iomem *addr, void * data, int bytelen)
{
        unsigned char *ptr = (unsigned char *)addr;
        unsigned char *ptr2 = (unsigned char *)data;
        while(bytelen) {
                *ptr2 = *ptr;
                ptr2++;
                bytelen--;
        }
}

static inline void readsw(const void __iomem *addr, void * data, int wordlen)
{
        unsigned short *ptr = (unsigned short *)addr;
        unsigned short *ptr2 = (unsigned short *)data;
        while(wordlen) {
                *ptr2 = *ptr;
                ptr2++;
                wordlen--;
        }
}

static inline void readsl(const void __iomem *addr, void * data, int longlen)
{
        unsigned int *ptr = (unsigned int *)addr;
        unsigned int *ptr2 = (unsigned int *)data;
        while(longlen) {
                *ptr2 = *ptr;
                ptr2++;
                longlen--;
        }
}
static inline void writesb(void __iomem *addr, const void * data, int bytelen)
{
        unsigned char *ptr = (unsigned char *)addr;
        unsigned char *ptr2 = (unsigned char *)data;
        while(bytelen) {
                *ptr = *ptr2;
                ptr2++;
                bytelen--;
        }
}
static inline void writesw(void __iomem *addr, const void * data, int wordlen)
{
        unsigned short *ptr = (unsigned short *)addr;
        unsigned short *ptr2 = (unsigned short *)data;
        while(wordlen) {
                *ptr = *ptr2;
                ptr2++;
                wordlen--;
        }
}
static inline void writesl(void __iomem *addr, const void * data, int longlen)
{
        unsigned int *ptr = (unsigned int *)addr;
        unsigned int *ptr2 = (unsigned int *)data;
        while(longlen) {
                *ptr = *ptr2;
                ptr2++;
                longlen--;
        }
}


#define insb(p,d,l)		BUG()
#define insw(p,d,l)		BUG()
#define insl(p,d,l)		BUG()
#define outsb(p,d,l)		BUG()
#define outsw(p,d,l)		BUG()
#define outsl(p,d,l)		BUG()

#endif	/* __KERNEL__ */
#endif	/* __ASM_NDS32_IO_H */
