#include <linux/module.h>
#include <linux/kallsyms.h>
#include <asm/uaccess.h>
#include <asm/elf.h>
#include <asm/fpu.h>
#include <linux/elf.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#define ELF_CHECKING_OS
#include "nds32-elf.h"

extern struct proc_dir_entry *proc_dir_cpu;

#ifdef CONFIG_ELFCHK_DEFAULT_ENABLE
static int elf_check_en = 1;
#else
static int elf_check_en = 0;
#endif
static DEFINE_MUTEX(nds32_elfchk_lock);

struct reg_struct {

	const char *name;
	int idx;
};

static struct reg_struct regs[] = {

	{"cpu_ver", CPU_SR_INDEX(0, 0, 0)},
	{"icm_cfg", CPU_SR_INDEX(0, 1, 0)},
	{"dcm_cfg", CPU_SR_INDEX(0, 2, 0)},
	{"mmu_cfg", CPU_SR_INDEX(0, 3, 0)},
	{"msc_cfg", CPU_SR_INDEX(0, 4, 0)},
	{"msc_cfg2", CPU_SR_INDEX(0, 4, 1)},
#ifdef CONFIG_FPU
	{"fucop_exist", CPU_SR_INDEX(0, 5, 0)}
#endif
};

#ifdef CONFIG_FPU
static unsigned int read_fpu_fpcfg(void)
{
	unsigned int fpcfg = 0;

	enable_fpu();
	asm volatile ("fmfcfg    %0\n\t":"=&r" (fpcfg));
	disable_fpu();

	return fpcfg;
}
#endif

static unsigned int read_cpu_sr(unsigned int idx)
{
	switch (idx) {

	case CPU_SR_INDEX(0, 0, 0):
		return GET_CPU_VER();
	case CPU_SR_INDEX(0, 1, 0):
		return GET_ICM_CFG();
	case CPU_SR_INDEX(0, 2, 0):
		return GET_DCM_CFG();
	case CPU_SR_INDEX(0, 3, 0):
		return GET_MMU_CFG();
	case CPU_SR_INDEX(0, 4, 0):
		return GET_MSC_CFG();
	case CPU_SR_INDEX(0, 4, 1):
		if (GET_MSC_CFG() & MSC_CFG_mskMSC_EXT)
			return GET_MSC_CFG2();
		return SR_NOT_EXIST;

	case CPU_SR_INDEX(0, 5, 0):
		return GET_FUCOP_EXIST();

	default:
		printk(KERN_ERR
		       "%s: invalid system register index (%d, %d, %d)\n",
		       __func__, (idx >> 7) & 0x7, (idx >> 3) & 0xf, idx & 0x7);

		return SR_NOT_EXIST;
	}
}

static unsigned int reg_read_callback(unsigned int idx)
{
	if (HW_IS_CPU(idx)) {

		return read_cpu_sr(SR_INDEX(idx));
	} else if (HW_IS_FPU(idx)) {

#ifdef CONFIG_FPU
		if (SR_FPU_FPCFG == SR_INDEX(idx))
			return read_fpu_fpcfg();
		else
			return SR_NOT_EXIST;
#endif
	} else if (HW_IS_AUDIO(idx)) {

		return SR_NOT_EXIST;
	}

	return SR_NOT_EXIST;
}

#define BUFLEN 1024

int do_elf_check_arch(const struct elf32_hdr *hdr)
{
	static char msg_buf[BUFLEN];
	unsigned int err;
	int buf_status = 0;

	if (!elf_check_en)
		return 1;

	mutex_lock(&nds32_elfchk_lock);
	err =
	    elf_check((void *)hdr, reg_read_callback, msg_buf, BUFLEN,
		      &buf_status);
	mutex_unlock(&nds32_elfchk_lock);

	if (err)
		printk(KERN_WARNING "%s", msg_buf);

	return !err;
}

static ssize_t proc_elf_check_read(struct file *file, char __user *userbuf,
						size_t count, loff_t *ppos)
{
	int ret;
	char buf[4];

	ret = sprintf(buf, "%d\n", elf_check_en);
	return simple_read_from_buffer(userbuf, count, ppos, buf, ret);
}

#define INPUTLEN 10
static ssize_t proc_elf_check_write(struct file *file,
			const char __user *ubuf, size_t count, loff_t *ppos)
{
	unsigned long en;
	char inbuf[INPUTLEN];

	if (count > INPUTLEN - 1)
		count = INPUTLEN - 1;

	if (copy_from_user(inbuf, ubuf, count))
		return -EFAULT;

	inbuf[count] = '\0';

	if (!sscanf(inbuf, "%lu", &en) || en > 3)
		return -EFAULT;

	elf_check_en = en & 0x01;

	return count;
}

static int cpu_sr_show(struct seq_file *m, void *v)
{
	unsigned long val = read_cpu_sr((unsigned int)m->private);

	seq_printf(m, "0x%08lx\n", val);
	return 0;
}

static const struct file_operations fops = {
	.open = simple_open,
	.read = proc_elf_check_read,
	.write = proc_elf_check_write,
};

static int cpu_sr_open(struct inode *inode, struct file *file)
{
	return single_open(file, cpu_sr_show, PDE_DATA(inode));
}


static const struct file_operations cpu_sr_fops = {
	.open 	= cpu_sr_open,
	.read 	= seq_read,
	.llseek	= seq_lseek,
	.release	= single_release,
};

int __init elf_check_init(void)
{
	static struct proc_dir_entry *res_elf_check;
	int i;

	if (!proc_dir_cpu) {
		if (!(proc_dir_cpu = proc_mkdir("cpu", NULL)))
			return -ENOMEM;
	}

	res_elf_check =
	    proc_create("elf_core_checking", S_IWUSR | S_IRUGO,
			      proc_dir_cpu, &fops);
	if (!res_elf_check)
		return -ENOMEM;

	for (i = 0; i < ARRAY_SIZE(regs); i++) {

		if (!proc_create_data(regs[i].name, S_IWUSR | S_IRUGO,
					    proc_dir_cpu,
					    &cpu_sr_fops,
					    (void *)regs[i].idx))

			return -ENOMEM;
	}

	return 0;
}

module_init(elf_check_init);
