/* ============================================================================
 *
 *  arch/nds32/kernel/kgdb.c
 *
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is for NDS32 KGDB support.
 *
 *  Author: Harry Pan
 *
 *  Revision History:
 *
 *    Nov.23.2007     Initial ported by Harry,
 *                    inherited from the KGDB in 2.6.11 and 2.4.35.
 *
 *  Note:
 *
 * ============================================================================
 */
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/spinlock.h>
#include <linux/personality.h>
#include <linux/ptrace.h>
#include <linux/elf.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/kgdb.h>
#include <linux/kdebug.h>

#include <asm/atomic.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/ptrace.h>
#include <asm/traps.h>
#ifdef CONFIG_NDS32_EDM_SUP
#include <asm/hw_breakpoint.h>
#endif //CONFIG_NDS32_EDM_SUP
struct dbg_reg_def_t dbg_reg_def[DBG_MAX_REG_NUM] =
{
	{ "r0", 4, offsetof(struct pt_regs, NDS32_r0)},
	{ "r1", 4, offsetof(struct pt_regs, NDS32_r1)},
	{ "r2", 4, offsetof(struct pt_regs, NDS32_r2)},
	{ "r3", 4, offsetof(struct pt_regs, NDS32_r3)},
	{ "r4", 4, offsetof(struct pt_regs, NDS32_r4)},
	{ "r5", 4, offsetof(struct pt_regs, NDS32_r5)},
	{ "r6", 4, offsetof(struct pt_regs, NDS32_r6)},
	{ "r7", 4, offsetof(struct pt_regs, NDS32_r7)},
	{ "r8", 4, offsetof(struct pt_regs, NDS32_r8)},
	{ "r9", 4, offsetof(struct pt_regs, NDS32_r9)},
	{ "r10", 4, offsetof(struct pt_regs, NDS32_r10)},
	{ "r11", 4, offsetof(struct pt_regs, NDS32_r11)},
	{ "r12", 4, offsetof(struct pt_regs, NDS32_r12)},
	{ "r13", 4, offsetof(struct pt_regs, NDS32_r13)},
	{ "r14", 4, offsetof(struct pt_regs, NDS32_r14)},
	{ "r15", 4, offsetof(struct pt_regs, NDS32_r15)},
	{ "r16", 4, offsetof(struct pt_regs, NDS32_r16)},
	{ "r17", 4, offsetof(struct pt_regs, NDS32_r17)},
	{ "r18", 4, offsetof(struct pt_regs, NDS32_r18)},
	{ "r19", 4, offsetof(struct pt_regs, NDS32_r19)},
	{ "r20", 4, offsetof(struct pt_regs, NDS32_r20)},
	{ "r21", 4, offsetof(struct pt_regs, NDS32_r21)},
	{ "r22", 4, offsetof(struct pt_regs, NDS32_r22)},
	{ "r23", 4, offsetof(struct pt_regs, NDS32_r23)},
	{ "r24", 4, offsetof(struct pt_regs, NDS32_r24)},
	{ "r25", 4, offsetof(struct pt_regs, NDS32_r25)},
	{ "r26", 4, -1},
	{ "r27", 4, -1},
	{ "fp", 4, offsetof(struct pt_regs, NDS32_fp)},
	{ "gp", 4, offsetof(struct pt_regs, NDS32_gp)},
	{ "lp", 4, offsetof(struct pt_regs, NDS32_lp)},
	{ "sp", 4, offsetof(struct pt_regs, NDS32_sp)},
	{ "pc", 4, offsetof(struct pt_regs, NDS32_ipc)},
};

char *dbg_get_reg(int regno, void *mem, struct pt_regs *regs)
{
	if (regno >= DBG_MAX_REG_NUM || regno < 0)
		return NULL;
	if (dbg_reg_def[regno].offset != -1)
		memcpy(mem, (void *)regs + dbg_reg_def[regno].offset,
		       dbg_reg_def[regno].size);
	else
		memset(mem, 0, dbg_reg_def[regno].size);
	return dbg_reg_def[regno].name;
}

int dbg_set_reg(int regno, void *mem, struct pt_regs *regs)
{
	if (regno >= DBG_MAX_REG_NUM || regno < 0)
		return -EINVAL;

	if (dbg_reg_def[regno].offset != -1)
		memcpy((void *)regs + dbg_reg_def[regno].offset, mem,
		       dbg_reg_def[regno].size);
	return 0;
}

// ----------------------------------------------------------------------------
// kgdb_get_user_regs()
//
// Get user process registers.
// ----------------------------------------------------------------------------
static inline struct pt_regs *kgdb_get_user_regs(struct task_struct *task)
{
	return (struct pt_regs *)
	    ((unsigned long)task_thread_info(task) + THREAD_SIZE -
	     8 - sizeof(struct pt_regs));
}

// ============================================================================
// sleeping_thread_to_gdb_regs()
//
// ============================================================================
void sleeping_thread_to_gdb_regs(unsigned long *gregs, struct task_struct *task)
{
	int regno;
	struct pt_regs *tregs;

	/* Just making sure... */
	if (task == NULL)
		return;

	/* Initialize to zero */
	for (regno = 0; regno < NDS32_NUM_REGS; regno++)
		gregs[regno] = 0;

	/* Otherwise, we have only some registers from switch_to() */
	tregs = kgdb_get_user_regs(task);

	gregs[0] = tregs->NDS32_r0;
	gregs[1] = tregs->NDS32_r1;
	gregs[2] = tregs->NDS32_r2;
	gregs[3] = tregs->NDS32_r3;
	gregs[4] = tregs->NDS32_r4;
	gregs[5] = tregs->NDS32_r5;
	gregs[6] = tregs->NDS32_r6;
	gregs[7] = tregs->NDS32_r7;
	gregs[8] = tregs->NDS32_r8;
	gregs[9] = tregs->NDS32_r9;
	gregs[10] = tregs->NDS32_r10;
	gregs[11] = tregs->NDS32_r11;
	gregs[12] = tregs->NDS32_r12;
	gregs[13] = tregs->NDS32_r13;
	gregs[14] = tregs->NDS32_r14;
	gregs[15] = tregs->NDS32_r15;
	gregs[16] = tregs->NDS32_r16;
	gregs[17] = tregs->NDS32_r17;
	gregs[18] = tregs->NDS32_r18;
	gregs[19] = tregs->NDS32_r19;
	gregs[20] = tregs->NDS32_r20;
	gregs[21] = tregs->NDS32_r21;
	gregs[22] = tregs->NDS32_r22;
	gregs[23] = tregs->NDS32_r23;
	gregs[24] = tregs->NDS32_r24;
	gregs[25] = tregs->NDS32_r25;
	gregs[26] = tregs->NDS32_pp0;
	gregs[27] = tregs->NDS32_pp1;
	gregs[28] = tregs->NDS32_fp;
	gregs[29] = tregs->NDS32_gp;
	gregs[30] = tregs->NDS32_lp;
	gregs[31] = tregs->NDS32_sp;
	gregs[32] = tregs->NDS32_ipc;
}
#ifdef CONFIG_NDS32_EDM_SUP


static int nds32_kgdb_set_hw_break(unsigned long addr, int len, enum kgdb_bptype type)
{
	int ret;

	switch (type) {
	case BP_HARDWARE_BREAKPOINT:
		type = NDS32_HW_BREAKPOINT;
		break;
	case BP_WRITE_WATCHPOINT:
		type = NDS32_WRITE_WATCHPOINT;
		break;
	case BP_READ_WATCHPOINT:
		type = NDS32_READ_WATCHPOINT;
		break;
	case BP_ACCESS_WATCHPOINT:
		type = NDS32_ACCESS_WATCHPOINT;
		break;
	default:
		return -ENOSPC;
	}
	ret = nds32_set_hw_break(addr, VIRTUAL_ADDR,len,(enum nds32_hw_bptype) type);

	if(ret >= 0)
		return 0;
	else
		return -ENOSPC;
}
static int nds32_kgdb_remove_hw_break(unsigned long addr, int len, enum kgdb_bptype type)
{
	switch (type) {
	case BP_HARDWARE_BREAKPOINT:
		type = NDS32_HW_BREAKPOINT;
		break;
	case BP_WRITE_WATCHPOINT:
		type = NDS32_WRITE_WATCHPOINT;
		break;
	case BP_READ_WATCHPOINT:
		type = NDS32_READ_WATCHPOINT;
		break;
	case BP_ACCESS_WATCHPOINT:
		type = NDS32_ACCESS_WATCHPOINT;
		break;
	default:
		return -ENOSPC;
	}

	nds32_remove_hw_break(addr, len,(enum nds32_hw_bptype) type);
	return 0;
}
static void nds32_kgdb_disable_hw_debug(struct pt_regs *regs)
{
	nds32_disable_all_edm_bpc();
}

#endif //CONFIG_NDS32_EDM_SUP

void kgdb_arch_set_pc(struct pt_regs *regs, unsigned long pc)
{
	regs->NDS32_ipc = pc;
}
void arch_kgdb_breakpoint(void)
{
	asm __volatile__ (
		".global compiled_breakinst\n\t"
		"compiled_breakinst: break 0x1f\n\t");
}


int kgdb_arch_handle_exception(int exception_vector, int signo,
			       int err_code, char *remcom_in_buffer,
			       char *remcom_out_buffer,
			       struct pt_regs *linux_regs)
{
	long addr;
	char *ptr;


	switch (remcom_in_buffer[0]) {
	case 'k':
	case 'D':
	case 'c':
	case 's':
		if ((linux_regs->NDS32_ipc == (unsigned long)compiled_breakinst))
			linux_regs->NDS32_ipc += 4;

		kgdb_contthread = NULL;

		/*
		 * Try to read optional parameter, pc unchanged if no parm.
		 * If this was a compiled breakpoint, we need to move
		 * to the next instruction or we will just breakpoint
		 * over and over again.
		 */
		ptr = &remcom_in_buffer[1];
		if (kgdb_hex2long(&ptr, &addr)) {
			linux_regs->NDS32_ipc = addr;
		}
		linux_regs->NDS32_ipsw &= ~0x800;
		if (remcom_in_buffer[0] == 's') {
			linux_regs->NDS32_ipsw |= 0x800;
		}

		return 0;
	}

	return -1;
}

static int kgdb_notify(struct notifier_block *self,
		       unsigned long cmd, void *ptr)
{
	struct die_args *args = ptr;
	if (!user_mode(args->regs)) {
		kgdb_handle_exception(args->trapnr, args->signr,
				      args->err, args->regs);
		return NOTIFY_STOP;
	}
	return NOTIFY_DONE;
}

static struct notifier_block kgdb_notifier = {
	.notifier_call = kgdb_notify,
};

int kgdb_arch_init(void)
{
	register_die_notifier(&kgdb_notifier);
	return 0;
}

void kgdb_arch_exit(void)
{
	unregister_die_notifier(&kgdb_notifier);
}

struct kgdb_arch arch_kgdb_ops = {
#ifdef __NDS32_EL__
	.gdb_bpt_instr = {0xea, 0x1f},
#else
	.gdb_bpt_instr = {0x1f, 0xea},
#endif
#ifdef CONFIG_NDS32_EDM_SUP
	.flags = KGDB_HW_BREAKPOINT,
	.set_hw_breakpoint = nds32_kgdb_set_hw_break,
	.remove_hw_breakpoint = nds32_kgdb_remove_hw_break,
	.disable_hw_break = nds32_kgdb_disable_hw_debug,
	.remove_all_hw_break = nds32_remove_all_hw_break,
	.correct_hw_break = nds32_correct_hw_break,
#endif //CONFIG_NDS32_EDM_SUP
};
