/*
 * linux/arch/nds32/kernel/process.c
 */
/* Copyright (C) 1996-2000 Russell King - Converted to ARM.
 * Original Copyright (C) 1995  Linus Torvalds
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/* ============================================================================
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is process implementation for NDS32 architecture, and it
 *    is original referred from ARM.
 *
 *  Revision History:
 *
 *    Oct.02.2007     Initial ported by Tom, Shawn, Steven, and Harry.
 *    Oct.03.2007     Updated get_wchan() for info under /proc.
 *
 *  Note:
 *
 * ============================================================================
 */
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/kallsyms.h>
#include <asm/leds.h>
#include <asm/uaccess.h>
#include <asm/elf.h>
#include <asm/fpu.h>
#include <asm/proc-fns.h>
#include <linux/elf.h>
#include <linux/tick.h>
#include <linux/proc_fs.h>
#include <linux/ptrace.h>

extern const char *processor_modes[];
extern void setup_mm_for_reboot(char mode);
extern struct task_struct *_switch(struct task_struct *last,
				   struct thread_info *prev,
				   struct thread_info *next);

#ifndef CONFIG_UNLAZY_FPU
struct task_struct *last_task_used_math = NULL;
#endif
static volatile int hlt_counter;

#ifdef CONFIG_PROC_FS
struct proc_dir_entry *proc_dir_cpu;
EXPORT_SYMBOL(proc_dir_cpu);
#endif

extern inline void arch_reset(char mode)
{
	if (mode == 's') {
		/* Use cpu handler, jump to 0 */
		cpu_reset(0);
	}
}

#ifndef arch_idle
static inline void arch_idle(void)
{
	cpu_do_idle();
}
#endif

void disable_hlt(void)
{
	hlt_counter++;
}

EXPORT_SYMBOL(disable_hlt);

void enable_hlt(void)
{
	hlt_counter--;
}

EXPORT_SYMBOL(enable_hlt);

static int __init nohlt_setup(char *__unused)
{
	hlt_counter = 1;
	return 1;
}

static int __init hlt_setup(char *__unused)
{
	hlt_counter = 0;
	return 1;
}

__setup("nohlt", nohlt_setup);
__setup("hlt", hlt_setup);

/*
 * The following aren't currently used.
 */

void (*pm_idle) (void) = NULL;
EXPORT_SYMBOL(pm_idle);

void (*pm_power_off) (void);
EXPORT_SYMBOL(pm_power_off);

/*
 * This is our default idle handler.  We need to disable
 * interrupts here to ensure we don't miss a wakeup call.
 */
void default_idle(void)
{
	local_irq_disable();
	if (!need_resched() && !hlt_counter)
		arch_idle();
	local_irq_enable();
}

/*
 * The idle thread.  We try to conserve power, while trying to keep
 * overall latency low.  The architecture specific idle is passed
 * a value to indicate the level of "idleness" of the system.
 */
void cpu_idle(void)
{

	/* endless idle loop with no priority at all */
	while (1) {
		void (*idle) (void) = pm_idle;

		if (!idle)
			idle = default_idle;
		tick_nohz_idle_enter();
		leds_event(led_idle_start);
		while (!need_resched())
			idle();
		leds_event(led_idle_end);
		tick_nohz_idle_exit();
		preempt_enable_no_resched();
		schedule();
		preempt_disable();
	}
}

static char reboot_mode = 'h';

int __init reboot_setup(char *str)
{
	reboot_mode = str[0];
	return 1;
}

static int cpub_pwroff(void)
{
	return 0;
}

__setup("reboot=", reboot_setup);

void machine_halt(void)
{
	cpub_pwroff();
}

EXPORT_SYMBOL(machine_halt);

void machine_power_off(void)
{
	if (pm_power_off)
		pm_power_off();
}

EXPORT_SYMBOL(machine_power_off);

void machine_restart(char *__unused)
{
	/*
	 * Clean and disable cache, and turn off interrupts
	 */
	cpu_proc_fin();

	/*
	 * Tell the mm system that we are going to reboot -
	 * we may need it to insert some 1:1 mappings so that
	 * soft boot works.
	 */
	setup_mm_for_reboot(reboot_mode);

	/*
	 * Now call the architecture specific reboot code.
	 */
	arch_reset(reboot_mode);

	/*
	 * Whoops - the architecture was unable to reboot.
	 * Tell the user!
	 */
	mdelay(1000);
	printk("Reboot failed -- System halted\n");
	while (1) ;
}

EXPORT_SYMBOL(machine_restart);

void show_regs(struct pt_regs *regs)
{
	print_symbol("PC is at %s\n", instruction_pointer(regs));
	print_symbol("LR is at %s\n", regs->NDS32_lp);
	printk("pc : [<%08lx>]    lp : [<%08lx>]    %s\n"
	       "sp : %08lx  fp : %08lx  gp : %08lx\n",
	       instruction_pointer(regs),
	       regs->NDS32_lp, print_tainted(), regs->NDS32_sp,
	       regs->NDS32_fp, regs->NDS32_gp);
	printk("r25: %08lx  r24: %08lx\n", regs->NDS32_r25, regs->NDS32_r24);

	printk("r23: %08lx  r22: %08lx  r21: %08lx  r20: %08lx\n",
	       regs->NDS32_r23, regs->NDS32_r22,
	       regs->NDS32_r21, regs->NDS32_r20);
	printk("r19: %08lx  r18: %08lx  r17: %08lx  r16: %08lx\n",
	       regs->NDS32_r19, regs->NDS32_r18,
	       regs->NDS32_r17, regs->NDS32_r16);
	printk("r15: %08lx  r14: %08lx  r13: %08lx  r12: %08lx\n",
	       regs->NDS32_r15, regs->NDS32_r14,
	       regs->NDS32_r13, regs->NDS32_r12);
	printk("r11: %08lx  r10: %08lx  r9 : %08lx  r8 : %08lx\n",
	       regs->NDS32_r11, regs->NDS32_r10,
	       regs->NDS32_r9, regs->NDS32_r8);
	printk("r7 : %08lx  r6 : %08lx  r5 : %08lx  r4 : %08lx\n",
	       regs->NDS32_r7, regs->NDS32_r6, regs->NDS32_r5, regs->NDS32_r4);
	printk("r3 : %08lx  r2 : %08lx  r1 : %08lx  r0 : %08lx\n",
	       regs->NDS32_r3, regs->NDS32_r2, regs->NDS32_r1, regs->NDS32_r0);
	printk("  IRQs o%s  Segment %s\n",
	       interrupts_enabled(regs) ? "n" : "ff",
	       segment_eq(get_fs(), get_ds())? "kernel" : "user");
}

EXPORT_SYMBOL(show_regs);

#ifdef CONFIG_FPU
void show_fpregs(struct user_fp *regs)
{
	int i;

	for (i = 0; i < 8; i++) {
		unsigned long *p;
		char type;

		p = (unsigned long *)(regs->fpregs + i);

		switch (regs->ftype[i]) {
		case 1:
			type = 'f';
			break;
		case 2:
			type = 'd';
			break;
		case 3:
			type = 'e';
			break;
		default:
			type = '?';
			break;
		}
		if (regs->init_flag)
			type = '?';

		printk("  f%d(%c): %08lx %08lx %08lx%c",
		       i, type, p[0], p[1], p[2], i & 1 ? '\n' : ' ');
	}

	printk("FPSR: %08lx FPCR: %08lx\n",
	       (unsigned long)regs->fpsr, (unsigned long)regs->fpcr);
}
#endif

/*
 * Task structure and kernel stack allocation.
 */
static unsigned long *thread_info_head;
static unsigned int nr_thread_info;

#define EXTRA_TASK_STRUCT   4
#define ll_alloc_task_struct() ((struct thread_info *) __get_free_pages(GFP_KERNEL,1))
#define ll_free_task_struct(p) free_pages((unsigned long)(p),1)

struct thread_info *alloc_thread_info(struct task_struct *task)
{
	struct thread_info *thread = NULL;

	if (EXTRA_TASK_STRUCT) {
		unsigned long *p = thread_info_head;

		if (p) {
			thread_info_head = (unsigned long *)p[0];
			nr_thread_info -= 1;
		}
		thread = (struct thread_info *)p;
	}

	if (!thread)
		thread = ll_alloc_task_struct();

#ifdef CONFIG_MAGIC_SYSRQ
	/*
	 * The stack must be cleared if you want SYSRQ-T to
	 * give sensible stack usage information
	 */
	if (thread) {
		char *p = (char *)thread;
		memzero(p + KERNEL_STACK_SIZE, KERNEL_STACK_SIZE);
	}
#endif
	return thread;
}

void free_thread_info(struct thread_info *thread)
{
	if (EXTRA_TASK_STRUCT && nr_thread_info < EXTRA_TASK_STRUCT) {
		unsigned long *p = (unsigned long *)thread;
		p[0] = (unsigned long)thread_info_head;
		thread_info_head = p;
		nr_thread_info += 1;
	} else
		ll_free_task_struct(thread);
}

/*
 * Free current thread data structures etc..
 */
void exit_thread(void)
{
        user_direct_debug_exception_to_DIMU(current);
#if defined(CONFIG_FPU)
# ifndef CONFIG_UNLAZY_FPU
	if (last_task_used_math == current) {
		last_task_used_math = NULL;
	}
# endif
#endif
}

void flush_thread(void)
{
	struct task_struct *tsk = current;

	memset(&tsk->thread.debug, 0, sizeof(struct debug_info));
#if defined(CONFIG_FPU)
	clear_fpu(task_pt_regs(tsk));
	clear_used_math();
# ifndef CONFIG_UNLAZY_FPU
	if (last_task_used_math == current) {
		last_task_used_math = NULL;
	}
# endif
#endif


}

void release_thread(struct task_struct *dead_task)
{
}

asmlinkage void ret_from_fork(void) __asm__("ret_from_fork");
int
copy_thread(unsigned long clone_flags, unsigned long stack_start,
	    unsigned long stk_sz, struct task_struct *p)
{
	struct thread_info *thread = task_thread_info(p);
	struct pt_regs *childregs;
	struct cpu_context_save *cpu_cs;

	childregs =
	    ((struct pt_regs *)((unsigned long)thread + THREAD_SIZE - 8)) - 1;

	thread->sp_save = cpu_cs = ((struct cpu_context_save *)(childregs)) - 1;

	/* FIXME */
	if (unlikely(p->flags & PF_KTHREAD)) {
		/* TODO: is this necessary?? */
		memset(childregs, 0, sizeof(struct pt_regs));
		/* kernel thread fn */
		cpu_cs->r6 = stack_start;
		/* kernel thread argument */
		cpu_cs->r7 = stk_sz;
	} else {
		*childregs = *current_pt_regs();
		if (stack_start)
			childregs->NDS32_sp = stack_start;
		/* child get zero as ret. */
		childregs->NDS32_r0 = 0;
		/* TODO: verify the necessity */
		childregs->NDS32_osp = 0;
		cpu_cs->r6 = 0;
		if (clone_flags & CLONE_SETTLS)
			childregs->NDS32_r25 = childregs->NDS32_r3;
	}
	/* cpu context switching  */
	thread->sp_save->pc = (unsigned long)ret_from_fork;

#ifdef CONFIG_FPU
	if (used_math()) {
# ifdef CONFIG_UNLAZY_FPU
		unlazy_fpu(current);
# else
		preempt_disable();
		if (last_task_used_math == current)
			save_fpu(current);
		preempt_enable();
# endif
		p->thread.fpu = current->thread.fpu;
		clear_fpu(task_pt_regs(p));
		set_stopped_child_used_math(p);
	}
#endif


#ifdef CONFIG_HWZOL
	childregs->NDS32_lb = 0;
	childregs->NDS32_le = 0;
	childregs->NDS32_lc = 0;
#endif

        clear_tsk_thread_flag(p, TIF_DEBUG);
        if ((clone_flags & CLONE_PTRACE) && test_thread_flag(TIF_DEBUG)) {
                user_direct_debug_exception_to_IVB(p);
        }

	return 0;
}

struct task_struct *__switch_to(struct task_struct *last,
				struct thread_info *prev,
				struct thread_info *next)
{
#ifdef CONFIG_FPU
#ifdef CONFIG_UNLAZY_FPU
	unlazy_fpu(prev->task);
#endif
	if (!(next->task->flags & PF_KTHREAD)) {
		clear_fpu(task_pt_regs(next->task));
	}
#endif
	return _switch(last, prev, next);
}

/*
 * fill in the fpe structure for a core dump...
 */
int dump_fpu(struct pt_regs *regs, elf_fpregset_t * fpu)
{
	int fpvalid = 0;

#if 0				// XXX defined(CONFIG_FPU)
	struct task_struct *tsk = current;

	fpvalid = !!tsk_used_math(tsk);
	if (fpvalid) {
		unlazy_fpu(tsk, regs);
		memcpy(fpu, &tsk->thread.fpu, sizeof(*fpu));
	}
#endif

	return fpvalid;
}

EXPORT_SYMBOL(dump_fpu);

unsigned long get_wchan(struct task_struct *p)
{
	unsigned long fp, lr;
	unsigned long stack_start, stack_end;
	int count = 0;

	if (!p || p == current || p->state == TASK_RUNNING)
		return 0;

#ifdef CONFIG_FRAME_POINTER
	stack_start = (unsigned long)end_of_stack(p);
	stack_end = (unsigned long)task_stack_page(p) + THREAD_SIZE;

	fp = thread_saved_fp(p);
	do {
		if (fp < stack_start || fp > stack_end)
			return 0;
		lr = ((unsigned long *)fp)[0];
		if (!in_sched_functions(lr))
			return lr;
		fp = *(unsigned long *)(fp + 4);
	} while (count++ < 16);
	return 0;
#else
	return 0;
#endif
}

EXPORT_SYMBOL(get_wchan);

extern int do_elf_check_arch(const struct elf32_hdr *hdr);

int elf_check_arch(const struct elf32_hdr *hdr)
{
	return do_elf_check_arch(hdr);
}
