/*
 * linux/arch/nds32/kernel/ptrace.c
 */
/* By Ross Biro 1/23/92
 * edited by Linus Torvalds
 * ARM modifications Copyright (C) 2000 Russell King
 * NDS32 modifications Copyright (C) 2007 Harry Pan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/* ============================================================================
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is for Andes NDS32 architecture.
 *
 *  Revision History:
 *
 *    Jul.31.2007     Initial ported by Tom, Shawn, and Steven,
 *                    revised by Harry.
 *                    Current implmentation is based on Andes Instruction
 *                    Set Architecture Specification (AS-0001-0001)
 *                    version:3.7 date:7-20-2007.
 *                    It is original taken from ARM, then fit to NDS32.
 *    Aug.15.2007     Mainly updated breakpoint handling base on NDS32 ISA.
 *                    I also did code revise.
 *    Nov.14.2007     Fixed up ptrace_set_bpt() while handling 16-bit insn.
 *    Nov.27.2007     Added checking duplicate breakpoints in
 *                    add_breakpoint(), based on Shawn's idea.
 *    Dec.06.2007     Added get_user_gpr().
 *    Apr.17.2009     Added support for FPU and Audio regs.
 *		      Added support for PTRACE_GETFPREGS, PTRACE_SETFPREGS,
 *		      PTRACE_GETAUREGS, PTRACE_SETAUREGS
 *
 *  Note:
 *
 *    Current layout: 0-31 GR, 32-34 SPR, 35-... SR, index start from zero.
 *
 *      +----------+-----+--------------+---+--------+
 *      |    GR    | SPR |      SR      |...|  Audio |
 *      +----------+-----+--------------+---+--------+
 *      0          32    35              ...500    531
 *
 * ============================================================================
 */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/ptrace.h>
#include <linux/user.h>
#include <linux/security.h>
#include <linux/init.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/traps.h>
#include <asm/fpu.h>
#include <nds32_intrinsic.h>

#include "ptrace.h"

#ifdef __NDS32_EL__
#define BREAKINST	0x01EA
#else
#define BREAKINST	0xEA01
#endif

struct pt_regs_offset {
        const char *name;
        int offset;
};

#define REG_OFFSET_NAME(r) \
        {.name = #r, .offset = offsetof(struct pt_regs, NDS32_##r)}
#define REG_OFFSET_END {.name = NULL, .offset = 0}

static const struct pt_regs_offset regoffset_table[] = {
        REG_OFFSET_NAME(ir0),
        REG_OFFSET_NAME(ipsw),
        REG_OFFSET_NAME(ipc),
        REG_OFFSET_NAME(sp),
        REG_OFFSET_NAME(ORIG_r0),
        REG_OFFSET_NAME(pipsw),
        REG_OFFSET_NAME(pipc),
        REG_OFFSET_NAME(pp0),
        REG_OFFSET_NAME(pp1),
#if defined(CONFIG_HWZOL)
        REG_OFFSET_NAME(lb),
        REG_OFFSET_NAME(le),
        REG_OFFSET_NAME(lc),
#endif
        REG_OFFSET_NAME(r0),
        REG_OFFSET_NAME(r1),
        REG_OFFSET_NAME(r2),
        REG_OFFSET_NAME(r3),
        REG_OFFSET_NAME(r4),
        REG_OFFSET_NAME(r5),
        REG_OFFSET_NAME(r6),
        REG_OFFSET_NAME(r7),
        REG_OFFSET_NAME(r8),
        REG_OFFSET_NAME(r9),
        REG_OFFSET_NAME(r10),
        REG_OFFSET_NAME(r11),
        REG_OFFSET_NAME(r12),
        REG_OFFSET_NAME(r13),
        REG_OFFSET_NAME(r14),
        REG_OFFSET_NAME(r15),
        REG_OFFSET_NAME(r16),
        REG_OFFSET_NAME(r17),
        REG_OFFSET_NAME(r18),
        REG_OFFSET_NAME(r19),
        REG_OFFSET_NAME(r20),
        REG_OFFSET_NAME(r21),
        REG_OFFSET_NAME(r22),
        REG_OFFSET_NAME(r23),
        REG_OFFSET_NAME(r24),
        REG_OFFSET_NAME(r25),
        REG_OFFSET_NAME(fp),
        REG_OFFSET_NAME(gp),
        REG_OFFSET_NAME(lp),
        REG_OFFSET_NAME(FUCOP_CTL),
        REG_OFFSET_NAME(osp),
        REG_OFFSET_END,
};

/**
 * regs_query_register_offset() - query register offset from its name
 * @name:       the name of a register
 *
 * regs_query_register_offset() returns the offset of a register in struct
 * pt_regs from its name. If the name is invalid, this returns -EINVAL;
 */
int regs_query_register_offset(const char *name)
{
        const struct pt_regs_offset *roff;
        for (roff = regoffset_table; roff->name != NULL; roff++)
                if (!strcmp(roff->name, name))
                        return roff->offset;
        return -EINVAL;
}

/**
 * regs_query_register_name() - query register name from its offset
 * @offset:     the offset of a register in struct pt_regs.
 *
 * regs_query_register_name() returns the name of a register from its
 * offset in struct pt_regs. If the @offset is invalid, this returns NULL;
 */
const char *regs_query_register_name(unsigned int offset)
{
        const struct pt_regs_offset *roff;
        for (roff = regoffset_table; roff->name != NULL; roff++)
                if (roff->offset == offset)
                        return roff->name;
        return NULL;
}

/**
 * regs_within_kernel_stack() - check the address in the stack
 * @regs:      pt_regs which contains kernel stack pointer.
 * @addr:      address which is checked.
 *
 * regs_within_kernel_stack() checks @addr is within the kernel stack page(s).
 * If @addr is within the kernel stack, it returns true. If not, returns false.
 */
bool regs_within_kernel_stack(struct pt_regs *regs, unsigned long addr)
{
        return ((addr & ~(THREAD_SIZE - 1))  ==
                (kernel_stack_pointer(regs) & ~(THREAD_SIZE - 1)));
}

/**
 * regs_get_kernel_stack_nth() - get Nth entry of the stack
 * @regs:       pt_regs which contains kernel stack pointer.
 * @n:          stack entry number.
 *
 * regs_get_kernel_stack_nth() returns @n th entry of the kernel stack which
 * is specified by @regs. If the @n th entry is NOT in the kernel stack,
 * this returns 0.
 */
unsigned long regs_get_kernel_stack_nth(struct pt_regs *regs, unsigned int n)
{
        unsigned long *addr = (unsigned long *)kernel_stack_pointer(regs);
        addr += n;
        if (regs_within_kernel_stack(regs, (unsigned long)addr))
                return *addr;
        else
                return 0;
}

/* get_user_reg()
 *
 * This routine will get a word off of the processes privileged stack.
 * the offset is how far from the base addr as stored in the THREAD.
 * this routine assumes that all the privileged stacks are in our
 * data space.
 */
static inline unsigned int get_user_reg(struct task_struct *task, int offset)
{
	return task_pt_regs(task)->uregs[offset];
}

/* put_user_reg()
 *
 * this routine will put a word on the processes privileged stack.
 * the offset is how far from the base addr as stored in the THREAD.
 * this routine assumes that all the privileged stacks are in our
 * data space.
 */
static inline int put_user_reg(struct task_struct *task, int offset, long data)
{
	struct pt_regs newregs, *regs = task_pt_regs(task);
	int ret = -EINVAL;

	newregs = *regs;
	newregs.uregs[offset] = data;

	if (valid_user_regs(&newregs)) {
		regs->uregs[offset] = data;
		ret = 0;
	}

	return ret;
}

void arch_ptrace_attach(struct task_struct *child)
{
        user_direct_debug_exception_to_IVB(child);
}

/*
 * Called by kernel/ptrace.c when detaching..
 *
 * Make sure the single step bit is not set and
 * restore the del_sel bit.
 */
void ptrace_disable(struct task_struct *child)
{
	user_disable_single_step(child);
        user_direct_debug_exception_to_DIMU(child);
}

/*
 * Handle hitting a breakpoint.
 */
void send_sigtrap(struct task_struct *tsk, struct pt_regs *regs,
		  int error_code, int si_code)
{
	struct siginfo info;

	tsk->thread.trap_no = 1;
	tsk->thread.error_code = error_code;

	memset(&info, 0, sizeof(info));
	info.si_signo = SIGTRAP;
	info.si_code = si_code;

	info.si_addr = (void __user *)instruction_pointer(regs);

	/* Send us the fake SIGTRAP */
	force_sig_info(SIGTRAP, &info, tsk);
}

/* ptrace_read_user()
 *
 * Read the word at offset "off" into the "struct user".  We
 * actually access the pt_regs stored on the kernel stack.
 */
static int
ptrace_read_user(struct task_struct *tsk, unsigned long off,
		 unsigned long __user * ret)
{
	unsigned long tmp = 0;

	if (off < 500) {
		if (off & 3)
			return -EIO;

		if (off < sizeof(struct pt_regs))
			tmp = get_user_reg(tsk, off >> 2);

		return put_user(tmp, ret);
	} else if (off < 532) {
		return put_user(tmp, ret);
	} else
		return -EIO;
}

/* ptrace_write_user()
 *
 * Write the word at offset "off" into "struct user".  We
 * actually access the pt_regs stored on the kernel stack.
 */
static int
ptrace_write_user(struct task_struct *tsk, unsigned long off, unsigned long val)
{
	if (off < 500) {
		if (off & 3)
			return -EIO;

		if (off >= sizeof(struct pt_regs))
			return 0;

		return put_user_reg(tsk, off >> 2, val);
	} else if (off < 532) {
		return 0;
	} else
		return -EIO;
}

/* ptrace_getregs()
 *
 * Get all user integer registers.
 */
static int ptrace_getregs(struct task_struct *tsk, void __user * uregs)
{
	struct pt_regs *regs = task_pt_regs(tsk);

	return copy_to_user(uregs, regs, sizeof(struct pt_regs)) ? -EFAULT : 0;
}

/* ptrace_setregs()
 *
 * Set all user integer registers.
 */
static int ptrace_setregs(struct task_struct *tsk, void __user * uregs)
{
	struct pt_regs newregs;
	int ret;

	ret = -EFAULT;
	if (copy_from_user(&newregs, uregs, sizeof(struct pt_regs)) == 0) {
		struct pt_regs *regs = task_pt_regs(tsk);

		ret = -EINVAL;
		if (valid_user_regs(&newregs)) {
			*regs = newregs;
			ret = 0;
		}
	}

	return ret;
}

/* ptrace_getfpregs()
 *
 * Get the child FPU state.
 */
static int ptrace_getfpregs(struct task_struct *tsk, void __user * ufpregs)
{
#ifdef CONFIG_FPU
	if (used_math()) {
# ifdef CONFIG_UNLAZY_FPU
		unlazy_fpu(tsk);
# else
		preempt_disable();
		if (last_task_used_math == tsk)
			save_fpu(tsk);
		preempt_enable();
# endif
		return copy_to_user(ufpregs, &tsk->thread.fpu,
				    sizeof(struct fpu_struct)) ? -EFAULT : 0;
	} else {
		/* First time FPU user.  */
		memset(ufpregs, -1, sizeof(struct fpu_struct));
		return 0;
	}

#else
	return -EFAULT;
#endif
}

/*
 * Set the child FPU state.
 */
static int ptrace_setfpregs(struct task_struct *tsk, void __user * ufpregs)
{
#ifdef CONFIG_FPU
	int ret;

# ifndef CONFIG_UNLAZY_FPU
	if (last_task_used_math == tsk)
# endif
		clear_fpu(task_pt_regs(tsk));

	ret =
	    copy_from_user(&tsk->thread.fpu, ufpregs,
			   sizeof(struct fpu_struct)) ? -EFAULT : 0;

	if (!ret && !used_math()) {
		/* First time Audio user.  */
		set_used_math();
	}

	return ret;
#else
	return -EFAULT;
#endif
}



/* do_ptrace()
 *
 * Provide ptrace defined service.
 */
long arch_ptrace(struct task_struct *child, long request, unsigned long addr,
		 unsigned long data)
{
	int ret;

	switch (request) {
	case PTRACE_PEEKUSR:
		ret =
		    ptrace_read_user(child, addr, (unsigned long __user *)data);
		break;

	case PTRACE_POKEUSR:
		ret = ptrace_write_user(child, addr, data);
		break;

	case PTRACE_GETREGS:
		ret = ptrace_getregs(child, (void __user *)data);
		break;

	case PTRACE_SETREGS:
		ret = ptrace_setregs(child, (void __user *)data);
		break;

	case PTRACE_GETFPREGS:
		ret = ptrace_getfpregs(child, (void __user *)data);
		break;


	case PTRACE_SETFPREGS:
		ret = ptrace_setfpregs(child, (void __user *)data);
		break;
/*
        case PTRACE_GET_THREAD_AREA:
            ret = put_user(task_thread_info(child)->tp_value,
                       (unsigned long __user *) data);
            break;
*/

	default:
		ret = ptrace_request(child, request, addr, data);
		break;
	}

	return ret;
}

void user_enable_single_step(struct task_struct *child)
{
	struct pt_regs *regs;
	regs = task_pt_regs(child);
	regs->NDS32_ipsw |= PSW_mskHSS;
	set_tsk_thread_flag(child, TIF_SINGLESTEP);
}

void user_disable_single_step(struct task_struct *child)
{
	struct pt_regs *regs;
	regs = task_pt_regs(child);
	regs->NDS32_ipsw &= ~PSW_mskHSS;
	clear_tsk_thread_flag(child, TIF_SINGLESTEP);
}

#define DEH_SEL_UNKNOWN_VAL 0xff
static spinlock_t edm_ctl_lock;
static long edm_ctl_count = 0;
static u8 orig_deh_sel = DEH_SEL_UNKNOWN_VAL;
void user_direct_debug_exception_to_IVB(struct task_struct *child)
{
        unsigned long edm_ctl = 0;

        if (!child || !test_and_set_tsk_thread_flag(child, TIF_DEBUG)) {
                spin_lock(&edm_ctl_lock);

                edm_ctl_count++;

                if (edm_ctl_count == 1 && orig_deh_sel == DEH_SEL_UNKNOWN_VAL) {
                        edm_ctl = __nds32__mfsr(NDS32_SR_EDM_CTL);
                        orig_deh_sel = (edm_ctl & EDM_CTL_mskDEH_SEL) >>
                                       EDM_CTL_offDEH_SEL;
                        edm_ctl = edm_ctl & ~EDM_CTL_mskDEH_SEL;
                        __nds32__mtsr(edm_ctl, NDS32_SR_EDM_CTL);
                }

                spin_unlock(&edm_ctl_lock);
        }
}

void user_direct_debug_exception_to_DIMU(struct task_struct *child)
{
        unsigned long edm_ctl = 0;

        if (!child || test_and_clear_tsk_thread_flag(child, TIF_DEBUG)) {
                spin_lock(&edm_ctl_lock);

                edm_ctl_count--;

                /* edm_ctl_count should be pairs through IVB to increase it */
                WARN_ON(edm_ctl_count < 0);

                if (edm_ctl_count == 0 && orig_deh_sel != DEH_SEL_UNKNOWN_VAL) {
                        edm_ctl = __nds32__mfsr(NDS32_SR_EDM_CTL);
                        edm_ctl = (edm_ctl & ~EDM_CTL_mskDEH_SEL) |
                                  (orig_deh_sel << EDM_CTL_offDEH_SEL);
                        __nds32__mtsr(edm_ctl, NDS32_SR_EDM_CTL);
                        orig_deh_sel = DEH_SEL_UNKNOWN_VAL;
                }

                spin_unlock(&edm_ctl_lock);
        }
}

/* sys_trace()
 *
 * syscall trace handler.
 */
static inline void do_syscall_trace(void)
{
	if (!test_thread_flag(TIF_SYSCALL_TRACE))
		return;
	if (!(current->ptrace & PT_PTRACED))
		return;

	/* the 0x80 provides a way for the tracing parent to distinguish
	   between a syscall stop and SIGTRAP delivery */
	ptrace_notify(SIGTRAP | ((current->ptrace & PT_TRACESYSGOOD)
				 ? 0x80 : 0));
	/*
	 * this isn't the same as continuing with a signal, but it will do
	 * for normal use.  strace only continues with a signal if the
	 * stopping signal is not SIGTRAP.  -brl
	 */
	if (current->exit_code) {
		send_sig(current->exit_code, current, 1);
		current->exit_code = 0;
	}
}

asmlinkage int syscall_trace_enter(int syscall, struct pt_regs *regs)
{
	int orig_r0;

	orig_r0 = regs->NDS32_ORIG_r0;
	regs->NDS32_ORIG_r0 = syscall;
	do_syscall_trace();
	syscall = regs->NDS32_ORIG_r0;
	regs->NDS32_ORIG_r0 = orig_r0;
	return syscall & 0xfff;
}

asmlinkage void syscall_trace_leave(struct pt_regs *regs)
{
	do_syscall_trace();

	/* synthsize a single-step */
	if (test_thread_flag(TIF_SINGLESTEP)) {
		printk(KERN_INFO "synthsize trap (tf=0x%0x\n",
		       (unsigned int)current_thread_info()->flags);
		send_sigtrap(current, regs, 0, TRAP_BRKPT);
	}
}

