/*
 *  linux/arch/nds32/kernel/setup.c
 *
 *  Copyright (C) 1995-2001 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/* ============================================================================
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *  This program is for Andes NDS32 architecture.
 *
 *  Revision History:
 *
 *  Jul.05.2007 Initial ported by Tom, revised and patched for KGDB
 *    by Harry.
 *  Aug.26.2008 Some reworks on CPU info output for SMP.
 *
 *  Note:
 *
 * ============================================================================
 */
#include <linux/cpu.h>
#include <linux/bootmem.h>
#include <linux/seq_file.h>
#include <linux/memblock.h>
#include <linux/console.h>
#include <linux/screen_info.h>
#include <linux/delay.h>
#include <linux/of_fdt.h>
#include <linux/of_platform.h>
#include <linux/of_address.h>
#include <linux/dma-mapping.h>
#include <asm/cpu.h>
#include <asm/setup.h>
#include <asm/cacheflush.h>
#include <asm/cpuver.h>
#include <asm/procinfo.h>
#include <asm/proc-fns.h>
#include <asm/traps.h>
#include <asm/fpu.h>
#include <asm/cache_info.h>
#include <asm/elf.h>
#include <nds32_intrinsic.h>
#include "vmlinux.h"

extern struct mm_struct init_mm;
unsigned long cpu_id, cpu_rev, cpu_cfgid;
char *endianness = NULL;

unsigned int elf_hwcap;
unsigned int __atags_pointer __initdata;
EXPORT_SYMBOL(elf_hwcap);


char elf_platform[ELF_PLATFORM_SIZE];
EXPORT_SYMBOL(elf_platform);


static struct proc_info_item proc_info;
char __initdata cmd_line[COMMAND_LINE_SIZE] = CONFIG_CMDLINE;

DEFINE_PER_CPU(struct cpuinfo_nds32, cpu_data);

extern void __init early_init_devtree(void *params);
extern void __init copy_fdt(void);
/*
 * The following string table, must sync with HWCAP_xx bitmask,
 * which is defined in <asm/procinfo.h>
 */
static const char *hwcap_str[] = {
	"mfusr_pc",
	"perf1",
	"perf2",
	"fpu",
	"audio",
	"16b",
	"string",
	"reduced_regs",
	"video",
	"encrypt",
	"edm",
	"lmdma",
	"pfm",
	"hsmp",
	"trace",
	"div",
	"mac",
	"l2c",
	"dx_regs",
	"v2",
	NULL,
};

struct cache_info L1_cache_info[2];

static void __init dump_cpu_info(int cpu)
{
	int i = 0, aliasing_num;
	printk("CPU%d Features: ", cpu);

	for (i = 0; hwcap_str[i]; i++) {
		if (elf_hwcap & (1 << i))
			printk("%s ", hwcap_str[i]);
	}

	printk("\n");

	L1_cache_info[ICACHE].cache_type = ICACHE;
	L1_cache_info[ICACHE].ways = CACHE_WAY(ICACHE);
	L1_cache_info[ICACHE].way_bits = ilog2(CACHE_WAY(ICACHE));
	L1_cache_info[ICACHE].line_size = CACHE_LINE_SIZE(ICACHE);
	L1_cache_info[ICACHE].line_bits = ilog2(CACHE_LINE_SIZE(ICACHE));
	L1_cache_info[ICACHE].sets = CACHE_SET(ICACHE);
	L1_cache_info[ICACHE].set_bits = ilog2(CACHE_SET(ICACHE));
	L1_cache_info[ICACHE].size =
	    CACHE_SET(ICACHE) * CACHE_WAY(ICACHE) * CACHE_LINE_SIZE(ICACHE) /
	    1024;
	printk("L1I:%dKB/%dS/%dW/%dB\n", L1_cache_info[ICACHE].size,
	       L1_cache_info[ICACHE].sets, L1_cache_info[ICACHE].ways,
	       L1_cache_info[ICACHE].line_size);
	aliasing_num =
	    L1_cache_info[ICACHE].size * 1024 / PAGE_SIZE /
	    L1_cache_info[ICACHE].ways;
	if (aliasing_num & 1 && aliasing_num != 1)
		printk
		    ("%s: not alising:%d, it should be multiple of 2 if it ia aliasing cache.\n",
		     __func__, aliasing_num);
	L1_cache_info[ICACHE].aliasing_num = aliasing_num;
	L1_cache_info[ICACHE].aliasing_mask = (aliasing_num - 1) << PAGE_SHIFT;
	L1_cache_info[ICACHE].not_aliasing_mask =
	    ~L1_cache_info[ICACHE].aliasing_mask;
	L1_cache_info[DCACHE].cache_type = DCACHE;
	L1_cache_info[DCACHE].ways = CACHE_WAY(DCACHE);
	L1_cache_info[DCACHE].way_bits = ilog2(CACHE_WAY(DCACHE));
	L1_cache_info[DCACHE].line_size = CACHE_LINE_SIZE(DCACHE);
	L1_cache_info[DCACHE].line_bits = ilog2(CACHE_LINE_SIZE(DCACHE));
	L1_cache_info[DCACHE].sets = CACHE_SET(DCACHE);
	L1_cache_info[DCACHE].set_bits = ilog2(CACHE_SET(DCACHE));
	L1_cache_info[DCACHE].size =
	    CACHE_SET(DCACHE) * CACHE_WAY(DCACHE) * CACHE_LINE_SIZE(DCACHE) /
	    1024;
	printk("L1D:%dKB/%dS/%dW/%dB\n", L1_cache_info[DCACHE].size,
	       L1_cache_info[DCACHE].sets, L1_cache_info[DCACHE].ways,
	       L1_cache_info[DCACHE].line_size);
	aliasing_num =
	    L1_cache_info[DCACHE].size * 1024 / PAGE_SIZE /
	    L1_cache_info[DCACHE].ways;
#ifdef CONFIG_HIGHMEM
	if (aliasing_num > 1 && CONFIG_HIGHMEM)
		WARN(1,
		     "%s: HIGHMEM is not supported for alising VIPT cache. aliasing_num:%d\n",
		     __func__, aliasing_num);
#else
	if (aliasing_num & 1 && aliasing_num != 1)
		printk
		    ("%s: not alising:%d, it should be multiple of 2 if it ia aliasing cache.\n",
		     __func__, aliasing_num);
#endif
	L1_cache_info[DCACHE].aliasing_num = aliasing_num;
	L1_cache_info[DCACHE].aliasing_mask = (aliasing_num - 1) << PAGE_SHIFT;
	L1_cache_info[DCACHE].not_aliasing_mask =
	    ~L1_cache_info[DCACHE].aliasing_mask;
#ifdef CONFIG_CPU_DCACHE_WRITETHROUGH
	printk("L1 D-Cache is WRITE-THROUGH\n");
#else
	printk("L1 D-Cache is WRITE-BACK\n");
#endif

	if ((GET_CPU_VER() & CPU_VER_mskCFGID_COP) &&
	    (GET_FUCOP_EXIST() &
	     (FUCOP_EXIST_mskCP0ISFPU | FUCOP_EXIST_mskCP0EX))) {
                /* Disable fpu and enable when it is used. */
                SET_FUCOP_CTL(GET_FUCOP_CTL() & ~FUCOP_CTL_mskCP0EN);
#ifdef CONFIG_FPU
                printk("FPU is able to use.\n");
#endif
        }
}
#ifdef CONFIG_CACHE_L2

void __iomem *atl2c_base;
static const struct of_device_id atl2c_ids[] __initconst = {
	{ .compatible = "andestech,atl2c", }
};

static int __init atl2c_of_init(void)
{
	struct device_node *np;
	struct resource res;
	unsigned long tmp = 0;
	unsigned long l2set, l2way, l2clsz;

	if (!(GET_MSC_CFG() & MSC_CFG_mskL2C))
		return -ENODEV;

	np = of_find_matching_node(NULL, atl2c_ids);
	if (!np)
		return -ENODEV;

	if (of_address_to_resource(np, 0, &res))
		return -ENODEV;

	atl2c_base = ioremap(res.start, resource_size(&res));
	if (!atl2c_base)
		return -ENOMEM;

	l2set =
	    64 << ((L2C_R_REG(L2_CA_CONF_OFF) & L2_CA_CONF_mskL2SET) >>
		   L2_CA_CONF_offL2SET);
	l2way =
	    1 +
	    ((L2C_R_REG(L2_CA_CONF_OFF) & L2_CA_CONF_mskL2WAY) >>
	     L2_CA_CONF_offL2WAY);
	l2clsz =
	    4 << ((L2C_R_REG(L2_CA_CONF_OFF) & L2_CA_CONF_mskL2CLSZ) >>
		  L2_CA_CONF_offL2CLSZ);
	printk("L2:%luKB/%luS/%luW/%luB\n",
	       l2set * l2way * l2clsz / 1024, l2set, l2way, l2clsz);

	tmp = L2C_R_REG(L2CC_PROT_OFF);
	tmp &= ~L2CC_PROT_mskMRWEN;
	L2C_W_REG(L2CC_PROT_OFF, tmp);

	tmp = L2C_R_REG(L2CC_SETUP_OFF);
	tmp &= ~L2CC_SETUP_mskPART;
	L2C_W_REG(L2CC_SETUP_OFF, tmp);

	tmp = L2C_R_REG(L2CC_CTRL_OFF);
	tmp |= L2CC_CTRL_mskEN;
	L2C_W_REG(L2CC_CTRL_OFF, tmp);

	return 0;
}
subsys_initcall(atl2c_of_init);
#endif
static void __init setup_cpuinfo(void)
{
	unsigned long tmp = 0;
	struct proc_info_list *list;
	extern struct proc_info_list __proc_info_begin, __proc_info_end;

	register unsigned coreid = GET_CPU_VER();
	for (list = &__proc_info_begin; list < &__proc_info_end; list++)
//              if (list->cpu_val == (GET_CPU_VER() & CPU_VER_mskCPUID))
		if (list->cpu_val == (coreid & list->cpu_mask))
			break;
	/*
	 * If the architecture type is not recognised, then we
	 * can co nothing...
	 */
	if (list >= &__proc_info_end) {
		printk
		    ("Processor configuration botched (CPU_VER 0x%lx), unable to continue.\n",
		     GET_CPU_VER());
		while (1) ;
	}

	proc_info = *list->info;

	cpu_dcache_inval_all();	// XXX head.S turn $$ on, need change.
	cpu_icache_inval_all();
	DSB();
	ISB();

	cpu_id = GET_CPU_ID();
	cpu_rev = GET_CPU_REV();
	cpu_cfgid = GET_CPU_CFGID();

	printk("CPU: %s, %s %s, CPU_VER 0x%08lx(id %lu, rev %lu, cfg %lu)\n",
	       list->arch_name,
	       proc_info.manufacturer, proc_info.cpu_name,
	       GET_CPU_VER(), cpu_id, cpu_rev, cpu_cfgid);

	elf_hwcap |= HWCAP_MFUSR_PC;

	if (((GET_MSC_CFG() & MSC_CFG_mskBASEV) >> MSC_CFG_offBASEV) == 0) {
		if (GET_MSC_CFG() & MSC_CFG_mskDIV)
			elf_hwcap |= HWCAP_DIV;

		if ((GET_MSC_CFG() & MSC_CFG_mskMAC)
		    || (cpu_id == 12 && cpu_rev < 4))
			elf_hwcap |= HWCAP_MAC;
	} else {
		elf_hwcap |= HWCAP_V2;
		elf_hwcap |= HWCAP_DIV;
		elf_hwcap |= HWCAP_MAC;
	}

	if (cpu_cfgid & 0x0001)
		elf_hwcap |= HWCAP_EXT;

	if (cpu_cfgid & 0x0002)
		elf_hwcap |= HWCAP_BASE16;

	if (cpu_cfgid & 0x0004)
		elf_hwcap |= HWCAP_EXT2;

	if (cpu_cfgid & 0x0008)
		elf_hwcap |= HWCAP_FPU;

	if (cpu_cfgid & 0x0010)
		elf_hwcap |= HWCAP_STRING;

	if (GET_MMU_CFG() & MMU_CFG_mskDE)
		endianness = "MSB";
	else
		endianness = "LSB";

	if (GET_MSC_CFG() & MSC_CFG_mskEDM)
		elf_hwcap |= HWCAP_EDM;

	if (GET_MSC_CFG() & MSC_CFG_mskLMDMA)
		elf_hwcap |= HWCAP_LMDMA;

	if (GET_MSC_CFG() & MSC_CFG_mskPFM)
		elf_hwcap |= HWCAP_PFM;

	if (GET_MSC_CFG() & MSC_CFG_mskHSMP)
		elf_hwcap |= HWCAP_HSMP;

	if (GET_MSC_CFG() & MSC_CFG_mskTRACE)
		elf_hwcap |= HWCAP_TRACE;

	if (GET_MSC_CFG() & MSC_CFG_mskAUDIO)
		elf_hwcap |= HWCAP_AUDIO;

	if (GET_MSC_CFG() & MSC_CFG_mskL2C)
		elf_hwcap |= HWCAP_L2C;

	tmp = GET_CACHE_CTL();
#ifndef CONFIG_CPU_DCACHE_DISABLE
	tmp |= CACHE_CTL_mskDC_EN;
#endif

#ifndef CONFIG_CPU_ICACHE_DISABLE
	tmp |= CACHE_CTL_mskIC_EN;
#endif
	SET_CACHE_CTL(tmp);
	DSB();
	ISB();

	sprintf(elf_platform, "%s %s", list->elf_name, endianness);

	dump_cpu_info(smp_processor_id());
}

static unsigned long __init setup_memory(void)
{
	unsigned long bootmap_size;
	unsigned long ram_start_pfn;
	unsigned long free_ram_start_pfn;
	phys_addr_t memory_start, memory_end;
	struct memblock_region *region;

	memory_end = memory_start = 0;

	/* Find main memory where is the kernel */
	for_each_memblock(memory, region) {
		memory_start = region->base;
		memory_end = region->base + region->size;
		printk(KERN_INFO "%s: Memory: 0x%x-0x%x\n", __func__,
		       memory_start, memory_end);
	}

	if (!memory_end) {
		panic("No memory!");
	}

	ram_start_pfn = PFN_UP(memblock_start_of_DRAM());
	/* free_ram_start_pfn is first page after kernel */
	free_ram_start_pfn = PFN_UP(__pa(&_end));
	max_pfn = PFN_DOWN(memblock_end_of_DRAM());

	/* it could update max_pfn */
	if (max_pfn - ram_start_pfn <= MAXMEM_PFN)
		max_low_pfn = max_pfn;
	else {
		max_low_pfn = MAXMEM_PFN + ram_start_pfn;
#ifndef CONFIG_HIGHMEM
		max_pfn = MAXMEM_PFN + ram_start_pfn;
#endif
	}
	/* high_memory is related with VMALLOC */
	high_memory = (void *)__va(max_low_pfn * PAGE_SIZE);
	min_low_pfn = free_ram_start_pfn;

	/*
	 * initialize the boot-time allocator (with low memory only).
	 *
	 * This makes the memory from the end of the kernel to the end of
	 * RAM usable.
	 * init_bootmem sets the global values min_low_pfn, max_low_pfn.
	 */
	bootmap_size = init_bootmem_node(NODE_DATA(0), free_ram_start_pfn,
					 ram_start_pfn, max_low_pfn);
	free_bootmem(PFN_PHYS(free_ram_start_pfn),
		     (max_low_pfn - free_ram_start_pfn) << PAGE_SHIFT);
	reserve_bootmem(PFN_PHYS(free_ram_start_pfn), bootmap_size,
			BOOTMEM_DEFAULT);

	for_each_memblock(reserved, region) {
		if (region->size != 0) {
			printk(KERN_INFO "Reserved - 0x%08x-0x%08x\n",
			       (u32) region->base, (u32) region->size);
			reserve_bootmem(region->base, region->size,
					BOOTMEM_DEFAULT);
		}
	}
	return max_low_pfn;
}

void __init setup_arch(char **cmdline_p)
{
	early_init_devtree(__atags_pointer ? \
		phys_to_virt(__atags_pointer) : __dtb_start);

	setup_cpuinfo();

	init_mm.start_code = (unsigned long)&_stext;
	init_mm.end_code = (unsigned long)&_etext;
	init_mm.end_data = (unsigned long)&_edata;
	init_mm.brk = (unsigned long)&_end;

	/* use generic way to parse */
	parse_early_param();

	/* setup bootmem allocator */
	setup_memory();

	/* paging_init() sets up the MMU and marks all pages as reserved */
	paging_init();

	copy_fdt();
	unflatten_device_tree();

#ifdef CONFIG_SMP
	smp_init_cpus();
#endif

#if defined(CONFIG_VT)
#if defined(CONFIG_VGA_CONSOLE)
	conswitchp = &vga_con;
#elif defined(CONFIG_DUMMY_CONSOLE)
	conswitchp = &dummy_con;	//+ Tom: we will reach here
#endif
#endif

	*cmdline_p = cmd_line;
	early_trap_init();
}

static int __init nds32_device_probe(void)
{
	return of_platform_populate(NULL, NULL, NULL, NULL);
}

device_initcall(nds32_device_probe);

/*
 * cpu_init - initialise one CPU.
 *
 * cpu_init dumps the cache information, initialises SMP specific
 * information.
 */

void __init cpu_init(void)
{
	unsigned int cpu = smp_processor_id(), tmp = 0;

	if (cpu >= NR_CPUS) {
		printk(KERN_CRIT "CPU%u: bad primary CPU number\n", cpu);
		BUG();
	}

	if (system_state == SYSTEM_BOOTING)
		dump_cpu_info(cpu);

	tmp = 1 << IVB_offESZ;
#ifdef CONFIG_EVIC
	tmp |= 1 << IVB_offEVIC;
#endif
	SET_IVB(tmp | IVB_BASE);
	tmp = 0x10003;
	SET_INT_MASK(tmp);
	ISB();
}

static int c_show(struct seq_file *m, void *v)
{
	int i;

	seq_printf(m, "Processor\t: %s %s (id %lu, rev %lu, cfg %lu)\n",
		   proc_info.manufacturer, proc_info.cpu_name,
		   cpu_id, cpu_rev, cpu_cfgid);

	seq_printf(m, "L1I\t\t: %luKB/%luS/%luW/%luB\n",
		   CACHE_SET(ICACHE) * CACHE_WAY(ICACHE) *
		   CACHE_LINE_SIZE(ICACHE) / 1024, CACHE_SET(ICACHE),
		   CACHE_WAY(ICACHE), CACHE_LINE_SIZE(ICACHE));

	seq_printf(m, "L1D\t\t: %luKB/%luS/%luW/%luB\n",
		   CACHE_SET(DCACHE) * CACHE_WAY(DCACHE) *
		   CACHE_LINE_SIZE(DCACHE) / 1024, CACHE_SET(DCACHE),
		   CACHE_WAY(DCACHE), CACHE_LINE_SIZE(DCACHE));

#if defined(CONFIG_SMP)
	for_each_online_cpu(i) {
		seq_printf(m, "Processor\t: %d\n", i);
		seq_printf(m, "BogoMIPS\t: %lu.%02lu\n\n",
			   per_cpu(cpu_data,
				   i).loops_per_jiffy / (500000UL / HZ),
			   (per_cpu(cpu_data, i).loops_per_jiffy /
			    (5000UL / HZ)) % 100);
	}
#else /* CONFIG_SMP */
	seq_printf(m, "BogoMIPS\t: %lu.%02lu\n",
		   loops_per_jiffy / (500000 / HZ),
		   (loops_per_jiffy / (5000 / HZ)) % 100);
#endif

	/* dump out the processor features */
	seq_puts(m, "Features\t: ");

	for (i = 0; hwcap_str[i]; i++)
		if (elf_hwcap & (1 << i))
			seq_printf(m, "%s ", hwcap_str[i]);

	seq_puts(m, "\n\n");
	seq_printf(m, "Hardware\t: %s\n", elf_platform);

	return 0;
}

static void *c_start(struct seq_file *m, loff_t * pos)
{
	return *pos < 1 ? (void *)1 : NULL;
}

static void *c_next(struct seq_file *m, void *v, loff_t * pos)
{
	++*pos;
	return NULL;
}

static void c_stop(struct seq_file *m, void *v)
{
}

struct seq_operations cpuinfo_op = {
	.start = c_start,
	.next = c_next,
	.stop = c_stop,
	.show = c_show
};
