/*
 *  linux/arch/nds32/kernel/signal.c
 *
 *  Copyright (C) 1995-2002 Russell King
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/ptrace.h>
#include <linux/personality.h>
#include <linux/freezer.h>
#include <linux/tracehook.h>

#include <asm/cacheflush.h>
#include <asm/ucontext.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/fpu.h>

#include "ptrace.h"
#include "signal.h"
#include "fpu.h"

#define _BLOCKABLE (~(sigmask(SIGKILL) | sigmask(SIGSTOP)))

/*
 * For NDS32 syscalls, we encode the syscall number into the instruction.
 */
#if defined( __NDS32_EL__)
#define SWI_SYS_SIGRETURN       (0xeb0e0a64)
#define SWI_SYS_RT_SIGRETURN    (0xab150a64)
#define SWI_SYS_RESTART		(0x0b000a64)	/* syscall __NR_restart_syscall */
#define SWI_SYS_RESTART_LWIBI	(0x0180af0d)	/* lwi.bi  $p0, [$sp], 4 */
#define SWI_SYS_RESTART_JRP0	(0x0068004a)	/* jr      $p0 */
#elif defined(__NDS32_EB__)
#define SWI_SYS_SIGRETURN	(0x6400000b|(__NR_sigreturn<<5))
#define SWI_SYS_RT_SIGRETURN	(0x6400000b|(__NR_rt_sigreturn<<5))
#define SWI_SYS_RESTART		(0x640a000b)	/* syscall __NR_restart_syscall */
#define SWI_SYS_RESTART_LWIBI	(0x0daf8001)	/* lwi.bi  $p0, [$sp], 4 */
#define SWI_SYS_RESTART_JRP0	(0x4a006800)	/* jr      $p0 */
#else
#error "NDS32, but neither __NDS32_EB__, nor __NDS32_EL__???"
#endif

const unsigned long retcodes[2] = {
	SWI_SYS_SIGRETURN,
	SWI_SYS_RT_SIGRETURN
};

const unsigned long syscall_restart_code[3] = {
	SWI_SYS_RESTART,	/* syscall __NR_restart_syscall */
	SWI_SYS_RESTART_LWIBI,	/* lwi.bi  $p0, [$sp], 4 */
	SWI_SYS_RESTART_JRP0	/* jr      $p0 */
};

#ifdef CONFIG_FPU
static inline int restore_sigcontext_fpu(struct pt_regs *regs,
					 struct sigcontext __user * sc)
{
	struct task_struct *tsk = current;
	unsigned long used_math_flag;
	int ret = 0;

	if (!(GET_FUCOP_EXIST() & FUCOP_EXIST_mskCP0ISFPU))
		return 0;

	__get_user_error(used_math_flag, &sc->used_math_flag, ret);

	if (!used_math_flag)
		return 0;

	set_used_math();

#ifdef CONFIG_UNLAZY_FPU
	clear_fpu(regs);
#else
	preempt_disable();
	if (current == last_task_used_math) {
		last_task_used_math = NULL;
		release_fpu(regs);
	}
	preempt_enable();
#endif

	return __copy_from_user(&tsk->thread.fpu, &sc->fpu,
				sizeof(struct fpu_struct));
}

static inline int setup_sigcontext_fpu(struct pt_regs *regs,
				       struct sigcontext __user * sc)
{
	struct task_struct *tsk = current;
	int ret = 0;

	if (!(GET_FUCOP_EXIST() & FUCOP_EXIST_mskCP0ISFPU))
		return 0;

	__put_user_error(used_math(), &sc->used_math_flag, ret);

	if (!used_math())
		return ret;

	preempt_disable();
#ifdef CONFIG_UNLAZY_FPU
	unlazy_fpu(tsk);
#else
	if (last_task_used_math != NULL)
		save_fpu(last_task_used_math);
#endif
	ret = __copy_to_user(&sc->fpu, &tsk->thread.fpu,
			     sizeof(struct fpu_struct));

	grab_fpu(task_pt_regs(tsk));
	fpload(&init_fpuregs);
#ifndef CONFIG_UNLAZY_FPU	//Lazy FPU
	last_task_used_math = current;
#endif
	preempt_enable();
	return ret;
}
#else
static inline int
restore_sigcontext_fpu(struct pt_regs *regs, struct sigcontext __user * sc)
{
	return 0;
}

static inline int
setup_sigcontext_fpu(struct pt_regs *regs, struct sigcontext __user * sc)
{
	return 0;
}
#endif /* CONFIG_FPU */

/*
 * Auxiliary signal frame.  This saves stuff like FP state.
 * The layout of this structure is not part of the user ABI.
 */
struct aux_sigframe {
};

/*
 * Do a signal return; undo the signal stack.  These are aligned to 64-bit.
 */
struct sigframe {
	struct ucontext uc;
	unsigned long retcode;
	struct aux_sigframe aux __attribute__((aligned(8)));
};

struct rt_sigframe {
	struct siginfo info;
	struct sigframe sig;
};

static int restore_sigframe(struct pt_regs *regs, struct sigframe __user *sf)
{
	sigset_t set;
	int err;

	err = __copy_from_user(&set, &sf->uc.uc_sigmask, sizeof(set));
	if (err == 0) {
		sigdelsetmask(&set, ~_BLOCKABLE);
		set_current_blocked(&set);
	}

	__get_user_error(regs->NDS32_r0, &sf->uc.uc_mcontext.nds32_r0, err);
	__get_user_error(regs->NDS32_r1, &sf->uc.uc_mcontext.nds32_r1, err);
	__get_user_error(regs->NDS32_r2, &sf->uc.uc_mcontext.nds32_r2, err);
	__get_user_error(regs->NDS32_r3, &sf->uc.uc_mcontext.nds32_r3, err);
	__get_user_error(regs->NDS32_r4, &sf->uc.uc_mcontext.nds32_r4, err);
	__get_user_error(regs->NDS32_r5, &sf->uc.uc_mcontext.nds32_r5, err);
	__get_user_error(regs->NDS32_r6, &sf->uc.uc_mcontext.nds32_r6, err);
	__get_user_error(regs->NDS32_r7, &sf->uc.uc_mcontext.nds32_r7, err);
	__get_user_error(regs->NDS32_r8, &sf->uc.uc_mcontext.nds32_r8, err);
	__get_user_error(regs->NDS32_r9, &sf->uc.uc_mcontext.nds32_r9, err);
	__get_user_error(regs->NDS32_r10, &sf->uc.uc_mcontext.nds32_r10, err);
	__get_user_error(regs->NDS32_r11, &sf->uc.uc_mcontext.nds32_r11, err);
	__get_user_error(regs->NDS32_r12, &sf->uc.uc_mcontext.nds32_r12, err);
	__get_user_error(regs->NDS32_r13, &sf->uc.uc_mcontext.nds32_r13, err);
	__get_user_error(regs->NDS32_r14, &sf->uc.uc_mcontext.nds32_r14, err);
	__get_user_error(regs->NDS32_r15, &sf->uc.uc_mcontext.nds32_r15, err);
	__get_user_error(regs->NDS32_r16, &sf->uc.uc_mcontext.nds32_r16, err);
	__get_user_error(regs->NDS32_r17, &sf->uc.uc_mcontext.nds32_r17, err);
	__get_user_error(regs->NDS32_r18, &sf->uc.uc_mcontext.nds32_r18, err);
	__get_user_error(regs->NDS32_r19, &sf->uc.uc_mcontext.nds32_r19, err);
	__get_user_error(regs->NDS32_r20, &sf->uc.uc_mcontext.nds32_r20, err);

	__get_user_error(regs->NDS32_r21, &sf->uc.uc_mcontext.nds32_r21, err);
	__get_user_error(regs->NDS32_r22, &sf->uc.uc_mcontext.nds32_r22, err);
	__get_user_error(regs->NDS32_r23, &sf->uc.uc_mcontext.nds32_r23, err);
	__get_user_error(regs->NDS32_r24, &sf->uc.uc_mcontext.nds32_r24, err);
	__get_user_error(regs->NDS32_r25, &sf->uc.uc_mcontext.nds32_r25, err);
	__get_user_error(regs->NDS32_fp, &sf->uc.uc_mcontext.nds32_fp, err);
	__get_user_error(regs->NDS32_gp, &sf->uc.uc_mcontext.nds32_gp, err);
	__get_user_error(regs->NDS32_lp, &sf->uc.uc_mcontext.nds32_lp, err);
	__get_user_error(regs->NDS32_sp, &sf->uc.uc_mcontext.nds32_sp, err);
	__get_user_error(regs->NDS32_ipc, &sf->uc.uc_mcontext.nds32_ipc, err);
#if defined(CONFIG_HWZOL)
	__get_user_error(regs->NDS32_lc, &sf->uc.uc_mcontext.zol.nds32_lc, err);
	__get_user_error(regs->NDS32_le, &sf->uc.uc_mcontext.zol.nds32_le, err);
	__get_user_error(regs->NDS32_lb, &sf->uc.uc_mcontext.zol.nds32_lb, err);
#endif


	err |= !valid_user_regs(regs);
	err |= restore_sigcontext_fpu(regs, &sf->uc.uc_mcontext);

	return err;
}

asmlinkage int sys_sigreturn(struct pt_regs *regs)
{
	struct sigframe __user *frame;

	/* Always make any pending restarted system calls return -EINTR */
	current_thread_info()->restart_block.fn = do_no_restart_syscall;

	/*
	 * Since we stacked the signal on a 64-bit boundary,
	 * then 'sp' should be word aligned here.  If it's
	 * not, then the user is trying to mess with us.
	 */
	if (regs->NDS32_sp & 7)
		goto badframe;

	frame = (struct sigframe __user *)regs->NDS32_sp;

	if (!access_ok(VERIFY_READ, frame, sizeof(*frame)))
		goto badframe;

	if (restore_sigframe(regs, frame))
		goto badframe;

	return regs->NDS32_r0;

badframe:
	force_sig(SIGSEGV, current);
	return 0;
}

asmlinkage int sys_rt_sigreturn(struct pt_regs *regs)
{
	struct rt_sigframe __user *frame;

	/* Always make any pending restarted system calls return -EINTR */
	current_thread_info()->restart_block.fn = do_no_restart_syscall;

	/*
	 * Since we stacked the signal on a 64-bit boundary,
	 * then 'sp' should be word aligned here.  If it's
	 * not, then the user is trying to mess with us.
	 */
	if (regs->NDS32_sp & 7)
		goto badframe;

	frame = (struct rt_sigframe __user *)regs->NDS32_sp;

	if (!access_ok(VERIFY_READ, frame, sizeof(*frame)))
		goto badframe;

	if (restore_sigframe(regs, &frame->sig))
		goto badframe;

	if (restore_altstack(&frame->sig.uc.uc_stack))
		goto badframe;

	return regs->NDS32_r0;

badframe:
	force_sig(SIGSEGV, current);
	return 0;
}

static int
setup_sigframe(struct sigframe __user *sf, struct pt_regs *regs, sigset_t *set)
{
	int err = 0;

	err |= setup_sigcontext_fpu(regs, &sf->uc.uc_mcontext);

	__put_user_error(regs->NDS32_r0, &sf->uc.uc_mcontext.nds32_r0, err);
	__put_user_error(regs->NDS32_r1, &sf->uc.uc_mcontext.nds32_r1, err);
	__put_user_error(regs->NDS32_r2, &sf->uc.uc_mcontext.nds32_r2, err);
	__put_user_error(regs->NDS32_r3, &sf->uc.uc_mcontext.nds32_r3, err);
	__put_user_error(regs->NDS32_r4, &sf->uc.uc_mcontext.nds32_r4, err);
	__put_user_error(regs->NDS32_r5, &sf->uc.uc_mcontext.nds32_r5, err);
	__put_user_error(regs->NDS32_r6, &sf->uc.uc_mcontext.nds32_r6, err);
	__put_user_error(regs->NDS32_r7, &sf->uc.uc_mcontext.nds32_r7, err);
	__put_user_error(regs->NDS32_r8, &sf->uc.uc_mcontext.nds32_r8, err);
	__put_user_error(regs->NDS32_r9, &sf->uc.uc_mcontext.nds32_r9, err);
	__put_user_error(regs->NDS32_r10, &sf->uc.uc_mcontext.nds32_r10, err);
	__put_user_error(regs->NDS32_r11, &sf->uc.uc_mcontext.nds32_r11, err);
	__put_user_error(regs->NDS32_r12, &sf->uc.uc_mcontext.nds32_r12, err);
	__put_user_error(regs->NDS32_r13, &sf->uc.uc_mcontext.nds32_r13, err);
	__put_user_error(regs->NDS32_r14, &sf->uc.uc_mcontext.nds32_r14, err);
	__put_user_error(regs->NDS32_r15, &sf->uc.uc_mcontext.nds32_r15, err);
	__put_user_error(regs->NDS32_r16, &sf->uc.uc_mcontext.nds32_r16, err);
	__put_user_error(regs->NDS32_r17, &sf->uc.uc_mcontext.nds32_r17, err);
	__put_user_error(regs->NDS32_r18, &sf->uc.uc_mcontext.nds32_r18, err);
	__put_user_error(regs->NDS32_r19, &sf->uc.uc_mcontext.nds32_r19, err);
	__put_user_error(regs->NDS32_r20, &sf->uc.uc_mcontext.nds32_r20, err);

	__put_user_error(regs->NDS32_r21, &sf->uc.uc_mcontext.nds32_r21, err);
	__put_user_error(regs->NDS32_r22, &sf->uc.uc_mcontext.nds32_r22, err);
	__put_user_error(regs->NDS32_r23, &sf->uc.uc_mcontext.nds32_r23, err);
	__put_user_error(regs->NDS32_r24, &sf->uc.uc_mcontext.nds32_r24, err);
	__put_user_error(regs->NDS32_r25, &sf->uc.uc_mcontext.nds32_r25, err);
	__put_user_error(regs->NDS32_fp, &sf->uc.uc_mcontext.nds32_fp, err);
	__put_user_error(regs->NDS32_gp, &sf->uc.uc_mcontext.nds32_gp, err);
	__put_user_error(regs->NDS32_lp, &sf->uc.uc_mcontext.nds32_lp, err);
	__put_user_error(regs->NDS32_sp, &sf->uc.uc_mcontext.nds32_sp, err);
	__put_user_error(regs->NDS32_ipc, &sf->uc.uc_mcontext.nds32_ipc, err);
#if defined(CONFIG_HWZOL)
	__put_user_error(regs->NDS32_lc, &sf->uc.uc_mcontext.zol.nds32_lc, err);
	__put_user_error(regs->NDS32_le, &sf->uc.uc_mcontext.zol.nds32_le, err);
	__put_user_error(regs->NDS32_lb, &sf->uc.uc_mcontext.zol.nds32_lb, err);
#endif

	__put_user_error(current->thread.trap_no, &sf->uc.uc_mcontext.trap_no, err);
	__put_user_error(current->thread.error_code, &sf->uc.uc_mcontext.error_code, err);
	__put_user_error(current->thread.address, &sf->uc.uc_mcontext.fault_address, err);
	__put_user_error(set->sig[0], &sf->uc.uc_mcontext.oldmask, err);

	err |= __copy_to_user(&sf->uc.uc_sigmask, set, sizeof(*set));

	return err;
}

static inline void __user *
get_sigframe(struct k_sigaction *ka, struct pt_regs *regs, int framesize)
{
	unsigned long sp = regs->NDS32_sp;
	void __user *frame;

	/*
	 * This is the X/Open sanctioned signal stack switching.
	 */
	if ((ka->sa.sa_flags & SA_ONSTACK) && !sas_ss_flags(sp))
		sp = current->sas_ss_sp + current->sas_ss_size;

	/*
	 * ATPCS B01 mandates 8-byte alignment
	 */
	frame = (void __user *)((sp - framesize) & ~7);

	/*
	 * Check that we can actually write to the signal frame.
	 */
	if (!access_ok(VERIFY_WRITE, frame, framesize))
		frame = NULL;

	return frame;
}

static int setup_return(struct pt_regs *regs, struct k_sigaction *ka,
			unsigned long __user * rc, void __user * frame,
			int usig)
{
	unsigned long handler = (unsigned long)ka->sa.sa_handler;
	unsigned long retcode;

	/*
	 * Maybe we need to deliver a 32-bit signal to a 26-bit task.
	 */
	if (ka->sa.sa_flags & SA_RESTORER) {
		retcode = (unsigned long)ka->sa.sa_restorer;
	} else {

		unsigned int idx = 0;
		unsigned long line_size = CACHE_LINE_SIZE(ICACHE);
		unsigned long start, end;

		if (ka->sa.sa_flags & SA_SIGINFO)
			idx++;

		if (__put_user(retcodes[idx], rc))
			return 1;

		/*
		 *  Ensure that the instruction cache sees
		 *  the return code written onto the stack.
		 */
		start = (unsigned long)rc & ~(line_size - 1);
		end = start + line_size;
		flush_icache_range(start, end);

		retcode = KERN_SIGRETURN_CODE + (idx << 2);
	}

	regs->NDS32_r0 = usig;
	regs->NDS32_sp = (unsigned long)frame;
	regs->NDS32_lp = retcode;
	regs->NDS32_ipc = handler;
	/* Also store handler address in r15 for updating GP in the handler. */
	regs->NDS32_r15 = handler;

	return 0;
}

static int
setup_frame(int usig, struct k_sigaction *ka, sigset_t *set, struct pt_regs *regs)
{
	struct sigframe __user *frame = get_sigframe(ka, regs, sizeof(*frame));
	int err = 0;

	if (!frame)
		return 1;

	err |= setup_sigframe(frame, regs, set);
	if (err == 0)
		err = setup_return(regs, ka, &frame->retcode, frame, usig);

	return err;
}

static int
setup_rt_frame(int usig, struct k_sigaction *ka, siginfo_t *info,
	       sigset_t *set, struct pt_regs *regs)
{
	struct rt_sigframe __user *frame =
	    get_sigframe(ka, regs, sizeof(*frame));
	stack_t stack;
	int err = 0;

	if (!frame)
		return 1;

	err |= copy_siginfo_to_user(&frame->info, info);

	__put_user_error(0, &frame->sig.uc.uc_flags, err);
	__put_user_error(NULL, &frame->sig.uc.uc_link, err);

	memset(&stack, 0, sizeof(stack));
	stack.ss_sp = (void __user *)current->sas_ss_sp;
	stack.ss_flags = sas_ss_flags(regs->NDS32_sp);	//NDS32_sp
	stack.ss_size = current->sas_ss_size;
	err |= __copy_to_user(&frame->sig.uc.uc_stack, &stack, sizeof(stack));

	err |= setup_sigframe(&frame->sig, regs, set);
	if (err == 0)
		err = setup_return(regs, ka, &frame->sig.retcode, frame, usig);

	if (err == 0) {
		/*
		 * For realtime signals we must also set the second and third
		 * arguments for the signal handler.
		 *   -- Peter Maydell <pmaydell@chiark.greenend.org.uk> 2000-12-06
		 */
		regs->NDS32_r1 = (unsigned long)&frame->info;
		regs->NDS32_r2 = (unsigned long)&frame->sig.uc;
	}

	return err;
}

/*
 * OK, we're invoking a handler
 */
static int
handle_signal(unsigned long sig, struct k_sigaction *ka,
	      siginfo_t *info, sigset_t *oldset,
	      struct pt_regs * regs)
{
	struct thread_info *thread = current_thread_info();
	struct task_struct *tsk = current;
	int usig = sig;
	int ret;

	/*
	 * translate the signal
	 */
	if (usig < 32 && thread->exec_domain
	    && thread->exec_domain->signal_invmap)
		usig = thread->exec_domain->signal_invmap[usig];

	/*
	 * Set up the stack frame
	 */
	if (ka->sa.sa_flags & SA_SIGINFO)
		ret = setup_rt_frame(usig, ka, info, oldset, regs);
	else
		ret = setup_frame(usig, ka, oldset, regs);

	/*
	 * Check that the resulting registers are actually sane.
	 */
	ret |= !valid_user_regs(regs);

	if (ret != 0) {
		force_sigsegv(sig, tsk);
		return ret;
	}

	signal_delivered(sig, info, ka, regs, test_thread_flag(TIF_SINGLESTEP));
	return 0;
}

/*
 * Note that 'init' is a special process: it doesn't get signals it doesn't
 * want to handle. Thus you cannot kill init even with a SIGKILL even by
 * mistake.
 *
 * Note that we go through the signals twice: once to check the signals that
 * the kernel can handle, and then we build all the user-level signal handling
 * stack-frames in one go after that.
 */
static void do_signal(struct pt_regs *regs, int syscall)
{
	unsigned int retval = 0, continue_addr = 0, restart_addr = 0;
	struct k_sigaction ka;
	siginfo_t info;
	int signr;

	/*
	 * We want the common case to go fast, which
	 * is why we may in certain cases get here from
	 * kernel mode. Just return without doing anything
	 * if so.
	 */
	if (!user_mode(regs))
		return;

	/*
	 * If we were from a system call, check for system call restarting...
	 */
	if (syscall) {
		continue_addr = regs->NDS32_ipc;
		restart_addr = continue_addr - 4;
		retval = regs->NDS32_r0;

		/*
		 * Prepare for system call restart.  We do this here so that a
		 * debugger will see the already changed PSW.
		 */
		switch (retval) {
		case -ERESTARTNOHAND:
		case -ERESTARTSYS:
		case -ERESTARTNOINTR:
			regs->NDS32_r0 = regs->NDS32_ORIG_r0;
			regs->NDS32_ipc = restart_addr;
			break;
		case -ERESTART_RESTARTBLOCK:
			regs->NDS32_r0 = -EINTR;
			break;
		}
	}

	if (try_to_freeze())
		goto no_signal;

	/*
	 * Get the signal to deliver.  When running under ptrace, at this
	 * point the debugger may change all our registers ...
	 */
	signr = get_signal_to_deliver(&info, &ka, regs, NULL);
	if (signr > 0) {
#if 0 //def USE_SWBRK
		if (current->ptrace & PT_SINGLESTEP)
			ptrace_cancel_bpt(current);
#endif
		sigset_t *oldset;

		/*
		 * Depending on the signal settings we may need to revert the
		 * decision to restart the system call.  But skip this if a
		 * debugger has chosen to restart at a different PC.
		 */
		if (regs->NDS32_ipc == restart_addr) {
			if (retval == -ERESTARTNOHAND
			    || (retval == -ERESTARTSYS
				&& !(ka.sa.sa_flags & SA_RESTART))) {
				regs->NDS32_r0 = -EINTR;
				regs->NDS32_ipc = continue_addr;
			}
		}

		if (test_thread_flag(TIF_RESTORE_SIGMASK))
			oldset = &current->saved_sigmask;
		else
			oldset = &current->blocked;
		if (handle_signal(signr, &ka, &info, oldset, regs) == 0) {
			/*
			 * A signal was successfully delivered; the saved
			 * sigmask will have been stored in the signal frame,
			 * and will be restored by sigreturn, so we can simply
			 * clear the TIF_RESTORE_SIGMASK flag.
			 */
			if (test_thread_flag(TIF_RESTORE_SIGMASK))
				clear_thread_flag(TIF_RESTORE_SIGMASK);
		}
		return;
#if 0 //def USE_SWBRK
		if (current->ptrace & PT_SINGLESTEP)
			ptrace_set_bpt(current);
#endif
	}

no_signal:
	if (syscall) {
		/*
		 * Handle restarting a different system call.  As above,
		 * if a debugger has chosen to restart at a different PC,
		 * ignore the restart.
		 */
		if (retval == -ERESTART_RESTARTBLOCK
		    && regs->NDS32_ipc == continue_addr) {
			u32 __user *usp;

			regs->NDS32_sp -= 4;
			usp = (u32 __user *)regs->NDS32_sp;

			if (put_user(regs->NDS32_ipc, usp) == 0)
				regs->NDS32_ipc = KERN_RESTART_CODE;
			else {
				regs->NDS32_sp += 4;
				force_sigsegv(0, current);
			}
		}

		/* If there's no signal to deliver, we just put the saved sigmask
		 * back.
		 */
		if (test_thread_flag(TIF_RESTORE_SIGMASK)) {
			clear_thread_flag(TIF_RESTORE_SIGMASK);
			sigprocmask(SIG_SETMASK, &current->saved_sigmask, NULL);
		}
	}
}
/* This way might be better. */
#if 0
/* TODO: discuss pass thread_flag for _TIF_* checking or not */
asmlinkage void do_notify_resume(struct pt_regs *regs)
{
	if (current_thread_info()->flags & _TIF_SIGPENDING)
		do_signal(regs);

	if (current_thread_info()->flags & _TIF_NOTIFY_RESUME) {
		clear_thread_flag(TIF_NOTIFY_RESUME);
		tracehook_notify_resume(regs);
	}
}
#endif
asmlinkage void
do_notify_resume(struct pt_regs *regs, unsigned int thread_flags, int syscall)
{
	if (thread_flags & _TIF_SIGPENDING)
		do_signal(regs, syscall);

	if (thread_flags & _TIF_NOTIFY_RESUME) {
		clear_thread_flag(TIF_NOTIFY_RESUME);
		tracehook_notify_resume(regs);
	}
}
