/*
 *  linux/arch/nds32/kernel/smp.c
 *
 *  Copyright (C) 2002 ARM Limited, All Rights Reserved.
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/cache.h>
#include <linux/profile.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/cpu.h>
#include <linux/smp.h>
#include <linux/seq_file.h>
#include <asm/atomic.h>
#include <asm/cacheflush.h>
#include <asm/cpu.h>
#include <asm/processor.h>
#include <asm/tlbflush.h>
#include <asm/ptrace.h>
#include <linux/err.h>
#include <asm/irq_regs.h>
#include <asm/mmu_context.h>
#include <asm/mmu.h>

void smp_init_cpus(void)
{
	int i;
	for (i = 0; i < NR_CPUS; i++)
		cpu_set(i, cpu_possible_map);
}

int setup_profiling_timer(unsigned int multiplier)
{
	return 0;
}

struct tlb_data {
	struct vm_area_struct *vma;
	unsigned long start;
	unsigned long end;
};

static void ipi_flush_tlb_all(void *data)
{
	local_flush_tlb_all();
}

void flush_tlb_all(void)
{
	on_each_cpu(ipi_flush_tlb_all, NULL, 1);
}

static void ipi_flush_tlb_mm(void *data)
{
	struct mm_struct *mm = (struct mm_struct *)data;
	local_flush_tlb_mm(mm);
}

void flush_tlb_mm(struct mm_struct *mm)
{
	on_each_cpu(ipi_flush_tlb_mm, mm, 1);
}

static void ipi_flush_tlb_page(void *data)
{
	struct tlb_data *t = (struct tlb_data *)data;
	local_flush_tlb_page(t->vma, t->start);
}

void flush_tlb_page(struct vm_area_struct *vma, unsigned long addr)
{
	struct tlb_data data;

	data.vma = vma;
	data.start = addr;
	on_each_cpu(ipi_flush_tlb_page, &data, 1);
}

static void ipi_flush_tlb_range(void *data)
{
	struct tlb_data *t = (struct tlb_data *)data;
	local_flush_tlb_range(t->vma, t->start, t->end);
}

void flush_tlb_range(struct vm_area_struct *vma,
		     unsigned long start, unsigned long end)
{
	struct tlb_data data;

	data.vma = vma;
	data.start = start;
	data.end = end;
	on_each_cpu(ipi_flush_tlb_range, &data, 1);
}

static void ipi_flush_tlb_kernel_range(void *data)
{
	struct tlb_data *t = (struct tlb_data *)data;
	local_flush_tlb_kernel_range(t->start, t->end);
}

void flush_tlb_kernel_range(unsigned long start, unsigned long end)
{
	struct tlb_data data;

	data.start = start;
	data.end = end;
	on_each_cpu(ipi_flush_tlb_kernel_range, &data, 1);
}

/* IPI implementation */
static inline unsigned long read_ipi_trigger(void)
{
	/* AMIC IPI trigger register */
	return *(volatile unsigned long *)(AMIC_VA_BASE + 0x40);
}

static inline void write_ipi_trigger(const struct cpumask *mask)
{
	unsigned long data = *cpus_addr(*mask);
	*(volatile unsigned long *)(AMIC_VA_BASE + 0x40) = data;
}

static inline void clear_ipi_status(void)
{
	*(volatile unsigned long *)(AMIC_VA_BASE + 0x44) = 0xf;
	asm("msync store\nisb");
}

static void __init send_IPI_boot(int cpu, struct task_struct *tsk)
{
	extern void secondary_startup(void);
	unsigned long *ptr = (unsigned long *)(0xc0006000 + (cpu << 9));
	ptr[5] = virt_to_phys(secondary_startup);
	ptr[6] = (unsigned long)task_stack_page(tsk) + THREAD_SIZE - 8;
	asm("cctl %0, L1D_VA_WB, alevel\n"::"r"(ptr));
	asm("cctl %0, L1D_VA_INVAL, alevel\nmsync\ndsb\n"::"r"(ptr));
	write_ipi_trigger(get_cpu_mask(cpu));
}

static int __init wait_cpu_boot_done(int cpu)
{
	int i, state;
	for (i = 0; i < 10000; i++) {
		state = read_ipi_trigger();
		if ((state & (1 << cpu)) == 0)
			break;
		udelay(100);
	}
	return (i == 10000) ? -1 : 0;
}

void __init secondary_start_kernel(void)
{
	unsigned long tmp;
	unsigned int cpu = smp_processor_id();

	atomic_inc(&init_mm.mm_count);
	current->active_mm = &init_mm;

	/*
	 * enable cache.
	 */

	tmp = GET_CACHE_CTL();
#ifndef CONFIG_CPU_DCACHE_DISABLE
	tmp |= CACHE_CTL_mskDC_EN;
#endif

#ifndef CONFIG_CPU_ICACHE_DISABLE
	tmp |= CACHE_CTL_mskIC_EN;
#endif
	SET_CACHE_CTL(tmp);
	DSB();
	ISB();

	preempt_disable();
	local_irq_enable();

	//assume all of cores runs the same speed
	//calibrate_delay();
	/* store cpu info ot the second cpu */
	smp_store_cpu_info(cpu);

	clear_ipi_status();
	cpu_idle();
}

/*
 * Called by both boot and secondaries to move global data into
 * per-processor storage.
 */
void __init smp_store_cpu_info(unsigned int cpuid)
{
	struct cpuinfo_nds32 *cpu_info = &per_cpu(cpu_data, cpuid);

	cpu_info->loops_per_jiffy = loops_per_jiffy;
}

/* functions be used by generic layer */
void arch_send_call_function_single_ipi(int cpu)
{
	write_ipi_trigger(get_cpu_mask(cpu));
}

void arch_send_call_function_ipi_mask(const struct cpumask *mask)
{
	write_ipi_trigger(mask);
}

static void __init smp_boot_one_cpu(int cpu)
{
	struct task_struct *idle;
	int ret;

	idle = fork_idle(cpu);
	if (IS_ERR(idle))
		panic(KERN_ERR "Fork failed for CPU %d", cpu);

	send_IPI_boot(cpu, idle);
	ret = wait_cpu_boot_done(cpu);
	if (ret == 0)
		cpu_set(cpu, cpu_online_map);
	else
		put_task_struct(idle);
}

static void ipi_timer(void *data)
{
	profile_tick(CPU_PROFILING);
	update_process_times(user_mode(get_irq_regs()));
}

void smp_send_timer(void)
{
	smp_call_function(ipi_timer, NULL, 0);
}

static void ipi_stop(void *data)
{
	cpu_clear(smp_processor_id(), cpu_online_map);
	local_irq_enable();
	for (;;) ;
}

void smp_send_stop(void)
{
	smp_call_function(ipi_stop, NULL, 0);
}

void smp_send_reschedule(int cpu)
{
	write_ipi_trigger(get_cpu_mask(cpu));
}

static void ipi_handler(unsigned int irq, struct irq_desc *desc)
{
	clear_ipi_status();
	generic_smp_call_function_single_interrupt();
	generic_smp_call_function_interrupt();
}

void __init smp_prepare_cpus(unsigned int max_cpus)
{
	/*
	 * XXX detect how many cores exist
	 */
	int i;
	unsigned int cpu = smp_processor_id();

	/* store master core cpu info */
	smp_store_cpu_info(cpu);
	for (i = 0; i < max_cpus; i++)
		cpu_set(i, cpu_present_map);

	/* setup IPI handler */
	set_irq_chip(32, &dummy_irq_chip);
	set_irq_chained_handler(32, ipi_handler);
}

int __cpuinit __cpu_up(unsigned int cpu)
{
	smp_boot_one_cpu(cpu);
	return cpu_online(cpu) ? 0 : -ENOSYS;
}

void __init smp_cpus_done(unsigned int max_cpus)
{
	int cpu;
	unsigned long bogosum = 0;

	for_each_online_cpu(cpu)
	    bogosum += per_cpu(cpu_data, cpu).loops_per_jiffy;

	printk(KERN_INFO "SMP: Total of %d processors activated "
	       "(%lu.%02lu BogoMIPS).\n",
	       num_online_cpus(),
	       bogosum / (500000 / HZ), (bogosum / (5000 / HZ)) % 100);

	/* now we can set all interruption don't cared */
	*(volatile unsigned long *)(AMIC_VA_BASE + 0x4) = 0xffffffff;
}

void __init smp_prepare_boot_cpu(void)
{
	unsigned int cpu = smp_processor_id();
	unsigned long *ptr = (unsigned long *)(0xc0006000 + (cpu << 9));
	ptr[4] = 1;
}
