/*
 *  linux/arch/nds32/kernel/traps.c
 *
 *  Copyright (C) 1995-2002 Russell King
 *  Fragments that appear the same as linux/arch/i386/kernel/traps.c (C) Linus Torvalds
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  'traps.c' handles hardware exceptions after we have saved some state in
 *  'linux/arch/nds32/lib/traps.S'.  Mostly a debugging aid, but will probably
 *  kill the offending process.
 */
/* ============================================================================
 *
 *  linux/arch/nds32/kernel/traps.c
 *
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is trap handling for NDS32 core, initial refer from ARM.
 *
 *  Revision History:
 *
 *    Jul.16.2007     Initial ported by Tom, revised for KGDB by Harry.
 *
 *  Note:
 *
 * ============================================================================
 */
#include <linux/module.h>
#include <linux/personality.h>
#include <linux/kallsyms.h>
#include <linux/hardirq.h>
#include <linux/kdebug.h>

#include <asm/cacheflush.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/traps.h>
#include <asm/fpu.h>
#include <asm/reg_access.h>
#include <asm/hw_breakpoint.h>

#include <linux/ptrace.h>
#include <nds32_intrinsic.h>
#include "ptrace.h"
#include "signal.h"

#define ETYPE_INFO_LENGTH 60
#define ENTRY_POINT_INFOR_LENGTH 20

extern void show_pte(struct mm_struct *mm, unsigned long addr);
static void do_alignment_check(unsigned long entry,
			       unsigned long addr, unsigned long itype,
			       struct pt_regs *regs, unsigned long oipc);
static void do_revinsn(unsigned long entry,
		       unsigned long addr, unsigned long itype,
		       struct pt_regs *regs, unsigned long oipc);
static void do_traps(unsigned long entry,
		     unsigned long addr, unsigned long itype,
		     struct pt_regs *regs, unsigned long oipc);
void unhandled_exceptions(unsigned long entry,
			  unsigned long addr, unsigned long itype,
			  struct pt_regs *regs, unsigned long oipc);
static void do_coprocesser_exception(unsigned long entry,
				     unsigned long addr, unsigned long itype,
				     struct pt_regs *regs, unsigned long oipc);
static void do_arithmetic(unsigned long entry,
			  unsigned long addr, unsigned long itype,
			  struct pt_regs *regs, unsigned long oipc);
void coprocsr_default_handler(unsigned long itype, struct pt_regs *regs,
			      unsigned long eva);

static bool fpu_coprocess_exist;
struct coprocsr_trap_t coprocsr_trap[MAX_COPROCESSOR_NUMBER];
char exception_entry_point_info[ENTRY_POINT_NUM][ENTRY_POINT_INFOR_LENGTH] = {
	"Reset/NMI",
	"TLB fill",
	"PTE not present",
	"TLB misc",
	"TLB VLPT miss",
	"Machine Error",
	"Debug related",
	"General exception",
	"syscall"
};
static struct general_exception_cb_t {
	char etype_info[ETYPE_INFO_LENGTH];
	void (*handler) (unsigned long entry, unsigned long addr,
			 unsigned long itype, struct pt_regs *regs,
			 unsigned long oipc);
} general_exception_cb[ITYPE_ETYPE_NUM] = {
	{ "Alignment check", do_alignment_check },
	{ "Reserved instruction", do_revinsn },
	{ "Trap", do_traps },
	{ "Arithmetic", do_arithmetic },
	{ "Precise bus error", unhandled_exceptions },
	{ "Imprecise bus error", unhandled_exceptions },
	{ "Coprocessor", do_coprocesser_exception },
	{ "Privileged instruction", unhandled_exceptions },
	{ "Reserved value", unhandled_exceptions },
	{ "Nonexistent memory address", unhandled_exceptions },
	{ "MPZIU Control", unhandled_exceptions },
	{ "Next-precise stack overflow", unhandled_exceptions },
	{ "Reserved", unhandled_exceptions },
	{ "Reserved", unhandled_exceptions },
	{ "Reserved", unhandled_exceptions },
	{ "Reserved", unhandled_exceptions }
};

/*
 * Dump out the contents of some memory nicely...
 */
void dump_mem(const char *str, unsigned long bottom, unsigned long top)
{
	unsigned long p = bottom & ~31;
	mm_segment_t fs;
	int i;

	/*
	 * We need to switch to kernel mode so that we can use __get_user
	 * to safely read from kernel space.  Note that we now dump the
	 * code first, just in case the backtrace kills us.
	 */
	fs = get_fs();
	set_fs(KERNEL_DS);

	printk("%s(0x%08lx to 0x%08lx)\n", str, bottom, top);

	for (p = bottom & ~31; p < top;) {
		printk("%04lx: ", p & 0xffff);

		for (i = 0; i < 8; i++, p += 4) {
			unsigned int val;

			if (p < bottom || p >= top)
				printk("         ");
			else {
				__get_user(val, (unsigned long *)p);
				printk("%08x ", val);
			}
		}
		printk("\n");
	}

	set_fs(fs);
}

EXPORT_SYMBOL(dump_mem);

/* These intrinsic functions are not supported in V2 toolchains of BSP321. */
#ifndef __NDS32_BASELINE_V2__
#define DEBUG_TLB_CACHE
#endif
#ifdef DEBUG_TLB_CACHE
/* This is the number of TLB entries. User should change it if necessary. */
#define TLB_NUM 128
unsigned int tlb_misc_new[TLB_NUM], tlb_vpn_new[TLB_NUM], tlb_data_new[TLB_NUM];
/* To get the whole TLB contents once it uses va=0x0 */
void dump_tlb(unsigned long va)
{
	unsigned int rd_num, tlb_vpn, tlb_misc, tlb_data, mmu_cfg, tlb_num;

	/* save tlb system registers */
	tlb_vpn = __nds32__mfsr(NDS32_SR_TLB_VPN);
	tlb_misc = __nds32__mfsr(NDS32_SR_TLB_MISC);
	tlb_data = __nds32__mfsr(NDS32_SR_TLB_DATA);
	mmu_cfg = __nds32__mfsr(NDS32_SR_MMU_CFG);
	tlb_num =
	    (((mmu_cfg & MMU_CFG_mskTBW) >> MMU_CFG_offTBW) +
	     1) * (1 << (((mmu_cfg & MMU_CFG_mskTBS) >> MMU_CFG_offTBS) + 2));

	for (rd_num = 0; rd_num < tlb_num; rd_num++) {
		/* read tlb entry with index */
		__nds32__tlbop_trd(rd_num);
		__nds32__dsb();
		tlb_vpn_new[rd_num] = __nds32__mfsr(NDS32_SR_TLB_VPN);
		tlb_misc_new[rd_num] = __nds32__mfsr(NDS32_SR_TLB_MISC);
		tlb_data_new[rd_num] = __nds32__mfsr(NDS32_SR_TLB_DATA);
	}

	/* restore tlb system registers */
	__nds32__mtsr(tlb_vpn, NDS32_SR_TLB_VPN);
	__nds32__mtsr(tlb_misc, NDS32_SR_TLB_MISC);
	__nds32__dsb();

	printk("cur VPN:%08x, MISC:%08x, DATA:%08x\n", tlb_vpn, tlb_misc,
	       tlb_data);
	/* to read out all the data */
	for (rd_num = 0; rd_num < tlb_num; rd_num++) {
		/*unsigned int vpn = tlb_vpn_new[rd_num] & PAGE_MASK; */
		if (tlb_data_new[rd_num] & 0x1)
			if (va == 0x0
			    || (va != 0x0
				&& (tlb_vpn_new[rd_num] == (va & PAGE_MASK))))
				printk
				    ("idx:0x%08x, VPN:%08x, MISC:%08x, DATA:%08x\n",
				     rd_num, tlb_vpn_new[rd_num],
				     tlb_misc_new[rd_num],
				     tlb_data_new[rd_num]);
	}
}

EXPORT_SYMBOL(dump_tlb);

struct cache_element {
	unsigned int pa;
	unsigned int wd;
	unsigned int cacheline[8];
	unsigned char dirty;
	unsigned char valid;
	unsigned char lock;
};

/* This is the number of cache entries. User should change it if necessary. */
#define CACHE_SET_NUM 0x100
#define CACHE_WAY_NUM 0x4
volatile struct cache_element ce[CACHE_SET_NUM][CACHE_WAY_NUM];
#define CCTL_mskDIRTY 0x400000
#define CCTL_offDIRTY 22
#define CCTL_mskVALID 0x2
#define CCTL_offVALID 1
#define CCTL_mskLOCK 0x1
#define CCTL_offLOCK 0
#define CCTL_mskTAG 0x3ffffc
#define CCTL_offTAG 0x2
extern unsigned long va2idx(unsigned long va, unsigned int cache_type,
			    unsigned long *way_offset);
#include <asm/cache_info.h>
extern struct cache_info L1_cache_info[2];
void dump_cache(unsigned int cache_type)
{
	volatile unsigned long idx, way, ra, tag, i;
	unsigned long sets, ways, line_size, set_bits, way_bits, line_bits,
	    way_offset;

	ways = L1_cache_info[DCACHE].ways;
	sets = L1_cache_info[DCACHE].sets;
	set_bits = L1_cache_info[cache_type].set_bits;
	way_bits = L1_cache_info[cache_type].way_bits;
	line_bits = L1_cache_info[cache_type].line_bits;
	line_size = L1_cache_info[cache_type].line_size;
	way_offset = set_bits + line_bits;

	if (cache_type != ICACHE && cache_type != DCACHE) {
		printk("%s not supported cache_type:%x\n", __func__,
		       cache_type);
		return;
	}

	/* NDS32_CCTL_L1I_IX_RTAG Read tag L1I cache */
	/* NDS32_CCTL_L1I_IX_RWD Read word data L1I cache */

	for (idx = 0; idx < sets; idx++) {
		for (way = 0; way < ways; way++) {
			ra = (way << way_offset) | (idx << line_bits);
			if (cache_type == ICACHE)
				tag =
				    __nds32__cctlidx_read
				    (NDS32_CCTL_L1I_IX_RTAG, ra);
			else
				tag =
				    __nds32__cctlidx_read
				    (NDS32_CCTL_L1D_IX_RTAG, ra);
			ce[idx][way].dirty =
			    (unsigned char)(tag & CCTL_mskDIRTY) >>
			    CCTL_offDIRTY;
			ce[idx][way].valid =
			    (unsigned char)(tag & CCTL_mskVALID) >>
			    CCTL_offVALID;
			ce[idx][way].lock =
			    (unsigned char)(tag & CCTL_mskLOCK) >> CCTL_offLOCK;
			ce[idx][way].pa =
			    (tag & CCTL_mskTAG) >> CCTL_offTAG << PAGE_SHIFT;
			for (i = 0; i < line_size / 4; i++) {
				if (cache_type == ICACHE)
					ce[idx][way].cacheline[i] =
					    __nds32__cctlidx_read
					    (NDS32_CCTL_L1I_IX_RWD,
					     (ra | i << 2));
				else
					ce[idx][way].cacheline[i] =
					    __nds32__cctlidx_read
					    (NDS32_CCTL_L1D_IX_RWD,
					     (ra | i << 2));
			}
		}
	}
	printk("dump %s\n", cache_type ? "DCACHE" : "ICACHE");
	printk("%8s %4s %4s %1s %1s %1s %8s %8s %8s %8s %8s %8s %8s %8s\n",
	       "ADDRESS", "SET", "WAY", "V", "D", "L", "00", "04", "08", "0C",
	       "10", "14", "18", "1C");
	for (idx = 0; idx < sets; idx++) {
		for (way = 0; way < ways; way++) {
			printk("%08lx %04lx %04lx %1u %1u %1u ",
			       ce[idx][way].pa +
			       ((idx * line_size) % PAGE_SIZE), idx, way,
			       ce[idx][way].valid, ce[idx][way].dirty,
			       ce[idx][way].lock);
			for (i = 0; i < line_size / 4; i++) {
				printk("%08x ", ce[idx][way].cacheline[i]);
			}
			printk("\n");
		}
	}
}

EXPORT_SYMBOL(dump_cache);

void dump_cache_va(unsigned int cache_type, unsigned int va)
{
	volatile struct cache_element cache_entry[4];
	volatile unsigned long idx, way, tag, ra, i;
	unsigned long ways, line_size, set_bits, line_bits, way_offset;

	ways = L1_cache_info[cache_type].ways;
	line_size = CACHE_LINE_SIZE(cache_type);
	set_bits = L1_cache_info[cache_type].set_bits;
	line_bits = L1_cache_info[cache_type].line_bits;

	if (cache_type != ICACHE && cache_type != DCACHE) {
		printk("%s not supported cache_type:%x\n", __func__,
		       cache_type);
		return;
	}
	idx = va2idx(va, cache_type, &way_offset);
	//idx = (va & (((1 << set_bits) - 1) << line_bits)) >> line_bits;
	for (way = 0; way < ways; way++) {
		ra = (way << way_offset) | idx;
		if (cache_type == ICACHE)
			tag = __nds32__cctlidx_read(NDS32_CCTL_L1I_IX_RTAG, ra);
		else
			tag = __nds32__cctlidx_read(NDS32_CCTL_L1D_IX_RTAG, ra);
		cache_entry[way].dirty =
		    (unsigned char)(tag & CCTL_mskDIRTY) >> CCTL_offDIRTY;
		cache_entry[way].valid =
		    (unsigned char)(tag & CCTL_mskVALID) >> CCTL_offVALID;
		cache_entry[way].lock =
		    (unsigned char)(tag & CCTL_mskLOCK) >> CCTL_offLOCK;
		cache_entry[way].pa =
		    (tag & CCTL_mskTAG) >> CCTL_offTAG << PAGE_SHIFT;
		for (i = 0; i < line_size / 4; i++) {
			if (cache_type == ICACHE)
				cache_entry[way].cacheline[i] =
				    __nds32__cctlidx_read(NDS32_CCTL_L1I_IX_RWD,
							  (ra | i << 2));
			else
				cache_entry[way].cacheline[i] =
				    __nds32__cctlidx_read(NDS32_CCTL_L1D_IX_RWD,
							  (ra | i << 2));
		}
	}
	printk("dump %s va:%x\n", cache_type ? "DCACHE" : "ICACHE", va);

	printk("%8s %4s %4s %1s %1s %1s %8s %8s %8s %8s %8s %8s %8s %8s\n",
	       "ADDRESS", "SET", "WAY", "V", "D", "L", "00", "04", "08", "0C",
	       "10", "14", "18", "1C");
	for (way = 0; way < ways; way++) {
		printk("%08lx %04lx %04lx %1u %1u %1u ",
		       cache_entry[way].pa +
		       (((idx >> line_bits) * line_size) % PAGE_SIZE),
		       (idx >> line_bits), way, cache_entry[way].valid,
		       cache_entry[way].dirty, cache_entry[way].lock);
		for (i = 0; i < 8; i++) {
			printk("%08x ", cache_entry[way].cacheline[i]);
		}
		printk("\n");
	}
}

EXPORT_SYMBOL(dump_cache_va);
#endif

static void dump_instr(struct pt_regs *regs)
{
	unsigned long addr = instruction_pointer(regs);
	const int width = 8;
	mm_segment_t fs;
	int i;

	return;
	/*
	 * We need to switch to kernel mode so that we can use __get_user
	 * to safely read from kernel space.  Note that we now dump the
	 * code first, just in case the backtrace kills us.
	 */
	fs = get_fs();
	set_fs(KERNEL_DS);

	printk("Code: ");
	for (i = -4; i < 1; i++) {
		unsigned int val, bad;

		bad = __get_user(val, &((u32 *) addr)[i]);

		if (!bad)
			printk(i == 0 ? "(%0*x) " : "%0*x ", width, val);
		else {
			printk("bad PC value.");
			break;
		}
	}
	printk("\n");

	set_fs(fs);
}

#ifdef CONFIG_FUNCTION_GRAPH_TRACER
#include <linux/ftrace.h>
static void
get_real_ret_addr(unsigned long *addr, struct task_struct *tsk, int *graph)
{
        if (*addr == (unsigned long)return_to_handler) {
                int index = tsk->curr_ret_stack;

                if (tsk->ret_stack && index >= *graph) {
                        index -= *graph;
                        *addr = tsk->ret_stack[index].ret;
                        (*graph)++;
                }
        }
}
#else
static inline void
get_real_ret_addr(unsigned long *addr, struct task_struct *tsk, int *graph)
{
}
#endif

#define LOOP_TIMES (100)
static void __dump(struct task_struct *tsk, unsigned long *base_reg)
{
        unsigned long ret_addr;
        int cnt = LOOP_TIMES, graph =0;
        printk("Call Trace:\n");
#ifndef CONFIG_FRAME_POINTER
        while (!kstack_end(base_reg)) {
                ret_addr = *base_reg++;
                if (__kernel_text_address(ret_addr)) {
                        get_real_ret_addr(&ret_addr, tsk, &graph);
                        printk("[<%08lx>] ? ", ret_addr);
                        print_symbol("%s\n", ret_addr);
                }
                if (--cnt < 0)
                        break;
        }
#else
        while (!kstack_end((void *)base_reg) &&
               !((unsigned long)base_reg & 0x3) &&
               ((unsigned long)base_reg >= TASK_SIZE)) {
                unsigned long next_fp;
#if !defined(NDS32_ABI_2)
                ret_addr = base_reg[0];
                next_fp = base_reg[1];
#else
                ret_addr = base_reg[-1];
                next_fp = base_reg[FP_OFFSET];
#endif
                if (__kernel_text_address(ret_addr)) {
                        get_real_ret_addr(&ret_addr, tsk, &graph);
                        printk("[<%08lx>] ", ret_addr);
                        print_symbol("%s\n", ret_addr);
                }

                if (--cnt < 0)
                        break;

                base_reg = (unsigned long *)next_fp;
        }
#endif
        printk("\n");
}

void dump_stack(void)
{
        unsigned long *base_reg;
#ifndef CONFIG_FRAME_POINTER
        __asm__ __volatile__("\tori\t%0, $sp, #0\n":"=r"(base_reg));
#else
        __asm__ __volatile__("\tori\t%0, $fp, #0\n":"=r"(base_reg));
#endif
        __dump(NULL, base_reg);
}
EXPORT_SYMBOL(dump_stack);

void show_stack(struct task_struct *tsk, unsigned long *sp)
{
        unsigned long *base_reg;

        if (!tsk)
                tsk = current;
#ifndef CONFIG_FRAME_POINTER
        if (tsk != current)
                base_reg = (unsigned long *)(task_thread_info(tsk)->sp_save);
        else
                __asm__ __volatile__("\tori\t%0, $sp, #0\n":"=r"(base_reg));
#else
        if (tsk != current)
                base_reg =
                    (unsigned long *)(task_thread_info(tsk)->sp_save->fp);
        else
                __asm__ __volatile__("\tori\t%0, $fp, #0\n":"=r"(base_reg));
#endif
        __dump(tsk, base_reg);
        barrier();
}

DEFINE_SPINLOCK(die_lock);

/*
 * This function is protected against re-entrancy.
 */
void die(const char *str, struct pt_regs *regs, int err)
{
	struct task_struct *tsk = current;
	static int die_counter;

	console_verbose();
	spin_lock_irq(&die_lock);
	bust_spinlocks(1);

	printk("Internal error: %s: %x [#%d]\n", str, err, ++die_counter);
	print_modules();
	printk("CPU: %i\n", smp_processor_id());
	show_regs(regs);
	printk("Process %s (pid: %d, stack limit = 0x%p)\n",
	       tsk->comm, tsk->pid, task_thread_info(tsk) + 1);

	if (!user_mode(regs) || in_interrupt()) {
		dump_mem("Stack: ", regs->NDS32_sp,
			 8192 + (unsigned long)task_thread_info(tsk));
		dump_instr(regs);
		dump_stack();
	}

	bust_spinlocks(0);
	spin_unlock_irq(&die_lock);
	do_exit(SIGSEGV);
}
EXPORT_SYMBOL(die);

void die_if_kernel(const char *str, struct pt_regs *regs, int err)
{
	if (user_mode(regs))
		return;

	die(str, regs, err);
}

int bad_syscall(int n, struct pt_regs *regs)
{
	struct thread_info *thread = current_thread_info();
	siginfo_t info;

	if (current->personality != PER_LINUX && thread->exec_domain->handler) {
		thread->exec_domain->handler(n, regs);
		return regs->NDS32_r0;
	}

	info.si_signo = SIGILL;
	info.si_errno = 0;
	info.si_code = ILL_ILLTRP;
	info.si_addr = (void __user *)instruction_pointer(regs) - 4;

	force_sig_info(SIGILL, &info, current);
	die_if_kernel("Oops - bad syscall", regs, n);
	return regs->NDS32_r0;
}

void __pte_error(const char *file, int line, unsigned long val)
{
	printk("%s:%d: bad pte %08lx.\n", file, line, val);
}

void __pmd_error(const char *file, int line, unsigned long val)
{
	printk("%s:%d: bad pmd %08lx.\n", file, line, val);
}

void __pgd_error(const char *file, int line, unsigned long val)
{
	printk("%s:%d: bad pgd %08lx.\n", file, line, val);
}

extern char exception_vector[73][64];
void __init trap_init(void)
{
	int id;

	if ((GET_CPU_VER() & CPU_VER_mskCFGID_COP) &&
	    (GET_FUCOP_EXIST() &
	     (FUCOP_EXIST_mskCP0ISFPU | FUCOP_EXIST_mskCP0EX)))
		fpu_coprocess_exist = true;

	/* Initialize exception handler of coprocessor */
	for(id = 0; id<MAX_COPROCESSOR_NUMBER; id++ ) {
		coprocsr_trap[id].disable_handler = coprocsr_default_handler;
		coprocsr_trap[id].excpt_handler = coprocsr_default_handler;
		coprocsr_trap[id].impre_excpt_handler
						= coprocsr_default_handler;
	}
	return;
}

void __init early_trap_init(void)
{
	unsigned long ivb = 0;
	unsigned long base = 0xc0000000;

	memcpy((unsigned long *)base, (unsigned long *)exception_vector,
	       sizeof(exception_vector));
	ivb = __nds32__mfsr(NDS32_SR_IVB);
#ifdef CONFIG_EVIC
	__nds32__mtsr((ivb & ~IVB_mskESZ) | (2 << IVB_offESZ) |
		      (1 << IVB_offEVIC) | IVB_BASE, NDS32_SR_IVB);
#else
	/* Check platform support. */
# if defined (CONFIG_IVIC_INTC)
	if (((ivb & IVB_mskNIVIC) >> IVB_offNIVIC) >= 2)
		panic
		    ("IVIC_INTC mode is not allowed on the platform without interrupt controller\n");
# elif defined(CONFIG_IVIC)
	if (((ivb & IVB_mskNIVIC) >> IVB_offNIVIC) < 2)
		panic
		    ("IVIC mode is not allowed on the platform with interrupt controller\n");
# endif
	__nds32__mtsr((ivb & ~IVB_mskESZ) | (2 << IVB_offESZ) | IVB_BASE,
		      NDS32_SR_IVB);
#endif
	__nds32__mtsr(0x10003 | INT_MASK_mskIDIVZE, NDS32_SR_INT_MASK);
	/*
	 * Copy signal return handlers into the vector page, and
	 * set sigreturn to be a pointer to these.
	 */
	memcpy((void *)KERN_SIGRETURN_CODE, retcodes, sizeof(retcodes));
	memcpy((void *)KERN_RESTART_CODE, syscall_restart_code,
	       sizeof(syscall_restart_code));

	/*
	 * 0x2000 is 8K-aligned of 0x1240 = 73 vectors * 64byte
	 * 0x1000 is page saving sigreturn & restart code
	 */
	flush_icache_range(base, base + 0x3000);

}

NMI_handler_t NMI_handler = 0;
void register_NMI_handler(NMI_handler_t handler)
{
	if(handler)
		NMI_handler = handler;
}
EXPORT_SYMBOL(register_NMI_handler);

void unregister_NMI_handler(void)
{
	NMI_handler = 0;
}
EXPORT_SYMBOL(unregister_NMI_handler);

void do_NMI(void)
{
	if(NMI_handler)
		NMI_handler();
}

/*
 * I modified original debug_trap to apply KGDB stuff,
 * Harry@Jul.18.2007
 */
void do_debug_trap(unsigned long entry, unsigned long addr,
		   unsigned long type, struct pt_regs *regs)
{
	if (notify_die(DIE_DEBUG, "debug", regs, addr, type, SIGTRAP)
	    == NOTIFY_STOP)
		return;

	if (user_mode(regs)) {
		/* trap_signal */
		send_sigtrap(current, regs, 0, TRAP_BRKPT);
	} else {
		/* kernel_trap */
		if (!fixup_exception(regs))
			die("unexpected kernel_trap", regs, 0);
	}
}

void unhandled_interruption(struct pt_regs *regs)
{
	siginfo_t si;

	printk("Unhandled Interruption\n");
	show_regs(regs);
	if (!user_mode(regs))
		do_exit(SIGKILL);
	si.si_signo = SIGKILL;
	si.si_errno = 0;
	force_sig_info(SIGKILL, &si, current);
}

void unhandled_exceptions(unsigned long entry, unsigned long addr,
			  unsigned long itype, struct pt_regs *regs,
			  unsigned long oipc)
{
	siginfo_t si;
	unsigned long etype = itype & ITYPE_mskETYPE;

	printk("Unhandled Exception\n");
        printk("entry: %lx (%s)\n", entry, exception_entry_point_info[entry]);
        printk("EVA addr:  %lx\n", addr);
        printk("etype: %lx", etype);
	if (GENERAL_EXCEPTION == entry) {
		printk(" (%s)\n", general_exception_cb[etype].etype_info);
	} else {
		printk("\n");
	}
	show_regs(regs);
	if (!user_mode(regs))
		do_exit(SIGKILL);
	si.si_signo = SIGKILL;
	si.si_errno = 0;
	si.si_addr = (void *)addr;
	force_sig_info(SIGKILL, &si, current);
}

extern int do_page_fault(unsigned long entry, unsigned long addr,
			 unsigned int error_code, struct pt_regs *regs);

/*
 * 2:DEF dispatch for TLB MISC exception handler
*/

void do_dispatch_tlb_misc(unsigned long entry, unsigned long addr,
			  unsigned long type, struct pt_regs *regs)
{
	type = type & (ITYPE_mskINST | ITYPE_mskETYPE);
	if ((type & 0xf) < 5)
		do_page_fault(entry, addr, type, regs);
	else
		unhandled_exceptions(entry, addr, type, regs, GET_OIPC());
}

int (*do_unaligned_access) (unsigned long entry, unsigned long addr,
			    unsigned long type, struct pt_regs * regs) = NULL;

EXPORT_SYMBOL(do_unaligned_access);

void do_revinsn(unsigned long entry, unsigned long addr,
		unsigned long itype, struct pt_regs *regs,
		unsigned long oipc)
{
	siginfo_t si;

	printk("Reserved Instruction exception");
	printk(" -- Invalid instruction on 0x%lx.\n", addr);
	show_regs(regs);
	if (!user_mode(regs))
		do_exit(SIGILL);
	si.si_signo = SIGILL;
	si.si_errno = 0;
	force_sig_info(SIGILL, &si, current);
}
void coprocsr_default_handler(unsigned long itype, struct pt_regs *regs,
			      unsigned long eva)
{
	int sub_type = ((itype & ITYPE_mskSTYPE) >> ITYPE_offSTYPE);
	int coprocsr_id = ((itype & ITYPE_mskCPID) >> ITYPE_offCPID);
	char *etype_info = general_exception_cb[COPROCESSOR].etype_info;

	if ((true == fpu_coprocess_exist)
	    && (0 == coprocsr_id)) {
		sprintf(etype_info,
			"FPU Coprocessor. Please enable CONFIG_FPU");
	} else {
		sprintf(etype_info,
			"COP %d stype:%d. "
			"check COP trap handler and enable CONFIG", coprocsr_id,
			sub_type);
	}
	unhandled_exceptions(GENERAL_EXCEPTION, eva, itype, regs, 0);

}
void do_coprocesser_exception(unsigned long entry, unsigned long addr,
			      unsigned long itype, struct pt_regs *regs,
			      unsigned long oipc)
{
	int sub_type = ((itype & ITYPE_mskSTYPE) >> ITYPE_offSTYPE);
	int coprocsr_id = ((itype & ITYPE_mskCPID) >> ITYPE_offCPID);
	/* Add more debug information in etype_info for coprocessor. */
	char *etype_info = general_exception_cb[COPROCESSOR].etype_info;

	switch (sub_type) {
	case COPROCESS_DISABLE_EXCEPTION:
		coprocsr_trap[coprocsr_id].disable_handler(itype, regs, addr);
		break;
	case COPROCESS_EXCEPTION:
		coprocsr_trap[coprocsr_id].excpt_handler(itype, regs, addr);
		break;
	case COPROCESS_IMPRECISE_EXCEPTION:
		coprocsr_trap[coprocsr_id].impre_excpt_handler(itype, regs,
							       addr);
		break;
	default:
		sprintf(etype_info, "Coprocessor. subtype:%d\n", sub_type);
		unhandled_exceptions(entry, addr, itype, regs, oipc);
		break;
	}
}

static void do_alignment_check(unsigned long entry, unsigned long addr,
			       unsigned long itype, struct pt_regs *regs,
			       unsigned long oipc)
{
	unsigned long type = itype & (ITYPE_mskINST | ITYPE_mskETYPE);

	if (do_unaligned_access) {
		int ret = do_unaligned_access(entry, addr, type, regs);
		if (ret == 0)
			return;
		if (ret == -EFAULT)
			printk("Unhandled unaligned access exception\n");
	}
	do_page_fault(entry, addr, type, regs);

}
static void do_arithmetic(unsigned long entry, unsigned long addr,
			  unsigned long itype, struct pt_regs *regs,
			  unsigned long oipc)
{
	int sub_type = ((itype & ITYPE_mskSTYPE) >> ITYPE_offSTYPE);
	char *etype_info = general_exception_cb[ARITHMETIC].etype_info;

	if (INT_DIVIDE_BY_ZERO == sub_type)
		sprintf(etype_info, "Arithmetic. subtype: INT Divide by Zero");
	else
		sprintf(etype_info, "Arithmetic. subtype: Integer Overflow");
	unhandled_exceptions(entry, addr, itype, regs, oipc);
}

static void do_traps(unsigned long entry, unsigned long addr,
		     unsigned long itype, struct pt_regs *regs,
		     unsigned long oipc)
{
	unsigned int swid = itype >> ITYPE_offSWID;

	if (swid == 0x1a) {
		/* trap, used on v3 EDM target debugging workaround */
		/*
		 * DIPC(OIPC) is passed as parameter before
		 * interrupt is enabled, so the DIPC will not be corrupted
		 * even though interrupts are coming in
		 */
		/*
		 * 1. update ipc
		 * 2. update pt_regs ipc with oipc
		 * 3. update pt_regs ipsw (clear DEX)
		 */
		__asm__ volatile ("mtsr %0, $IPC\n\t"::"r" (oipc));
		regs->NDS32_ipc = oipc;
		if (regs->NDS32_pipsw & PSW_mskDEX) {
			printk("Nested Debug exception is possibly happened\n");
			printk("ipc:%08x pipc:%08x\n",
			       (unsigned int)regs->NDS32_ipc,
			       (unsigned int)regs->NDS32_pipc);
		}
#ifdef CONFIG_NDS32_EDM_SUP
		nds32_disable_all_edm_bpc();
#endif
		do_debug_trap(entry, addr, itype, regs);
		regs->NDS32_ipsw &= ~PSW_mskDEX;
	} else {
		unhandled_exceptions(entry, addr, itype, regs, oipc);
	}

}

/*
 * 7:DEF dispatch for General exception handler
 */
void do_dispatch_general(unsigned long entry, unsigned long addr,
			 unsigned long itype, struct pt_regs *regs,
			 unsigned long oipc)
{
	unsigned long etype = itype & ITYPE_mskETYPE;

	general_exception_cb[etype].handler(entry, addr, itype, regs, oipc);
}
