/*
 * linux/arch/nds32/math-emu/fpuemu.c: a nds32 coprocessor (fpu) instruction emulator
 *
 * MIPS floating point support
 * Copyright (C) 1994-2000 Algorithmics Ltd.
 * http://www.algor.co.uk
 *
 * Kevin D. Kissell, kevink@mips.com and Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 2000  MIPS Technologies, Inc.
 * Copyright (C) 2009 Andes Technology Corporation
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * A complete emulator for MIPS coprocessor 1 instructions.  This is
 * required for #float(switch) or #float(trap), where it catches all
 * COP1 instructions via the "CoProcessor Unusable" exception.
 *
 * More surprisingly it is also required for #float(ieee), to help out
 * the hardware fpu at the boundaries of the IEEE-754 representation
 * (denormalised values, infinities, underflow, etc).  It is made
 * quite nasty because emulation of some non-COP1 instructions is
 * required, e.g. in branch delay slots.
 *
 * Note if you know that you won't have an fpu, then you'll get much
 * better performance by compiling with -msoft-float!
 */
#include <linux/sched.h>
#include <linux/debugfs.h>

#include <asm/processor.h>
#include <asm/ptrace.h>
#include <asm/bitfield.h>
#include <asm/uaccess.h>

#include "ieee754.h"
#include "insn.h"
#include "fpu_emulator.h"

/* Function which emulates a floating point instruction. */

static int fpu_emu(struct pt_regs *, struct fpu_struct *, unsigned long);

/* Further private data for which no space exists in fpu_struct */

struct fpu_emulator_stats fpuemustats;

/* rounding mode */
#define FPU_FPCSR_RN      0x0	/* nearest */
#define FPU_FPCSR_RU      0x1	/* towards +Infinity */
#define FPU_FPCSR_RD      0x2	/* towards -Infinity */
#define FPU_FPCSR_RZ      0x3	/* towards zero */

/* Convert Mips rounding mode (0..3) to IEEE library modes. */
static const unsigned char ieee_rm[4] = {
	[FPU_FPCSR_RN] = IEEE754_RN,
	[FPU_FPCSR_RU] = IEEE754_RU,
	[FPU_FPCSR_RD] = IEEE754_RD,
	[FPU_FPCSR_RZ] = IEEE754_RZ,
};

/* Convert IEEE library modes to NDS32 rounding mode (0..3). */
static const unsigned char nds32_rm[4] = {
	[IEEE754_RN] = FPU_FPCSR_RN,
	[IEEE754_RZ] = FPU_FPCSR_RZ,
	[IEEE754_RD] = FPU_FPCSR_RD,
	[IEEE754_RU] = FPU_FPCSR_RU,
};

/*
 * In the Linux kernel, we support selection of FPR format on the
 * basis of the Status.FR bit.  This does imply that, if a full 32
 * FPRs are desired, there needs to be a flip-flop that can be written
 * to one at that bit position.  In any case, O32 MIPS ABI uses
 * only the even FPRs (Status.FR = 0).
 */

#define CP0_STATUS_FR_SUPPORT

#ifdef CP0_STATUS_FR_SUPPORT
#define FR_BIT ST0_FR
#else
#define FR_BIT 0
#endif

#define DIFROMREG(di, x) ((di) = (unsigned long long)ptr[2*x+1] << 32 | (unsigned long long)ptr[2*x])
#define DITOREG(di, x) ptr[2*x+1] = (di) >> 32; ptr[2*x] = (unsigned long)(di)
#ifdef __NDS32_EL__
#define SIFROMREG(si, x)\
	((si) = *((unsigned long *)ctx + (x^1)))
#define SITOREG(si, x)\
	(*((unsigned long *)ctx + (x^1))= (si))
#else
#define SIFROMREG(si, x) ((si) = *((unsigned long *)ctx + x))
#define SITOREG(si, x)	(*((unsigned long *)ctx + x)= (si))
#endif

#define SPFROMREG(sp, x) SIFROMREG((sp).bits, x)
#define SPTOREG(sp, x)	SITOREG((sp).bits, x)
#define DPFROMREG(dp, x)	DIFROMREG((dp).bits, x)
#define DPTOREG(dp, x)	DITOREG((dp).bits, x)

#define DEF3OP(name, p, f1, f2, f3) \
static ieee754##p fpemu_##p##_##name(ieee754##p r, ieee754##p s, \
    ieee754##p t) \
{ \
        struct _ieee754_csr ieee754_csr_save; \
        s = f1(s, t); \
        ieee754_csr_save = ieee754_csr; \
        s = f2(s, r); \
        ieee754_csr_save.cx |= ieee754_csr.cx; \
        ieee754_csr_save.sx |= ieee754_csr.sx; \
        s = f3(s); \
        ieee754_csr.cx |= ieee754_csr_save.cx; \
        ieee754_csr.sx |= ieee754_csr_save.sx; \
        return s; \
}

DEF3OP(madd, sp, ieee754sp_mul, ieee754sp_add,);
DEF3OP(msub, sp, ieee754sp_mul, ieee754sp_sub,);
DEF3OP(nmadd, sp, ieee754sp_mul, ieee754sp_add, ieee754sp_neg);
DEF3OP(nmsub, sp, ieee754sp_mul, ieee754sp_sub, ieee754sp_neg);
DEF3OP(madd, dp, ieee754dp_mul, ieee754dp_add,);
DEF3OP(msub, dp, ieee754dp_mul, ieee754dp_sub,);
DEF3OP(nmadd, dp, ieee754dp_mul, ieee754dp_add, ieee754dp_neg);
DEF3OP(nmsub, dp, ieee754dp_mul, ieee754dp_sub, ieee754dp_neg);

/*
 * Emulate the floating point instruction w/ denorm input pointed at by IPC.
 * According to spec, only FPU arithmetic, data format conversion, and compare instructions.
 */

static int fpuEmulate(struct pt_regs *regs, struct fpu_struct *fpu)
{
	unsigned long insn = 0, addr = regs->NDS32_ipc;
	unsigned long emulpc, contpc;
	unsigned char *pc = (void *)&insn;
	char c;
	int i = 0;

	for (i = 0; i < 4; i++) {
		if (__get_user(c, (unsigned char *)addr++))
			return SIGBUS;
		*pc++ = c;
	}

	insn = be32_to_cpu(insn);

	emulpc = regs->NDS32_ipc;
	contpc = regs->NDS32_ipc + 4;

	fpuemustats.emulated++;
	switch (NDS32Insn_OPCODE(insn)) {
	case cop0_op:
		switch (NDS32Insn_OPCODE_COP0(insn)) {

		case fs1_op:
		case fs2_op:
		case fd1_op:
		case fd2_op:
			{
				int sig;

				/* a real fpu computation instruction */
				if ((sig = fpu_emu(regs, fpu, insn)))
					return sig;
			}
			break;

		default:
			return SIGILL;
		}
		break;
	default:
		return SIGILL;
	}

	/* we did it !! */
	regs->NDS32_ipc = contpc;

	return 0;
}

/*
 * Conversion table from NDS32 fcmp TYP
 * cond = ieee754dp_cmp(x,y,IEEE754_UN,sig);
 */
static const unsigned char cmptab[8] = {
	IEEE754_CEQ,
	IEEE754_CLT,
	IEEE754_CLT | IEEE754_CEQ,
	IEEE754_CUN,
	0,
	0,
	0,
	0,
};

/*
 * Emulate a single FPU arithmetic instruction.
 */
static int fpu_emu(struct pt_regs *xcp, struct fpu_struct *ctx,
		   unsigned long insn)
{
	int rfmt;		/* resulting format */
	unsigned rfpcsr = 0;	/* resulting csr */
	unsigned long *ptr = (unsigned long *)ctx;
	union {
		ieee754dp d;
		ieee754sp s;
		int w;
	} rv;			/* resulting value */

	fpuemustats.fpuops++;
	switch (rfmt = NDS32Insn_OPCODE_COP0(insn)) {
	case fs1_op:{
			union {
				ieee754sp(*t) (ieee754sp, ieee754sp, ieee754sp);
				ieee754sp(*b) (ieee754sp, ieee754sp);
				ieee754sp(*u) (ieee754sp);
			} handler;

			switch (NDS32Insn_OPCODE_BIT69(insn)) {
			case fadds_op:
				handler.b = ieee754sp_add;
				goto scopbop;
			case fsubs_op:
				handler.b = ieee754sp_sub;
				goto scopbop;
			case fmadds_op:
				handler.t = fpemu_sp_madd;
				goto scoptop;
			case fmsubs_op:
				handler.t = fpemu_sp_msub;
				goto scoptop;
			case fnmadds_op:
				handler.t = fpemu_sp_nmadd;
				goto scoptop;
			case fnmsubs_op:
				handler.t = fpemu_sp_nmsub;
				goto scoptop;
			case fmuls_op:
				handler.b = ieee754sp_mul;
				goto scopbop;
			case fdivs_op:
				handler.b = ieee754sp_div;
				goto scopbop;

				/* binary op on handler */
scoptop:
				{
					ieee754sp fd, fr, fs, ft;

					SPFROMREG(fr,
						  NDS32Insn_OPCODE_Rt(insn));
					SPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					SPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));
					fd = (*handler.t) (fr, fs, ft);
					SPTOREG(fd, NDS32Insn_OPCODE_Rt(insn));
				}
scopbop:
				{
					ieee754sp fs, ft;

					SPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					SPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));

					rv.s = (*handler.b) (fs, ft);
					goto copcsr;
				}
scopuop:
				{
					ieee754sp fs;

					SPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					rv.s = (*handler.u) (fs);
					goto copcsr;
				}
copcsr:
				if (ieee754_cxtest(IEEE754_INEXACT))
					rfpcsr |= FPCSR_mskIEX;
				if (ieee754_cxtest(IEEE754_UNDERFLOW))
					rfpcsr |= FPCSR_mskUDF;
				if (ieee754_cxtest(IEEE754_OVERFLOW))
					rfpcsr |= FPCSR_mskOVF;
				if (ieee754_cxtest(IEEE754_ZERO_DIVIDE))
					rfpcsr |= FPCSR_mskDBZ;
				if (ieee754_cxtest(IEEE754_INVALID_OPERATION))
					rfpcsr |= FPCSR_mskIVO;
				break;

			case fs1_f2op_op:
				switch (NDS32Insn_OPCODE_BIT1014(insn)) {
				case fs2d_op:{
						ieee754sp fs;
						SPFROMREG(fs,
							  NDS32Insn_OPCODE_Ra
							  (insn));
						rv.d = ieee754dp_fsp(fs);
						rfmt = fd1_op;
						goto copcsr;
					}
				case fsqrts_op:
					handler.u = ieee754sp_sqrt;
					goto scopuop;
				case fabss_op:
					handler.u = ieee754sp_abs;
					goto scopuop;
				default:
					return SIGILL;
				}
			default:
				return SIGILL;
			}
			break;
		}

	case fs2_op:
		switch (NDS32Insn_OPCODE_BIT69(insn)) {
		case fcmpeqs_op:
		case fcmpeqs_e_op:
		case fcmplts_op:
		case fcmplts_e_op:
		case fcmples_op:
		case fcmples_e_op:
		case fcmpuns_op:
		case fcmpuns_e_op:
			{
				unsigned int cmpop =
				    NDS32Insn_OPCODE_BIT69(insn);
				if (cmpop < 0x8) {
					ieee754sp fs, ft;

					SPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					SPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));
					rv.w =
					    ieee754sp_cmp(fs, ft,
							  cmptab[(cmpop >> 1) &
								 0x7],
							  cmpop & 0x1);
					if ((cmpop & 0x1)
					    &&
					    ieee754_cxtest
					    (IEEE754_INVALID_OPERATION))
						rfpcsr = FPCSR_mskIVO;
					else
						goto copcsr;
				} else
					return SIGILL;
			}
			break;

		default:
			return SIGILL;
		}
		break;

	case fd1_op:
		{
			union {
				ieee754dp(*t) (ieee754dp, ieee754dp, ieee754dp);
				ieee754dp(*b) (ieee754dp, ieee754dp);
				ieee754dp(*u) (ieee754dp);
			} handler;

			switch (NDS32Insn_OPCODE_BIT69(insn)) {
			case faddd_op:
				handler.b = ieee754dp_add;
				goto dcopbop;
			case fsubd_op:
				handler.b = ieee754dp_sub;
				goto dcopbop;
			case fmaddd_op:
				handler.t = fpemu_dp_madd;
				goto tdcoptop;
			case fmsubd_op:
				handler.t = fpemu_dp_msub;
				goto tdcoptop;
			case fnmaddd_op:
				handler.t = fpemu_dp_nmadd;
				goto tdcoptop;
			case fnmsubd_op:
				handler.t = fpemu_dp_nmsub;
				goto tdcoptop;

tdcoptop:
				{
					ieee754dp fd, fr, fs, ft;

					DPFROMREG(fr,
						  NDS32Insn_OPCODE_Rt(insn));
					DPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					DPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));
					fd = (*handler.t) (fr, fs, ft);
					DPTOREG(fd, NDS32Insn_OPCODE_Rt(insn));
					goto copcsr;
				}

			case fmuld_op:
				handler.b = ieee754dp_mul;
				goto dcopbop;
			case fdivd_op:
				handler.b = ieee754dp_div;
				goto dcopbop;

				/* binary op on handler */
dcopbop:			{
					ieee754dp fs, ft;

					DPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					DPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));

					rv.d = (*handler.b) (fs, ft);
					goto copcsr;
				}
dcopuop:			{
					ieee754dp fs;

					DPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					rv.d = (*handler.u) (fs);
					goto copcsr;
				}

			case fd1_f2op_op:
				switch (NDS32Insn_OPCODE_BIT1014(insn)) {
				case fd2s_op:{
						ieee754dp fs;

						DPFROMREG(fs,
							  NDS32Insn_OPCODE_Ra
							  (insn));
						rv.s = ieee754sp_fdp(fs);
						rfmt = fd1_op;
						goto copcsr;
					}
				case fsqrtd_op:
					handler.u = ieee754dp_sqrt;
					goto dcopuop;
				case fabsd_op:
					handler.u = ieee754dp_abs;
					goto dcopuop;
				default:
					return SIGILL;
				}
			default:
				return SIGILL;
			}
			break;
		}

	case fd2_op:
		switch (NDS32Insn_OPCODE_BIT69(insn)) {
		case fcmpeqd_op:
		case fcmpeqd_e_op:
		case fcmpltd_op:
		case fcmpltd_e_op:
		case fcmpled_op:
		case fcmpled_e_op:
		case fcmpund_op:
		case fcmpund_e_op:
			{
				unsigned cmpop = NDS32Insn_OPCODE_BIT69(insn);
				if (cmpop < 0x8) {
					ieee754dp fs, ft;

					DPFROMREG(fs,
						  NDS32Insn_OPCODE_Ra(insn));
					DPFROMREG(ft,
						  NDS32Insn_OPCODE_Rb(insn));
					rv.w =
					    ieee754dp_cmp(fs, ft,
							  cmptab[(cmpop >> 1) &
								 0x7],
							  cmpop & 0x1);
					rfmt = fs2_op;
					if ((cmpop & 0x1)
					    &&
					    ieee754_cxtest
					    (IEEE754_INVALID_OPERATION))
						rfpcsr = FPCSR_mskIVO;
					else
						goto copcsr;
				} else
					return SIGILL;
			}
			break;
		default:
			return SIGILL;
		}
		break;

	default:
		return SIGILL;
	}

	/*
	 * Now we can safely write the result back to the register file.
	 */
	switch (rfmt) {
	case fd1_op:
	case fd2_op:
		DPTOREG(rv.d, NDS32Insn_OPCODE_Rt(insn));
		break;
	case fs1_op:
	case fs2_op:
		SPTOREG(rv.s, NDS32Insn_OPCODE_Rt(insn));
		break;
	default:
		return SIGILL;
	}

	/*
	 * Update the fpu CSR register for this operation.
	 * If an exception is required, generate a tidy SIGFPE exception,
	 * without updating the result register.
	 * Note: cause exception bits do not accumulate, they are rewritten
	 * for each op; only the flag/sticky bits accumulate.
	 */
	ctx->fpcsr = (ctx->fpcsr & ~FPCSR_mskALL) | rfpcsr;
	if ((ctx->fpcsr << 5) & ctx->fpcsr & FPCSR_mskALLE) {
		return SIGFPE;
	}

	return 0;
}

int do_fpu_denorm(struct pt_regs *regs, struct fpu_struct *fpu)
{
	int sig = 0;

	/*
	 * The 'ieee754_csr' is an alias of
	 * fpcsr->RM.  No need to copy fpcsr->RM to
	 * ieee754_csr.  But ieee754_csr.rm is ieee
	 * library modes. (not NDS32 rounding mode)
	 */
	/* convert to ieee library modes */
	ieee754_csr.rm = ieee_rm[ieee754_csr.rm];
	sig = fpuEmulate(regs, fpu);
	/* revert to NDS32 rounding mode */
	ieee754_csr.rm = nds32_rm[ieee754_csr.rm];

	return sig;
}
