/*
 *  linux/arch/nds32/mm/ioremap.c
 *
 * Re-map IO memory to kernel address space so that we can access it.
 *
 * (C) Copyright 1995 1996 Linus Torvalds
 * Copyright (C) 2009 Andes Technology Corporation
 *
 * Hacked for ARM by Phil Blundell <philb@gnu.org>
 * Hacked to allow all architectures to build, and various cleanups
 * by Russell King
 *
 * This allows a driver to remap an arbitrary region of bus memory into
 * virtual space.  One should *only* use readl, writel, memcpy_toio and
 * so on with such remapped areas.
 *
 * Because the ARM only has a 32-bit address space we can't address the
 * whole of the (physical) PCI space at once.  PCI huge-mode addressing
 * allows us to circumvent this restriction by splitting PCI space into
 * two 2GB chunks and mapping only one at a time into processor memory.
 * We use MMU protection domains to trap any attempt to access the bank
 * that is not currently mapped.  (This isn't fully implemented yet.)
 */
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/io.h>
#include <asm/cacheflush.h>

/*
 * Remap an arbitrary physical address space into the kernel virtual
 * address space. Needed when the kernel wants to access high addresses
 * directly.
 *
 * NOTE! We need to allow non-page-aligned mappings too: we will obviously
 * have to convert them into an offset in a page-aligned mapping, but the
 * caller shouldn't need to know that small detail.
 *
 * 'flags' are the extra L_PTE_ flags that you want to specify for this
 * mapping.  See include/asm-arm/proc-armv/pgtable.h for more information.
 */
void __iomem *__ioremap(unsigned long phys_addr, size_t size,
			unsigned long flags, unsigned long align)
{
	struct vm_struct *area;
	unsigned long addr, offset, last_addr;
	pgprot_t prot;

	/* Don't allow wraparound or zero size */
	last_addr = phys_addr + size - 1;
	if (!size || last_addr < phys_addr)
		return NULL;

	/*
	 * Mappings have to be page-aligned
	 */
	offset = phys_addr & ~PAGE_MASK;
	phys_addr &= PAGE_MASK;
	size = PAGE_ALIGN(last_addr + 1) - phys_addr;

	/*
	 * Ok, go for it..
	 */
	area = get_vm_area(size, VM_IOREMAP);
	if (!area)
		return NULL;

	area->phys_addr = phys_addr;
	addr = (unsigned long)area->addr;
	/* TODO: verify this value for ioremap */
	prot = __pgprot(_PAGE_V | _PAGE_M_KRW | _PAGE_D |
			_PAGE_G | _PAGE_C_DEV | flags);
	/* TODO: verify this use generic ioremap_page_range instead of
	 * self's remap_area_pages */
	if (ioremap_page_range(addr, addr + size, phys_addr, prot)) {
		vunmap((void *)addr);
		return NULL;
	}
	return (__force void __iomem *)(offset + (char *)addr);

}

EXPORT_SYMBOL(__ioremap);

void __iounmap(void __iomem * addr)
{
	vunmap((void *)(PAGE_MASK & (unsigned long)addr));
}

EXPORT_SYMBOL(__iounmap);
