/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2004, 2005 by Ralf Baechle
 * Copyright (C) 2005 by MIPS Technologies, Inc.
 * Copyright (C) 2007 Andes Technology Corporation
 */
#include <linux/oprofile.h>
#include <linux/irqdomain.h>
#include <linux/interrupt.h>
#include <linux/smp.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <asm/irq_regs.h>
#include <asm/nds32.h>
#include <asm/uaccess.h>
#include <asm/pfm.h>

#include "op_impl.h"

#define NDS32_PERFCTL_EN(num)		(1 << num)
#define NDS32_PERFCTL_INTEN             (1UL << 3)
#define NDS32_PERFCTL_OVERFLOW		(1UL << 6)
#define NDS32_PERFCTL_NOKERNEL		(1UL << 9)
#define NDS32_PERFCTL_NOUSER		(1UL << 12)
#define NDS32_PERFCTL_EVENT_0(event)	(event << 15)
#define NDS32_PERFCTL_EVENT_1(event)	(event << 16)
#define NDS32_PERFCTL_EVENT_2(event)	(event << 22)

static unsigned long long ov0, ov1, ov2;
static int syscall = 0;

struct pmu_counter {
	volatile unsigned long ovf;
	unsigned long reset_counter;
};

enum { PFMC0, PFMC1, PFMC2, MAX_COUNTERS };

static struct pmu_counter results[MAX_COUNTERS];

static inline unsigned int read_perfcntr(int counter)
{
	switch (counter) {
	case PFMC0:
		return GET_PFMC0();
		break;
	case PFMC1:
		return GET_PFMC1();
		break;
	case PFMC2:
		return GET_PFMC2();
		break;
	default:
		printk(KERN_ERR
		       "Oprofile read_perfcntr: CPU has no %d performance counters\n",
		       counter);
	}

	return 0;
}

static inline unsigned int read_perfctrl(void)
{
	return GET_PFM_CTL();
}

static inline void write_perfcntr(int counter, unsigned int value)
{
	switch (counter) {
	case PFMC0:
		SET_PFMC0(value);
		break;
	case PFMC1:
		SET_PFMC1(value);
		break;
	case PFMC2:
		SET_PFMC2(value);
		break;
	default:
		printk(KERN_ERR
		       "Oprofile write_perfcntr: CPU has no %d performance counters\n",
		       counter);
	}
}

static inline void write_perfctrl(unsigned int value)
{
	SET_PFM_CTL(value);
}

struct op_nds32_model op_model_nds32_ops;

static struct nds32_register_config {
	unsigned int control[3];
	unsigned int counter[3];
} reg;

/* Compute all of the registers in preparation for enabling profiling.  */

static void nds32_reg_setup(struct op_counter_config *ctr)
{
	unsigned int counters = op_model_nds32_ops.num_counters;
	int i;

	/* Compute the performance counter control word.  */
	/* For now count kernel and user mode */
	for (i = 0; i < counters; i++) {
		reg.control[i] = 0;
		reg.counter[i] = 0;

		if (!ctr[i].enabled)
			continue;

		switch (i) {
		case 0:
			reg.control[i] = NDS32_PERFCTL_EVENT_0(ctr[i].event) |
			    (NDS32_PERFCTL_INTEN << i);
			break;
		case 1:
			reg.control[i] = NDS32_PERFCTL_EVENT_1(ctr[i].event) |
			    (NDS32_PERFCTL_INTEN << i);
			break;
		case 2:
			reg.control[i] = NDS32_PERFCTL_EVENT_2(ctr[i].event) |
			    (NDS32_PERFCTL_INTEN << i);
			break;
		default:
			printk(KERN_ERR
			       "Oprofile nds32_reg_setup: CPU has no %d performance counters\n",
			       i);
		}
		if (!(ctr[i].kernel))
			reg.control[i] |= (NDS32_PERFCTL_NOKERNEL << i);
		if (!(ctr[i].user))
			reg.control[i] |= (NDS32_PERFCTL_NOUSER << i);
		reg.counter[i] = -ctr[i].count;
	}
}

/* Program all of the registers in preparation for enabling profiling.  */

static void nds32_cpu_setup(void *args)
{
	unsigned int counters = op_model_nds32_ops.num_counters;

	switch (counters) {
	case 3:
		write_perfcntr(2, reg.counter[2]);
	case 2:
		write_perfcntr(1, reg.counter[1]);
	case 1:
		write_perfcntr(0, reg.counter[0]);
	}
	write_perfctrl(0);
}

/* Start all counters on current CPU */
static void nds32_cpu_start(void *args)
{
	unsigned int counters = op_model_nds32_ops.num_counters;
	unsigned int value = 0;

	switch (counters) {
	case 3:
		if (reg.control[2])
			value |= (NDS32_PERFCTL_EN(2) | reg.control[2]);
	case 2:
		if (reg.control[1])
			value |= (NDS32_PERFCTL_EN(1) | reg.control[1]);
	case 1:
		if (reg.control[0])
			value |= (NDS32_PERFCTL_EN(0) | reg.control[0]);
	}
	write_perfctrl(value);
}

/* Stop all counters on current CPU */
static void nds32_cpu_stop(void *args)
{
	write_perfctrl(0);
}

static irqreturn_t nds32_perfcount_handler(int irq, void *dev_id)
{
	unsigned int control, i;

	control = read_perfctrl();
	write_perfctrl(0);

	if (syscall) {
		if (control & PFM_CTL_mskOVF0)
			ov0++;
		if (control & PFM_CTL_mskOVF1)
			ov1++;
		if (control & PFM_CTL_mskOVF2)
			ov2++;
	} else {
		for (i = 0; i < MAX_COUNTERS; i++) {
			if ((control & (NDS32_PERFCTL_INTEN << i))
			    && (control & (NDS32_PERFCTL_OVERFLOW << i))) {
				oprofile_add_sample(get_irq_regs(), i);
				write_perfcntr(i, reg.counter[i]);
				break;
			}
		}
	}
	write_perfctrl(control);

	return IRQ_HANDLED;
}

static inline int n_counters(void)
{
	return 3;
}

static inline void reset_counters(int counters)
{
	switch (counters) {
	case 3:
		write_perfcntr(2, 0);
	case 2:
		write_perfcntr(1, 0);
	case 1:
		write_perfcntr(0, 0);
	}
	write_perfctrl(0);
}

static int __init nds32_init(void)
{
	int counters, irq_num, ret;
	struct device_node *pmu_node;
	struct platform_device *pmu_device;

	counters = n_counters();
	if (counters == 0) {
		printk(KERN_ERR "Oprofile: CPU has no performance counters\n");
		return -ENODEV;
	}

	reset_counters(counters);
	op_model_nds32_ops.cpu_type = "nds32";
	op_model_nds32_ops.num_counters = counters;

	pmu_node = of_find_node_by_name(NULL, "pmu");
	pmu_device = of_find_device_by_node(pmu_node);
	irq_num = platform_get_irq(pmu_device, 0);

	ret = request_irq(irq_num, nds32_perfcount_handler, IRQF_SHARED,
			  "NDS32 PFMCNTR", (void *)results);
	if (ret < 0) {
		printk(KERN_ERR "oprofile: unable to request IRQ%d\n", irq_num);
		return ret;
	}
	return 0;
}

static void nds32_exit(void)
{
	struct device_node *pmu_node = of_find_node_by_name(NULL, "pmu");
	struct platform_device *pmu_device = of_find_device_by_node(pmu_node);
	int irq_num = platform_get_irq(pmu_device, 0);
	free_irq(irq_num, results);
	reset_counters(op_model_nds32_ops.num_counters);
}

void sys_pfmctl(int event0, int event1, int event2, int start)
{
	unsigned int ctl = 0;

	if (start) {
		syscall = 1;
		if (event0 >= 0)
			ctl |=
			    (NDS32_PERFCTL_EVENT_0(event0) |
			     (NDS32_PERFCTL_INTEN << 0) | NDS32_PERFCTL_EN(0));
		if (event1 >= 0)
			ctl |=
			    (NDS32_PERFCTL_EVENT_1(event1) |
			     (NDS32_PERFCTL_INTEN << 1) | NDS32_PERFCTL_EN(1));
		if (event2 >= 0)
			ctl |=
			    (NDS32_PERFCTL_EVENT_2(event2) |
			     (NDS32_PERFCTL_INTEN << 2) | NDS32_PERFCTL_EN(2));
	} else {
		syscall = 0;
		if (event0 >= 0)
			ctl &=
			    ~((NDS32_PERFCTL_INTEN << 0) | NDS32_PERFCTL_EN(0));
		if (event1 >= 0)
			ctl &=
			    ~((NDS32_PERFCTL_INTEN << 1) | NDS32_PERFCTL_EN(1));
		if (event2 >= 0)
			ctl &=
			    ~((NDS32_PERFCTL_INTEN << 2) | NDS32_PERFCTL_EN(2));
	}

	write_perfctrl(ctl);

}

int sys_getpfm(struct pcounter __user * p)
{
	struct pcounter pfm;
	unsigned int control;

	control = read_perfctrl();
	write_perfctrl(0);

	pfm.pfm0 = ov0 << 32 | GET_PFMC0();
	pfm.pfm1 = ov1 << 32 | GET_PFMC1();
	pfm.pfm2 = ov2 << 32 | GET_PFMC2();

	if (copy_to_user(p, &pfm, sizeof(pfm)))
		return -EFAULT;

	write_perfctrl(control);

	return 0;
}

int sys_setpfm(int pfm0, int pfm1, int pfm2, struct pcounter __user * p)
{
	struct pcounter pfm;
	unsigned int control;

	control = read_perfctrl();
	write_perfctrl(0);

	if (copy_from_user(&pfm, p, sizeof(pfm)))
		return -EFAULT;

	if (pfm0) {
		SET_PFMC0((unsigned int)(pfm.pfm0 & 0xffffffff));
		ov0 = pfm.pfm0 >> 32;
	}
	if (pfm1) {
		SET_PFMC1((unsigned int)(pfm.pfm1 & 0xffffffff));
		ov1 = pfm.pfm1 >> 32;
	}
	if (pfm2) {
		SET_PFMC2((unsigned int)(pfm.pfm2 & 0xffffffff));
		ov2 = pfm.pfm2 >> 32;
	}

	write_perfctrl(control);

	return 0;
}

struct op_nds32_model op_model_nds32_ops = {
	.reg_setup = nds32_reg_setup,
	.cpu_setup = nds32_cpu_setup,
	.init = nds32_init,
	.exit = nds32_exit,
	.cpu_start = nds32_cpu_start,
	.cpu_stop = nds32_cpu_stop,
};
