/*
 *  Andestech ATCPIT100 Timer Device Driver Implementation
 *
 *  Copyright (C) 2016 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/irq.h>
#include <linux/clocksource.h>
#include <linux/clockchips.h>

#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/sched.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>

void __iomem *base;
static u32 freq;
#define tmr_writel(value, reg) __raw_writel(value, base + (reg))
#define tmr_readl(reg) __raw_readl(base + (reg))

/*
 * Definition of register offsets
 */

/* ID and Revision Register */
#define ID_REV		0x0

/* Configuration Register */
#define CFG		0x10

/* Interrupt Enable Register */
#define INT_EN		0x14
#define CH_INT_EN(c,i)	(1<<i)<<(4*c)

/* Interrupt Status Register */
#define INT_STA		0x18
#define CH_INT_STA(c,i)	(1<<i)<<(4*c)

/* Channel Enable Register */
#define CH_EN		0x1C
#define CH_TMR_EN(c,t)	(1<<t)<<(4*c)

/* Ch n Control REgister */
#define CH_CTL(n)	(0x20+0x10*n)
/* Channel clock source , bit 3 , 0:External clock , 1:APB clock */
#define APB_CLK		(1<<3)
/* Channel mode , bit 0~2 */
#define TMR_32		1
#define TMR_16		2
#define TMR_8		3
#define PWM		4

#define CH_REL(n)	(0x24+0x10*n)
#define CH_CNT(n)	(0x28+0x10*n)

static inline cycle_t clocksource_read_cycles(struct clocksource *cs)
{
	return (cycle_t) ~(tmr_readl(CH_CNT(1))+0xffffffff);
}

static void clksrc_atcpi1001_resume(struct clocksource *cs)
{
	tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(1,0), CH_EN);
}

static struct clocksource clksrc_atcpit1001 = {
	.name	= "atcpit1001_tm1",
	.rating	= 300,
	.read	= clocksource_read_cycles,
	.mask	= CLOCKSOURCE_MASK(32),
	.shift	= 21,
	.flags	= CLOCK_SOURCE_IS_CONTINUOUS,
	.resume = clksrc_atcpi1001_resume,
};

static void atcpit1001_clocksource_init(void)
{
	clksrc_atcpit1001.mult =
		clocksource_hz2mult(freq, clksrc_atcpit1001.shift);

	tmr_writel(0xffffffff, CH_REL(1));
	tmr_writel(APB_CLK|TMR_32, CH_CTL(1));
	tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(1,0), CH_EN);
	if (clocksource_register(&clksrc_atcpit1001))
		printk(KERN_ERR "Error: failed to register %s\n", clksrc_atcpit1001.name);
}

static int atcpit1001_set_next_event(unsigned long cycles,
		struct clock_event_device *evt)
{
	tmr_writel(cycles, CH_REL(0));
	return 0;
}

static void atcpit1001_set_mode(enum clock_event_mode mode,
		struct clock_event_device *evt)
{
	switch (mode) {
		case CLOCK_EVT_MODE_ONESHOT:
			tmr_writel(0xffffffff, CH_REL(0));
			tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(0,0), CH_EN);
			break;

		case CLOCK_EVT_MODE_PERIODIC:
			tmr_writel(freq / HZ - 1, CH_CNT(0));
			tmr_writel(freq / HZ - 1, CH_REL(0));
			tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(0,0), CH_EN);
			break;
		case CLOCK_EVT_MODE_UNUSED:
			break;
		case CLOCK_EVT_MODE_SHUTDOWN:
			tmr_writel(tmr_readl(CH_EN) & ~CH_TMR_EN(0,0), CH_EN);
			break;
		case CLOCK_EVT_MODE_RESUME:
			tmr_writel(tmr_readl(INT_STA) | CH_INT_STA(0,0), INT_STA);
			tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(0,0), CH_EN);
			break;
	}
}

static struct clock_event_device clockevent_atcpit1001 = {
	.name		= "atcpit1001_tm1",
	.features       = CLOCK_EVT_FEAT_ONESHOT | CLOCK_EVT_FEAT_PERIODIC,
	.shift		= 32,
	.cpumask	= cpu_all_mask,
	.set_next_event	= atcpit1001_set_next_event,
	.set_mode	= atcpit1001_set_mode,
};

static irqreturn_t timer1_interrupt(int irq, void *dev_id)
{
	struct clock_event_device *evt = dev_id;

	tmr_writel(tmr_readl(INT_STA) | CH_INT_STA(0,0), INT_STA);

	evt->event_handler(evt);

	return IRQ_HANDLED;
}

static struct irqaction timer1_irq = {
	.name		= "Timer Tick",
	.flags		= IRQF_DISABLED | IRQF_TIMER | IRQF_IRQPOLL,
	.handler	= timer1_interrupt,
	.dev_id		= &clockevent_atcpit1001
};

static void __init atcpit1001_clockevent_init(int irq)
{
	clockevent_atcpit1001.mult = div_sc(freq, NSEC_PER_SEC, clockevent_atcpit1001.shift);
	clockevent_atcpit1001.max_delta_ns = clockevent_delta2ns(0xffffffff, &clockevent_atcpit1001);
	clockevent_atcpit1001.min_delta_ns = clockevent_delta2ns(3, &clockevent_atcpit1001);

	clockevents_register_device(&clockevent_atcpit1001);
	setup_irq(irq, &timer1_irq);
}

int clksrc_init;
static void __init atcpit100_init(struct device_node *dev)
{
	int irq;

	base = of_iomap(dev, 0);
	if (!base)
		panic("Can't remap registers");

	if (of_property_read_u32(dev, "clock-frequency", &freq))
		panic("Can't read clock-frequency");

	irq = irq_of_parse_and_map(dev, 0);
	if (irq <= 0)
		panic("Failed to map timer IRQ\n");

	printk("ATCPIT100 timer 1 installed on IRQ %d, with clock %d at %d HZ. in 0x%08x\r\n",
			irq, freq, HZ, (u32)base);

	tmr_writel(APB_CLK|TMR_32, CH_CTL(0));
	tmr_writel(tmr_readl(INT_EN) | CH_INT_EN(0,0), INT_EN);
	tmr_writel(tmr_readl(CH_EN) | CH_TMR_EN(0,0), CH_EN);

	atcpit1001_clocksource_init();
	atcpit1001_clockevent_init(irq);
	clksrc_init = 1;
}

unsigned long long notrace sched_clock(void)
{
	if (clksrc_init)
		return clocksource_cyc2ns(clocksource_read_cycles(&clksrc_atcpit1001),
			clksrc_atcpit1001.mult, clksrc_atcpit1001.shift);
	else
		return (unsigned long long)(jiffies - INITIAL_JIFFIES)
	                                        * (NSEC_PER_SEC / HZ);
}

CLOCKSOURCE_OF_DECLARE(ae3xx_atcpit100, "andestech,atcpit100", atcpit100_init);
