/*
 *  Andes ATFTMR010 Timer Device Driver Implementation
 *
 *  Copyright (C) 2005 Faraday Corp. (http://www.faraday-tech.com)
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/irq.h>
#include <linux/clocksource.h>
#include <linux/clockchips.h>

#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/sched.h>
#include <linux/of_address.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>

void __iomem *base;
static u32 freq;
#define tmr_writel(value, reg) __raw_writel(value, base + (reg))
#define tmr_readl(reg) __raw_readl(base + (reg))

/*
 * Definition of register offsets
 */

#define TIMER1_COUNT                0x0
#define TIMER1_LOAD                 0x4
#define TIMER1_MATCH1               0x8
#define TIMER1_MATCH2               0xC

#define TIMER2_COUNT                0x10
#define TIMER2_LOAD                 0x14
#define TIMER2_MATCH1               0x18
#define TIMER2_MATCH2               0x1C

#define TIMER3_COUNT                0x20
#define TIMER3_LOAD                 0x24
#define TIMER3_MATCH1               0x28
#define TIMER3_MATCH2               0x2C

#define TIMER_TMCR                  0x30
#define TIMER_INTRSTATE             0x34
#define TIMER_INTRMASK              0x38

/* Each timer's register address is offset by 0x10 */
#define TIMER_OFFSET                0x10

/*
 * Definition of TMCR bits
 */

#define TM1ENABLE                   1
#define TM1CLOCK                    (1<<1)
#define TM1OFENABLE                 (1<<2)

#define TM2ENABLE                   (1<<3)
#define TM2CLOCK                    (1<<4)
#define TM2OFENABLE                 (1<<5)

#define TM3ENABLE                   (1<<6)
#define TM3CLOCK                    (1<<7)
#define TM3OFENABLE                 (1<<8)

#define TM1UPDOWN                   (1<<9)
#define TM2UPDOWN                   (1<<10)
#define TM3UPDOWN                   (1<<11)


#define TM1MATCH1		(1 << 0)
#define TM1MATCH2		(1 << 1)
#define TM1OVERFLOW		(1 << 2)
#define TM2MATCH1		(1 << 3)
#define TM2MATCH2		(1 << 4)
#define TM2OVERFLOW		(1 << 5)
#define TM3MATCH1		(1 << 6)
#define TM3MATCH2		(1 << 7)
#define TM3OVERFLOW		(1 << 8)


static inline cycle_t clocksource_read_cycles(struct clocksource *cs)
{
	return (cycle_t)tmr_readl(TIMER3_COUNT);
}

static void clksrc_atftmr0101_resume(struct clocksource *cs)
{
	tmr_writel(tmr_readl(TIMER_INTRMASK) | TM3MATCH1 | TM3MATCH2 | TM3OVERFLOW, TIMER_INTRMASK);
	tmr_writel(tmr_readl(TIMER_TMCR) | TM3UPDOWN | TM3ENABLE, TIMER_TMCR);
}

static struct clocksource clksrc_atftmr0101 = {
	.name	= "atftmr0101_tm1",
	.rating	= 300,
	.read	= clocksource_read_cycles,
	.mask	= CLOCKSOURCE_MASK(32),
	.shift	= 21,
	.flags	= CLOCK_SOURCE_IS_CONTINUOUS,
	.resume = clksrc_atftmr0101_resume,
};

static void __init atftmr0101_clocksource_init(void)
{
	clksrc_atftmr0101.mult =
		clocksource_hz2mult(freq, clksrc_atftmr0101.shift);

	tmr_writel(0, TIMER3_LOAD);
	tmr_writel(tmr_readl(TIMER_INTRMASK) | TM3MATCH1 | TM3MATCH2 | TM3OVERFLOW, TIMER_INTRMASK);
	tmr_writel(tmr_readl(TIMER_TMCR) | TM3UPDOWN | TM3ENABLE, TIMER_TMCR);
	if (clocksource_register(&clksrc_atftmr0101))
		printk(KERN_ERR "Error: failed to register %s\n", clksrc_atftmr0101.name);
}

static int atftmr0101_set_next_event(unsigned long cycles,
		struct clock_event_device *evt)
{
	tmr_writel(cycles, TIMER1_LOAD);
	return 0;
}

static void atftmr0101_set_mode(enum clock_event_mode mode,
		struct clock_event_device *evt)
{
	switch (mode) {
		case CLOCK_EVT_MODE_ONESHOT:
			tmr_writel(0xffffffff, TIMER1_LOAD);
			tmr_writel(tmr_readl(TIMER_TMCR) | TM1ENABLE, TIMER_TMCR);
			break;

		case CLOCK_EVT_MODE_PERIODIC:
			tmr_writel(freq / HZ - 1, TIMER1_COUNT);
			tmr_writel(freq / HZ - 1, TIMER1_LOAD);
			tmr_writel(tmr_readl(TIMER_TMCR) | TM1ENABLE, TIMER_TMCR);
			break;
		case CLOCK_EVT_MODE_UNUSED:
			break;
		case CLOCK_EVT_MODE_SHUTDOWN:
			tmr_writel(tmr_readl(TIMER_TMCR) & ~TM1ENABLE, TIMER_TMCR);
			break;
		case CLOCK_EVT_MODE_RESUME:
			tmr_writel(tmr_readl(TIMER_INTRMASK) | TM1MATCH1 | TM1MATCH2, TIMER_INTRMASK);
			tmr_writel(tmr_readl(TIMER_TMCR) | TM1ENABLE | TM1OFENABLE, TIMER_TMCR);
			break;
	}
}

static struct clock_event_device clockevent_atftmr0101 = {
	.name		= "atftmr0101_tm1",
	.features       = CLOCK_EVT_FEAT_ONESHOT | CLOCK_EVT_FEAT_PERIODIC,
	.shift		= 32,
	.cpumask	= cpu_all_mask,
	.set_next_event	= atftmr0101_set_next_event,
	.set_mode	= atftmr0101_set_mode,
};

static irqreturn_t timer1_interrupt(int irq, void *dev_id)
{
	struct clock_event_device *evt = dev_id;

	tmr_writel(tmr_readl(TIMER_INTRSTATE) | TM1MATCH1 | TM1MATCH2 | TM1OVERFLOW, TIMER_INTRSTATE);

	evt->event_handler(evt);

	return IRQ_HANDLED;
}

static struct irqaction timer1_irq = {
	.name		= "Timer Tick",
	.flags		= IRQF_DISABLED | IRQF_TIMER | IRQF_IRQPOLL,
	.handler	= timer1_interrupt,
	.dev_id		= &clockevent_atftmr0101
};

static void __init atftmr0101_clockevent_init(int irq)
{
	clockevent_atftmr0101.mult = div_sc(freq, NSEC_PER_SEC, clockevent_atftmr0101.shift);
	clockevent_atftmr0101.max_delta_ns = clockevent_delta2ns(0xffffffff, &clockevent_atftmr0101);
	clockevent_atftmr0101.min_delta_ns = clockevent_delta2ns(3, &clockevent_atftmr0101);

	clockevents_register_device(&clockevent_atftmr0101);
	setup_irq(irq, &timer1_irq);
}

int clksrc_init;
static void __init atftmr010_init(struct device_node *dev)
{
	int irq;

	base = of_iomap(dev, 0);
	if (!base)
		panic("Can't remap registers");

	if (of_property_read_u32(dev, "clock-frequency", &freq))
		panic("Can't read clock-frequency");

	irq = irq_of_parse_and_map(dev, 0);
	if (irq <= 0)
		panic("Failed to map timer IRQ\n");

	printk("ATFTMR010 timer 1 installed on IRQ %d, with clock %d at %d HZ. in 0x%08x\r\n",
			irq, freq, HZ, (u32)base);

	tmr_writel(tmr_readl(TIMER_TMCR) & ~(TM1ENABLE | TM1CLOCK | TM1OFENABLE | TM1UPDOWN), TIMER_TMCR);
	tmr_writel(tmr_readl(TIMER_INTRMASK) | TM1MATCH1 | TM1MATCH2, TIMER_INTRMASK);
	tmr_writel(tmr_readl(TIMER_TMCR) | TM1OFENABLE, TIMER_TMCR);

	atftmr0101_clocksource_init();
	atftmr0101_clockevent_init(irq);
	clksrc_init = 1;
}

unsigned long long notrace sched_clock(void)
{
	if (clksrc_init)
		return clocksource_cyc2ns(clocksource_read_cycles(&clksrc_atftmr0101),
			clksrc_atftmr0101.mult, clksrc_atftmr0101.shift);
	else
		return (unsigned long long)(jiffies - INITIAL_JIFFIES)
	                                        * (NSEC_PER_SEC / HZ);
}

CLOCKSOURCE_OF_DECLARE(ag101p_atftmr010, "andestech,atftmr010", atftmr010_init);
