#include <linux/module.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/gpio.h>
#include <linux/of_platform.h>
#include <linux/of_gpio.h>
#include <linux/err.h>

#define GPIO_DATA_OUT			0x24
#define GPIO_DATA_IN			0x20
#define PIN_DIR				0x28
#define PIN_PULL_ENABLE			0x40
#define PIN_PULL_TYPE			0x44
#define INT_ENABLE			0x50
#define INT_STATE			0x64
#define INT_MODE(ch)			(0x54+((ch/8)<<2))
#define HIGH_LEVEL			2
#define LOW_LEVEL			3
#define NEGATIVE_EDGE			5
#define POSITIVE_EDGE			6
#define DUAL_EDGE			7
#define DEBOUNCE_ENABLE			0x70

static resource_size_t atcgpio100_va_base;

#define GPIO_READL(offset)		\
	__raw_readl(atcgpio100_va_base + (offset))

#define GPIO_WRITEL(val, offset)	\
	__raw_writel((val), atcgpio100_va_base + (offset))


#define ATCGPIO100_VIRTUAL_IRQ_BASE	32

static int atcirq_to_gpio(unsigned int irq)
{
	return irq - ATCGPIO100_VIRTUAL_IRQ_BASE;
}

static int atcgpio_to_irq(struct gpio_chip *gc, unsigned int offset)
{
	return ATCGPIO100_VIRTUAL_IRQ_BASE + offset;
}

static int atcgpio_get(struct gpio_chip *gc, unsigned int gpio)
{
	return (GPIO_READL(GPIO_DATA_IN) >> gpio & 1);
}

static void atcgpio_set(struct gpio_chip *gc, unsigned int gpio, int data)
{
	unsigned long val;

	if (data)
		val = GPIO_READL(GPIO_DATA_OUT) | (0x1UL << gpio);
	else
		val = GPIO_READL(GPIO_DATA_OUT) & ~(0x1UL << gpio);

	GPIO_WRITEL(val, GPIO_DATA_OUT);
}

static int atcgpio_dir_in(struct gpio_chip *gc, unsigned int gpio)
{
	unsigned long val;
	val = GPIO_READL(PIN_DIR) & ~(0x1UL << gpio);
	GPIO_WRITEL(val, PIN_DIR);
	return 0;
}

static int atcgpio_dir_out(struct gpio_chip *gc, unsigned int gpio, int data)
{
	unsigned long val;
	val = GPIO_READL(PIN_DIR) | (0x1UL << gpio);
	GPIO_WRITEL(val, PIN_DIR);
	gc->set(gc, gpio, data);
	return 0;
}

static struct gpio_chip atcgpio_chip = {

	.label			= "ATCGPIO100",
	.base			= 0,
	.ngpio			= 16,
	.direction_input	= atcgpio_dir_in,
	.direction_output	= atcgpio_dir_out,
	.get			= atcgpio_get,
	.set			= atcgpio_set,
	.to_irq			= atcgpio_to_irq,
};

static void atcgpio_irq_ack(struct irq_data *data)
{
	GPIO_WRITEL(0x1UL << atcirq_to_gpio(data->irq), INT_STATE);
}

static void atcgpio_irq_unmask(struct irq_data *data)
{
	unsigned long val;
	val = GPIO_READL(INT_ENABLE) | (0x1UL << atcirq_to_gpio(data->irq));
	GPIO_WRITEL(val, INT_ENABLE);
}

static void atcgpio_irq_mask(struct irq_data *data)
{
	unsigned long val;
	val = GPIO_READL(INT_ENABLE) & ~(0x1UL << atcirq_to_gpio(data->irq));
	GPIO_WRITEL(val, INT_ENABLE);
}

static int atcgpio_irq_set_type(struct irq_data *data, unsigned int flow_type)
{
	unsigned long gpio = atcirq_to_gpio(data->irq);
	unsigned long ch = gpio%8;
	unsigned long mode_off = ch<<2;
	unsigned long val;
	val = GPIO_READL(INT_MODE(gpio));
	val &= ~(7<<mode_off);
	if (flow_type & IRQF_TRIGGER_RISING && flow_type & IRQF_TRIGGER_FALLING)
		GPIO_WRITEL(val | DUAL_EDGE<<mode_off, INT_MODE(gpio));
	else if (flow_type & IRQF_TRIGGER_FALLING)
		GPIO_WRITEL(val | NEGATIVE_EDGE<<mode_off, INT_MODE(gpio));
	else if (flow_type & IRQF_TRIGGER_RISING)
		GPIO_WRITEL(val | POSITIVE_EDGE<<mode_off, INT_MODE(gpio));
	return 0;
}

static struct irq_chip atcgpio_irq_chip = {

	.name = "ATCGPIO100_irq",
	.irq_ack = atcgpio_irq_ack,
	.irq_mask = atcgpio_irq_mask,
	.irq_unmask = atcgpio_irq_unmask,
	.irq_set_type = atcgpio_irq_set_type,
};

static void gpio_irq_router(unsigned int irq, struct irq_desc *desc)
{
	unsigned long status;
	int i = 0;

	status = GPIO_READL(INT_STATE);
	status &= ~((1 << 22) | (1 << 25) | (1 << 26));

	while (status) {
		if (status & 0x1UL)
			generic_handle_irq(gpio_to_irq(i));
		status >>= 1;
		i++;
	}
}

static int atcgpio100_gpio_probe(struct platform_device *pdev)
{
	int i;

	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	struct resource *irq_res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);

	if (!res) {
		dev_err(&pdev->dev, "failed to request atcgpio100 resource\n");
		return -ENOENT;
	}

	atcgpio100_va_base = (resource_size_t)devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR((void *)atcgpio100_va_base))
		return PTR_ERR((void *)atcgpio100_va_base);

	/* disable interrupt */
	GPIO_WRITEL(0x00000000UL, INT_ENABLE);

	/* clear interrupt */
	GPIO_WRITEL(0x0000FFFFUL, INT_STATE);

	/* enable de-bouncing */
	GPIO_WRITEL(0x0000FFFFUL, DEBOUNCE_ENABLE);

	/* enable interrupt */
	GPIO_WRITEL(0x0000FFFFUL, INT_ENABLE);

	gpiochip_add(&atcgpio_chip);

	for (i = 0; i < atcgpio_chip.ngpio; i++) {

		irq_set_chip(gpio_to_irq(i), &atcgpio_irq_chip);
		irq_set_handler(gpio_to_irq(i), handle_level_irq);
	}
	irq_set_chained_handler(irq_res->start, gpio_irq_router);

	pr_info("ATCGPIO100 module inserted\n");

	return 0;
}

static int atcgpio100_gpio_remove(struct platform_device *pdev)
{
	struct resource *irq_res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);

	irq_set_handler(irq_res->start, NULL);

	/* disable interrupt */
	GPIO_WRITEL(0x00000000UL, INT_ENABLE);
	if (gpiochip_remove(&atcgpio_chip))
		pr_info("failed to remove gpiochip\n");

	pr_info("GPIO module removed\n");
	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id atcgpio100_gpio_match[] = {
	{
		.compatible = "andestech,atcgpio100",
	},
	{ },
};
#endif

static struct platform_driver atcgpio100_gpio_driver = {
	.probe		= atcgpio100_gpio_probe,
	.driver		= {
		.name	= "atcgpio100_gpio",
		.of_match_table = of_match_ptr(atcgpio100_gpio_match),
	},
	.remove		= atcgpio100_gpio_remove,
};

module_platform_driver(atcgpio100_gpio_driver);
MODULE_DESCRIPTION("ATCGPIO100");
MODULE_LICENSE("GPL");
