#include <linux/module.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/gpio.h>
#include <linux/of_platform.h>
#include <linux/of_gpio.h>
#include <linux/err.h>

#define GPIO_DATA_OUT			0x00
#define GPIO_DATA_IN			0x04
#define PIN_DIR				0x08
#define PIN_BYPASS			0x0C
#define GPIO_DATA_SET			0x10
#define GPIO_DATA_CLEAR			0x14
#define PIN_PULL_ENABLE			0x18
#define PIN_PULL_TYPE			0x1C
#define INT_ENABLE			0x20
#define INT_RAW_STATE			0x24
#define INT_MASKED_STATE		0x28
#define INT_MASK			0x2C
#define INT_CLEAR			0x30
#define INT_TRIGGER			0x34
#define INT_BOTH			0x38
#define INT_RISE_NEG			0x3C
#define BOUNCE_ENABLE			0x40
#define BOUNCE_PRE_SCALE		0x44
static resource_size_t atfgpio010_va_base;

#define GPIO_READL(offset)		\
	__raw_readl(atfgpio010_va_base + (offset))

#define GPIO_WRITEL(val, offset)	\
	__raw_writel((val), atfgpio010_va_base + (offset))

#define ATFGPIO010_VIRTUAL_IRQ_BASE	32

static int atfirq_to_gpio(unsigned int irq)
{
	return irq - ATFGPIO010_VIRTUAL_IRQ_BASE;
}

static int atfgpio_to_irq(struct gpio_chip *gc, unsigned int offset)
{
	return ATFGPIO010_VIRTUAL_IRQ_BASE + offset;
}

static int atfgpio_get(struct gpio_chip *gc, unsigned int gpio)
{
	return (GPIO_READL(GPIO_DATA_IN) >> gpio & 1);
}

static void atfgpio_set(struct gpio_chip *gc, unsigned int gpio, int data)
{
	unsigned long val;

	if (data)
		val = GPIO_READL(GPIO_DATA_OUT) | (0x1UL << gpio);
	else
		val = GPIO_READL(GPIO_DATA_OUT) & ~(0x1UL << gpio);

	GPIO_WRITEL(val, GPIO_DATA_OUT);
}

static int atfgpio_dir_in(struct gpio_chip *gc, unsigned int gpio)
{
	unsigned long val;
	val = GPIO_READL(PIN_DIR) & ~(0x1UL << gpio);
	GPIO_WRITEL(val, PIN_DIR);
	return 0;
}

static int atfgpio_dir_out(struct gpio_chip *gc, unsigned int gpio, int data)
{
	unsigned long val;
	val = GPIO_READL(PIN_DIR) | (0x1UL << gpio);
	GPIO_WRITEL(val, PIN_DIR);
	gc->set(gc, gpio, data);
	return 0;
}

static struct gpio_chip atfgpio_chip = {

	.label			= "ATFGPIO010",
	.base			= 0,
	.ngpio			= 16,
	.direction_input	= atfgpio_dir_in,
	.direction_output	= atfgpio_dir_out,
	.get			= atfgpio_get,
	.set			= atfgpio_set,
	.to_irq			= atfgpio_to_irq,
};

static void atfgpio_irq_ack(struct irq_data *data)
{
	GPIO_WRITEL(0x1UL << atfirq_to_gpio(data->irq), INT_CLEAR);
}

static void atfgpio_irq_mask(struct irq_data *data)
{
	unsigned long val;
	val = GPIO_READL(INT_MASK) | (0x1UL << atfirq_to_gpio(data->irq));
	GPIO_WRITEL(val, INT_MASK);

}

static void atfgpio_irq_unmask(struct irq_data *data)
{
	unsigned long val;
	val = GPIO_READL(INT_MASK) & ~(0x1UL << atfirq_to_gpio(data->irq));
	GPIO_WRITEL(val, INT_MASK);
}

static int atfgpio_irq_set_type(struct irq_data *data, unsigned int flow_type)
{
	unsigned long bit = 0x1UL << atfirq_to_gpio(data->irq);
	unsigned long val;
	val = GPIO_READL(INT_BOTH);
	if (flow_type & IRQF_TRIGGER_RISING && flow_type & IRQF_TRIGGER_FALLING)
		GPIO_WRITEL(val | bit, INT_BOTH);
	else
		GPIO_WRITEL(val & ~bit, INT_BOTH);

	val = GPIO_READL(INT_RISE_NEG);

	if (flow_type & IRQF_TRIGGER_FALLING)
		GPIO_WRITEL(val | bit, INT_RISE_NEG);
	else if (flow_type & IRQF_TRIGGER_RISING)
		GPIO_WRITEL(val & ~bit, INT_RISE_NEG);
	return 0;
}

static struct irq_chip atfgpio_irq_chip = {

	.name = "ATFGPIO010_irq",
	.irq_ack = atfgpio_irq_ack,
	.irq_mask = atfgpio_irq_mask,
	.irq_unmask = atfgpio_irq_unmask,
	.irq_set_type = atfgpio_irq_set_type,
};

static void gpio_irq_router(unsigned int irq, struct irq_desc *desc)
{
	unsigned long status;
	int i = 0;

	status = GPIO_READL(INT_RAW_STATE);

	status &= ~((1 << 22) | (1 << 25) | (1 << 26));

	while (status) {
		if (status & 0x1UL)
			generic_handle_irq(gpio_to_irq(i));
		status >>= 1;
		i++;
	}
}
static int atfgpio010_gpio_probe(struct platform_device *pdev)
{
	int i;
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	struct resource *irq_res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);

	if (!res) {
		dev_err(&pdev->dev, "failed to request atfgpio010 resource\n");
		return -ENOENT;
	}
	atfgpio010_va_base = (resource_size_t)devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR((void *)atfgpio010_va_base))
		return PTR_ERR((void *)atfgpio010_va_base);

	/* disable interrupt */
	GPIO_WRITEL(0x00000000UL, INT_ENABLE);

	/* mask interrupt */
	GPIO_WRITEL(0x0000FFFFUL, INT_MASK);

	/* triggered interrupt on both edge */
	GPIO_WRITEL(0x0000FFFFUL, INT_BOTH);

	/* clear interrupt */
	GPIO_WRITEL(0x0000FFFFUL, INT_CLEAR);

	/* enable de-bouncing */
	GPIO_WRITEL(0x0000FFFFUL, BOUNCE_ENABLE);

	/* enable interrupt */
	GPIO_WRITEL(0x0000FFFFUL, INT_ENABLE);

	gpiochip_add(&atfgpio_chip);

	for (i = 0; i < atfgpio_chip.ngpio; i++) {

		irq_set_chip(gpio_to_irq(i), &atfgpio_irq_chip);
		irq_set_handler(gpio_to_irq(i), handle_level_irq);
	}

	irq_set_chained_handler(irq_res->start, gpio_irq_router);

	pr_info("GPIO module inserted\n");

	return 0;
}

static int atfgpio010_gpio_remove(struct platform_device *pdev)
{
	struct resource *irq_res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);

	irq_set_handler(irq_res->start, NULL);

	/* disable interrupt */
	GPIO_WRITEL(0x00000000UL, INT_ENABLE);

	if (gpiochip_remove(&atfgpio_chip))
		pr_info("failed to remove gpiochip\n");

	pr_info("GPIO module removed\n");
	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id atfgpio010_gpio_match[] = {
	{
		.compatible = "andestech,atfgpio010",
	},
	{ },
};
#endif

static struct platform_driver atfgpio010_gpio_driver = {
	.probe		= atfgpio010_gpio_probe,
	.driver		= {
		.name	= "atfgpio010_gpio",
		.of_match_table = of_match_ptr(atfgpio010_gpio_match),
	},
	.remove		= atfgpio010_gpio_remove,
};

module_platform_driver(atfgpio010_gpio_driver);
MODULE_DESCRIPTION("ATFGPIO010");
MODULE_LICENSE("GPL");
