/*
 *  Faraday FTINTC010 Master Interrupt Controller Device Driver Implementation
 *
 *  Copyright (C) 2005 Faraday Corp. (http://www.faraday-tech.com)
 *  Copyright (C) 2016 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Note
 *
 *  This program implements only the master INTC of the platform. Slave INTCs must
 *  be initialized by themselves.
 *
 * ChangeLog
 *
 *  Luke Lee  09/14/2005  Created.
 */
#include <linux/irq.h>
#include <linux/of.h>
#include <linux/of_irq.h>
#include <linux/of_address.h>
#include <linux/irqdomain.h>
#include <linux/interrupt.h>
#include "irqchip.h"
#include <asm/io.h>
#include <asm/atfintc010.h>

static resource_size_t atfintc010_va_base;

/*
 * Edge trigger IRQ chip methods
 */
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
spinlock_t irq_chip_lock;
#endif

static void atfintc010_ack_irq(struct irq_data *data)
{
	unsigned int tmp;
	// ack and disable
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_irq(&irq_chip_lock);
#endif
	*(volatile unsigned *)(atfintc010_va_base + IRQ_CLEAR_REG) = 1 << data->hwirq;
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_CLEAR_REG);
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MODE_REG);

	if (!(tmp & (1UL << data->hwirq))) {	/* level trigger */

		*(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG) &=
		    ~(1 << data->hwirq);
		tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG);
	}
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_unlock_irq(&irq_chip_lock);
#endif
}

static void atfintc010_mask_irq(struct irq_data *data)
{
	unsigned int tmp;
	// disable
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_irq(&irq_chip_lock);
#endif
	*(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG) &=
	    ~(1 << data->hwirq);
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG);
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_unlock_irq(&irq_chip_lock);
#endif
}

static void atfintc010_mask_ack_irq(struct irq_data *data)
{
	unsigned int tmp;
	// disable
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_irq(&irq_chip_lock);
#endif
	*(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG) &=
	    ~(1 << data->hwirq);
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG);

	*(volatile unsigned *)(atfintc010_va_base + IRQ_CLEAR_REG) = 1 << data->hwirq;
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_CLEAR_REG);

#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_unlock_irq(&irq_chip_lock);
#endif
}

static void atfintc010_unmask_irq(struct irq_data *data)
{
	unsigned int tmp;
	// enable
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_irq(&irq_chip_lock);
#endif
	*(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG) |= 1 << data->hwirq;
	tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MASK_REG);
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_unlock_irq(&irq_chip_lock);
#endif
}

static int atfintc010_set_type(struct irq_data *data,
				   unsigned int flow_type)
{
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_irq(&irq_chip_lock);
#endif
	/*
	 * IRQ Trigger Mode Register 1: edge
	 * IRQ Trigger Level Register 1: active high
	 */

	int tmp;

	if (flow_type & (IRQF_TRIGGER_FALLING | IRQF_TRIGGER_LOW)) {

		*(volatile unsigned *)(atfintc010_va_base + IRQ_LEVEL_REG) &=
		    ~(1UL << data->hwirq);
		tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_LEVEL_REG);
	}

	if (flow_type & (IRQF_TRIGGER_RISING | IRQF_TRIGGER_HIGH)) {

		*(volatile unsigned *)(atfintc010_va_base + IRQ_LEVEL_REG) |=
		    (1UL << data->hwirq);
		tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_LEVEL_REG);
	}

	if (flow_type & (IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING)) {

		*(volatile unsigned *)(atfintc010_va_base + IRQ_MODE_REG) |=
		    (1UL << data->hwirq);
		tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MODE_REG);
		(irq_desc + data->irq)->handle_irq = handle_edge_irq;
	}

	if (flow_type & (IRQF_TRIGGER_HIGH | IRQF_TRIGGER_LOW)) {

		*(volatile unsigned *)(atfintc010_va_base + IRQ_MODE_REG) &=
		    ~(1UL << data->hwirq);
		tmp = *(volatile unsigned *)(atfintc010_va_base + IRQ_MODE_REG);
		(irq_desc + data->irq)->handle_irq = handle_level_irq;
	}
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_unlock_irq(&irq_chip_lock);
#endif
	return 0;
}

static struct irq_chip atfintc010_chip = {

	.irq_ack = atfintc010_ack_irq,
	.irq_mask = atfintc010_mask_irq,
	.irq_mask_ack = atfintc010_mask_ack_irq,
	.irq_unmask = atfintc010_unmask_irq,
	.irq_set_type = atfintc010_set_type,
};


struct irq_domain *root_domain;
static int atfintc010_irq_domain_map(struct irq_domain *id, unsigned int virq,
				  irq_hw_number_t hw)
{
	if (PLATFORM_IRQ_TRIGGER_MODE & (1 << hw))
		irq_set_chip_and_handler(virq, &atfintc010_chip, handle_edge_irq);
	else
		irq_set_chip_and_handler(virq, &atfintc010_chip, handle_level_irq);

	return 0;
}

static struct irq_domain_ops atfintc010_ops = {
	.map = atfintc010_irq_domain_map,
	.xlate = irq_domain_xlate_onecell
};

int __init atfintc010_init_irq(struct device_node *node, struct device_node *parent)
{
	int i;
#if defined(CONFIG_SMP) && defined(CONFIG_IRQ_PER_CPU)
	spin_lock_init(&irq_chip_lock);
#endif

	atfintc010_va_base = (resource_size_t)of_iomap(node, 0);
	/* Initialize the INTC */
	outl(0x00000000, atfintc010_va_base + IRQ_MASK_REG);
	outl(0x00000000, atfintc010_va_base + FIQ_MASK_REG);
	outl(0xffffffff, atfintc010_va_base + IRQ_CLEAR_REG);
	outl(0xffffffff, atfintc010_va_base + FIQ_CLEAR_REG);
	outl(PLATFORM_IRQ_TRIGGER_MODE, atfintc010_va_base + IRQ_MODE_REG);
	/* FTINTC010: bit 0=active high or rising edge, 1=active low or falling edge. */
	outl(PLATFORM_IRQ_TRIGGER_LEVEL, atfintc010_va_base + IRQ_LEVEL_REG);



	root_domain = irq_domain_add_linear(node, PLATFORM_IRQ_TOTALCOUNT,
		&atfintc010_ops, NULL);

	if (!root_domain)
		panic("%s: unable to create IRQ domain\n", node->full_name);

	for(i = 0; i < PLATFORM_IRQ_TOTALCOUNT; i++)
		irq_create_mapping(root_domain, i);

	return 0;

}

unsigned int get_IntSrc(void)
{
	unsigned int irq_status, irq = 31;

	irq_status = inl(atfintc010_va_base + IRQ_STATUS_REG);
	if (irq_status == 0)
		return 32;
	if (irq_status & 0x0000ffff) {
		irq -= 16;
		irq_status <<= 16;
	}
	if (irq_status & 0x00ff0000) {
		irq -= 8;
		irq_status <<= 8;
	}
	if (irq_status & 0x0f000000) {
		irq -= 4;
		irq_status <<= 4;
	}
	if (irq_status & 0x30000000) {
		irq -= 2;
		irq_status <<= 2;
	}
	if (irq_status & 0x40000000) {
		irq -= 1;
	}

	return irq;
}
IRQCHIP_DECLARE(atfintc010, "andestech,atfintc010", atfintc010_init_irq);
