/*
 *  Andes Interrupt Controller Device Driver Implementation
 *
 *  Copyright (C) 2016 Andes Technology Corporation
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/irq.h>
#include <linux/of.h>
#include <linux/of_irq.h>
#include <linux/of_address.h>
#include <linux/irqdomain.h>
#include "irqchip.h"

static void nointc_ack_irq(struct irq_data *data)
{
	SET_INT_PEND2(1 << data->hwirq);
}

static void nointc_mask_irq(struct irq_data *data)
{
	unsigned long int_mask2 = GET_INT_MASK2();
	SET_INT_MASK2(int_mask2 & (~(1 << data->hwirq)));
}

static void nointc_mask_ack_irq(struct irq_data *data)
{
	unsigned long int_mask2 = GET_INT_MASK2();
	SET_INT_MASK2(int_mask2 & (~(1 << data->hwirq)));
	SET_INT_PEND2(1 << data->hwirq);

}

static void nointc_unmask_irq(struct irq_data *data)
{
	unsigned long int_mask2 = GET_INT_MASK2();
	SET_INT_MASK2(int_mask2 | (1 << data->hwirq));
}

static int nointc_set_type(struct irq_data *data, unsigned int flow_type)
{
	printk(KERN_WARNING "interrupt type is not configurable\n");
	return 0;
}

static struct irq_chip nointc_chip = {
	.name = "nointc",
	.irq_ack = nointc_ack_irq,
	.irq_mask = nointc_mask_irq,
	.irq_mask_ack = nointc_mask_ack_irq,
	.irq_unmask = nointc_unmask_irq,
	.irq_set_type = nointc_set_type,
};

static unsigned int __initdata nivic_map[6] = { 6, 2, 10, 16, 24, 32 };

struct irq_domain *root_domain;
static int nointc_irq_domain_map(struct irq_domain *id, unsigned int virq,
				  irq_hw_number_t hw)
{

	unsigned long int_trigger_type;
	int_trigger_type = GET_INT_TRIGGER();
	if (int_trigger_type & (1 << hw))
		irq_set_chip_and_handler(virq, &nointc_chip, handle_edge_irq);
	else
		irq_set_chip_and_handler(virq, &nointc_chip, handle_level_irq);

	return 0;
}

static struct irq_domain_ops nointc_ops = {
	.map = nointc_irq_domain_map,
	.xlate = irq_domain_xlate_onecell
};

int __init nointc_init_irq(struct device_node *node, struct device_node *parent)
{
	unsigned long int_vec_base, nivic, i;

	if (WARN(parent, "non-root atnointc are not supported"))
		return -EINVAL;

	int_vec_base = GET_IVB();

	if (((int_vec_base & IVB_mskIVIC_VER) >> IVB_offIVIC_VER) == 0) {
		panic("Unable to use NOINTC option to boot on this cpu\n");
	}

	nivic = (int_vec_base & IVB_mskNIVIC) >> IVB_offNIVIC;
	if (nivic >= (sizeof nivic_map / sizeof nivic_map[0])) {
		panic
		    ("The number of input for IVIC Controller is not supported on this cpu\n");
	}
	nivic = nivic_map[nivic];

	root_domain = irq_domain_add_linear(node, nivic,
			&nointc_ops, NULL);

	if (!root_domain)
		panic("%s: unable to create IRQ domain\n", node->full_name);

	for(i = 0; i < nivic; i++)
		irq_create_mapping(root_domain, i);

	return 0;

}
IRQCHIP_DECLARE(nointc, "andestech,nointc", nointc_init_irq);
