/*
 * Faraday RTC Support
 *
 * Copyright (C) 2006, 2007, 2008  Paul Mundt
 * Copyright (C) 2006  Jamie Lenehan
 * Copyright (C) 2008  Angelo Castello
 * Copyright (C) 2008  Roy Lee
 *
 * Based on the old arch/sh/kernel/cpu/rtc.c by:
 *
 *  Copyright (C) 2000  Philipp Rumpf <prumpf@tux.org>
 *  Copyright (C) 1999  Tetsuya Okada & Niibe Yutaka
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/rtc.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/seq_file.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/of.h>

#define DRV_NAME		"atfrtc010"

#define RTC_REG( off)		( *( volatile unsigned long *)( rtc->regbase + ( off)))

#define RTC_CR			RTC_REG( 0x20)	/* Control */
#define RTC_CR_IE		( 0x1UL << 0)
#define RTC_CR_IES		( 0x1UL << 1)
#define RTC_CR_IEM		( 0x1UL << 2)
#define RTC_CR_IEH		( 0x1UL << 3)
#define RTC_CR_IED		( 0x1UL << 4)
#define RTC_CR_ALRM		( 0x1UL << 5)
#define RTC_CR_LOAD		( 0x1UL << 6)

#define RTC_RR			RTC_REG( 0x1C)	/* RTC Record Register */
#define RTC_DIV			RTC_REG( 0x38)	/* RTC Divede Register */
#define RTC_REV			RTC_REG( 0x3C)	/* RTC Divede Register */

#define RTC_IR			RTC_REG( 0x34)	/* RTC interrupt state */
#define RTC_IR_IES		( 0x1UL << 0)	/* RTC interrupt state */
#define RTC_IR_IEM		( 0x1UL << 1)	/* RTC interrupt state */
#define RTC_IR_IEH		( 0x1UL << 2)	/* RTC interrupt state */
#define RTC_IR_IED		( 0x1UL << 3)	/* RTC interrupt state */
#define RTC_IR_ALRM		( 0x1UL << 4)	/* RTC interrupt state */

#define RTC_SECOND		RTC_REG( 0x00)	/* RTC sec */
#define RTC_MINUTE		RTC_REG( 0x04)	/* RTC min */
#define RTC_HOUR		RTC_REG( 0x08)	/* RTC hour */
#define RTC_DAYS		RTC_REG( 0x0C)	/* RTC day */

#define RTC_ALRM_SECOND		RTC_REG( 0x10)	/* RTC alarm sec */
#define RTC_ALRM_MINUTE		RTC_REG( 0x14)	/* RTC alarm min */
#define RTC_ALRM_HOUR		RTC_REG( 0x18)	/* RTC alarm hour */

#define RTC_WRITE_SECOND	RTC_REG( 0x24)	/* RTC write port for sec */
#define RTC_WRITE_MINUTE	RTC_REG( 0x28)	/* RTC write port for min */
#define RTC_WRITE_HOUR		RTC_REG( 0x2C)	/* RTC write port for hour */
#define RTC_WRITE_DAYS		RTC_REG( 0x30)	/* RTC write port for day */

struct atf_rtc{

	void __iomem *regbase;
	struct resource *res;
	unsigned int alarm_irq;
	unsigned int interrupt_irq;
	struct rtc_device *rtc_dev;
	spinlock_t		lock;		/* Protects this structure */
};

struct atf_rtc rtc_platform_data;

static irqreturn_t atf_rtc_interrupt( int irq, void *dev_id){

	struct atf_rtc *rtc = dev_id;
	if( RTC_IR & RTC_IR_IES){

		RTC_IR &= ~RTC_IR_IES;
		rtc_update_irq( rtc->rtc_dev, 1, RTC_UF | RTC_IRQF);

		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}

static irqreturn_t atf_rtc_alarm( int irq, void *dev_id){

	struct atf_rtc *rtc = dev_id;
	if( RTC_IR & RTC_IR_ALRM){

		RTC_CR &= ~RTC_CR_ALRM;
		RTC_IR &= ~RTC_IR_ALRM;
		rtc_update_irq( rtc->rtc_dev, 1, RTC_AF | RTC_IRQF);

		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}

static void atf_rtc_release( struct device *dev){

	struct atf_rtc *rtc = dev_get_drvdata( dev);

	RTC_CR &= ~RTC_CR_IES;
	RTC_CR &= ~RTC_CR_ALRM;
}

static int atf_alarm_irq_enable(struct device *dev, unsigned int enabled)
{

	struct atf_rtc *rtc = dev_get_drvdata( dev);
	spin_lock_irq(&rtc->lock);
	if (enabled)
		RTC_CR |= RTC_CR_ALRM;
	else
		RTC_CR &= ~RTC_CR_ALRM;
	spin_unlock_irq(&rtc->lock);
	return 0;
}

static int atf_rtc_read_time( struct device *dev, struct rtc_time *tm){
	struct atf_rtc *rtc = dev_get_drvdata( dev);
	unsigned long time = RTC_DAYS * 86400 + RTC_HOUR * 3600 + RTC_MINUTE * 60 + RTC_SECOND;
	rtc_time_to_tm( time, tm);
	if( rtc_valid_tm( tm) < 0) {
		dev_err( dev, "invalid date\n");
		rtc_time_to_tm( 0, tm);
	}
	return 0;
}

static int atf_rtc_set_time( struct device *dev, struct rtc_time *tm){

	struct atf_rtc *rtc = dev_get_drvdata( dev);
	unsigned long time = 0;

	rtc_tm_to_time( tm, &time);

	RTC_WRITE_DAYS = time / 86400;
	time %= 86400;

	RTC_WRITE_HOUR = time / 3600;
	time %= 3600;

	RTC_WRITE_MINUTE = time / 60;
	time %= 60;

	RTC_WRITE_SECOND = time;

	RTC_CR |= RTC_CR_LOAD;

	return 0;
}

static int atf_rtc_read_alarm( struct device *dev, struct rtc_wkalrm *wkalrm){

	struct atf_rtc *rtc = dev_get_drvdata( dev);
	struct rtc_time *tm = &wkalrm->time;

	tm->tm_sec	= RTC_ALRM_SECOND;
	tm->tm_min	= RTC_ALRM_MINUTE;
	tm->tm_hour	= RTC_ALRM_HOUR;

	wkalrm->enabled = ( RTC_CR & RTC_CR_ALRM) ? 1 : 0;

	return 0;
}

static int atf_rtc_set_alarm( struct device *dev, struct rtc_wkalrm *wkalrm){

	struct atf_rtc *rtc = dev_get_drvdata( dev);
	struct rtc_time *tm = &wkalrm->time;
	int err = rtc_valid_tm( tm);
	if( err < 0){

		dev_err( dev, "invalid alarm value\n");
		return err;
	}

	/* disable alarm interrupt and clear the alarm flag */
	RTC_CR &= ~RTC_CR_ALRM;

	/* set alarm time */
	RTC_ALRM_SECOND = tm->tm_sec;
	RTC_ALRM_MINUTE = tm->tm_min;
	RTC_ALRM_HOUR = tm->tm_hour;

	if( wkalrm->enabled)
		RTC_CR |= RTC_CR_ALRM;

	return 0;
}

static struct rtc_class_ops atf_rtc_ops = {

	.release	= atf_rtc_release,
	.alarm_irq_enable = atf_alarm_irq_enable,

	.read_time	= atf_rtc_read_time,
	.set_time	= atf_rtc_set_time,
	.read_alarm	= atf_rtc_read_alarm,
	.set_alarm	= atf_rtc_set_alarm,
};

static int atf_rtc_probe( struct platform_device *pdev){

	struct atf_rtc *rtc = &rtc_platform_data;
	int ret = -ENOENT;
	spin_lock_init(&rtc->lock);

	if( ( rtc->alarm_irq = platform_get_irq( pdev, 0)) < 0)
		goto err_exit;

	if( ( rtc->interrupt_irq = platform_get_irq( pdev, 1)) < 0)
		goto err_exit;

	if( !( rtc->res = platform_get_resource( pdev, IORESOURCE_MEM, 0)))
		goto err_exit;

	if( !( rtc->res = request_mem_region( rtc->res->start, rtc->res->end - rtc->res->start + 1, pdev->name)))
		goto err_request_region;

	ret = -EINVAL;

	if( !( rtc->regbase = ioremap_nocache( rtc->res->start, rtc->res->end - rtc->res->start + 1)))
		goto err_ioremap1;

	if( ( ret = request_irq( rtc->alarm_irq , atf_rtc_alarm, 0, "RTC Alarm     : atfrtc010", rtc)))
		goto err_exit;

	if( ( ret = request_irq( rtc->interrupt_irq , atf_rtc_interrupt, 0, "RTC Interrupt : atfrtc010", rtc)))
		goto err_interrupt_irq;

	ret = -EBUSY;


	//RTC_DIV = ( 0x1UL << 31) | 2;	/* AG101 */
	RTC_CR |= RTC_CR_IE;
	platform_set_drvdata( pdev, rtc);
	device_init_wakeup(&pdev->dev, true);

	rtc->rtc_dev = rtc_device_register( DRV_NAME, &pdev->dev, &atf_rtc_ops, THIS_MODULE);

	if( IS_ERR( rtc->rtc_dev)) {

		ret = PTR_ERR( rtc->rtc_dev);
		goto err_unmap;
	}

	rtc->rtc_dev->max_user_freq = 256;
	rtc->rtc_dev->irq_freq = 1;
	return 0;

err_unmap:
	iounmap( rtc->regbase);
err_ioremap1:
	release_resource( rtc->res);
err_request_region:
	free_irq( rtc->interrupt_irq, rtc);
err_interrupt_irq:
	free_irq( rtc->alarm_irq, rtc);
err_exit:
	return ret;
}

static int atf_rtc_remove( struct platform_device *pdev){

	struct atf_rtc *rtc = platform_get_drvdata( pdev);

	rtc_device_unregister( rtc->rtc_dev);

	/*
	 Because generic rtc will not execute rtc_device_release() when call rtc_device_unregister(),
	 rtc id will increase when unloading a rtc driver. This can work around to recycle rtc id.
	 But if kernel fix this issue, it shell be removed away.
	*/
	rtc->rtc_dev->dev.release(&rtc->rtc_dev->dev);

	RTC_CR &= ~RTC_CR_ALRM;

	free_irq( rtc->alarm_irq, rtc);
	free_irq( rtc->interrupt_irq, rtc);

	iounmap( rtc->regbase);
	release_resource( rtc->res);

	platform_set_drvdata( pdev, NULL);

	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id atf_rtc_dt_match[] = {
	{ .compatible = "andestech,atfrtc010" },
	{},
};
MODULE_DEVICE_TABLE(of, atf_rtc_dt_match);
#endif

static struct platform_driver atf_rtc_platform_driver = {

	.driver		= {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
		.of_match_table	= of_match_ptr(atf_rtc_dt_match),
	},
	.probe		= atf_rtc_probe,
	.remove		= atf_rtc_remove,
};

module_platform_driver(atf_rtc_platform_driver);

MODULE_DESCRIPTION( "SuperH on-chip RTC driver");
MODULE_AUTHOR( "Paul Mundt <lethal@linux-sh.org>, "
	      "Jamie Lenehan <lenehan@twibble.org>, "
	      "Angelo Castello <angelo.castello@st.com>, "
	      "Roy Lee <roylee@andestech.com>");
MODULE_LICENSE( "GPL");
MODULE_ALIAS( "platform:" DRV_NAME);
