/*
 *  Watchdog driver for the ATCWDT200 Watch Dog Driver
 *
 *      (c) Copyright 2016 Andes Technology Corp. (www.andestech.com)
 *      Based on sa1100_wdt.c by Oleg Drokin <green@crimea.edu>
 *      Based on SoftDog driver by Alan Cox <alan@redhat.com>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *      11/07/2016 Initial release
 */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/traps.h>
#include <linux/reboot.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/reboot.h>

#define DRIVER_NAME	"atcwdt200"
#define DEBUG( str, ...)			\
	do{					\
		if( debug)			\
		printk( str, ##__VA_ARGS__);	\
	} while(0)
/* ID and Revision Register */
#define idrev		(*( volatile unsigned long *)( wdt_base + 0x00))
#define ID_OFF		12
#define ID_MSK		(0xfffff<<ID_OFF)
#define ID		(0x03002<<ID_OFF)

/* Control Register */
#define ctrl		(*( volatile unsigned long *)( wdt_base + 0x10))
#define RST_TIME_OFF	8
#define RST_TIME_MSK	(0x3<<RST_TIME_OFF)
#define RST_CLK_128	(0<<RST_TIME_OFF)
#define RST_CLK_256	(1<<RST_TIME_OFF)
#define RST_CLK_512	(2<<RST_TIME_OFF)
#define RST_CLK_1024	(3<<RST_TIME_OFF)
#define INT_TIME_OFF	4
#define INT_TIME_MSK	(0x7<<INT_TIME_OFF)
#define INT_CLK_64	(0<<INT_TIME_OFF)
#define INT_CLK_256	(1<<INT_TIME_OFF)
#define INT_CLK_1024	(2<<INT_TIME_OFF)
#define INT_CLK_2048	(3<<INT_TIME_OFF)
#define INT_CLK_4096	(4<<INT_TIME_OFF)
#define INT_CLK_8192	(5<<INT_TIME_OFF)
#define INT_CLK_16384	(6<<INT_TIME_OFF)
#define INT_CLK_32768	(7<<INT_TIME_OFF)
#define RST_EN		(1<<3)
#define INT_EN		(1<<2)
#define CLK_PCLK	(1<<1)
#define WDT_EN		(1<<0)

/* Restart Register */
#define restart		(*( volatile unsigned long *)( wdt_base + 0x14))
#define RESTART_MAGIC	0xcafe

/* Write Enable Register */
#define wren		(*( volatile unsigned long *)( wdt_base + 0x18))
#define WP_MAGIC	0x5aa5

/* Status Register */
#define st		(*( volatile unsigned long *)( wdt_base + 0x1c))
#define INT_EXPIRED	(1<<0)

#define TIMER_MARGIN	60	/* (secs) Default is 1 minute */

static int expect_close = 0;
static int debug = 0;
static int timeout = TIMER_MARGIN;  /* in seconds */
static bool nowayout = WATCHDOG_NOWAYOUT;
int irq_count = 0;
module_param(nowayout, bool, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started "
	"(default=" __MODULE_STRING(WATCHDOG_NOWAYOUT) ")");
static int wdt_panic = 0;
module_param(wdt_panic, int, 0);
MODULE_PARM_DESC(wdt_panic,
	"Watchdog action, set to 1 to panic, 0 to reboot (default=0)");

module_param(debug, int, 0);
module_param(timeout, int, 0);
static u32 wdt_freq;
static void __iomem *wdt_base;

void wdt_start(int rst)
{
	wren = WP_MAGIC;
	if(rst)
		ctrl = (RST_CLK_1024|RST_EN|WDT_EN);
	else
		ctrl = (INT_CLK_32768|INT_EN|WDT_EN);
	irq_count = 0;
	wren = WP_MAGIC;
	restart = RESTART_MAGIC;
	st |= INT_EXPIRED;
}

void wdt_irq(void)
{
	irq_count++;
	st |= INT_EXPIRED;
	wren = WP_MAGIC;
	restart = RESTART_MAGIC;
	if(irq_count == timeout)
	{
		if (wdt_panic) {
			panic("Software Watchdog Timer expired");
		} else {
			wdt_start(1);
			emergency_restart();
			pr_crit("Reboot didn't ?????\n");
		}
	}
}

static int wdtdog_open(struct inode *inode, struct file *file){
	wdt_start(0);
	DEBUG("Activating WDT..\n");
	return 0;
}

static int wdtdog_release(struct inode *inode, struct file *file){
#ifndef CONFIG_WATCHDOG_NOWAYOUT
	/*
	 *  Shut off the timer.
	 *  Lock it in if it's a module and we defined ...NOWAYOUT
	 */
	if(expect_close){
		wren = WP_MAGIC;
		ctrl = 0;
		irq_count = 0;
		DEBUG( "Deactivating WDT..\n");
	}
#endif
	expect_close = 0;
	return 0;
}

static ssize_t wdtdog_write(struct file *file, const char *data, size_t len, loff_t *ppos){
	if (!len)
		return 0;

	if (!nowayout){
		size_t i;
		for (i = 0; i < len; i++){
			char c;
			if (get_user(c, data + i))
				return -EFAULT;
			if (c == 'V'){
				expect_close = 42;
				break;
			}
		}
	}
	irq_count = 0;
	return len;
}

static long wdtdog_ioctl(struct file *file, unsigned int cmd, unsigned long arg){
	static struct watchdog_info ident = {
		.identity = "FTWDT010 watchdog",
	};

	int time = 0;
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	int new_options, retval = -EINVAL;

	switch( cmd){
		case WDIOC_GETSUPPORT:
			return copy_to_user(argp, &ident, sizeof( ident));

		case WDIOC_GETSTATUS:
			return put_user(0, p);

		case WDIOC_GETBOOTSTATUS:
			return put_user(st ? WDIOF_CARDRESET : 0, p);

		case WDIOC_SETTIMEOUT:
			if( get_user( time, p))
				return 0;

			if( time <= 0 || time > 255)
				return -EINVAL;

			timeout = time;
			return put_user(timeout, p);

		case WDIOC_SETOPTIONS:
			if (get_user(new_options, p))
				return -EFAULT;

			if (new_options & WDIOS_DISABLECARD){
				wren = WP_MAGIC;
				ctrl &= ~WDT_EN;
				retval = 0;
			}

			if (new_options & WDIOS_ENABLECARD){
				wdt_start(0);
				retval = 0;
			}
			return retval;

		case WDIOC_GETTIMEOUT:
			return put_user(timeout, p);

		case WDIOC_KEEPALIVE:
			wren = WP_MAGIC;
			restart = RESTART_MAGIC;
			irq_count = 0;
			return 0;

		default:
			return -ENOIOCTLCMD;
	}
}

static struct file_operations wdtdog_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= wdtdog_write,
	.unlocked_ioctl	= wdtdog_ioctl,
	.open		= wdtdog_open,
	.release	= wdtdog_release,
};

static struct miscdevice atcwdt200_dog_miscdev = {
	WATCHDOG_MINOR,
	"ATCWDT200 watchdog",
	&wdtdog_fops
};

/*
 *	Notifier for system down
 */
static int wdt_notify_sys(struct notifier_block *this,
					unsigned long code, void *unused)
{
	if (code == SYS_RESTART) {
		/*
		 * Cobalt devices have no way of rebooting themselves other
		 * than getting the watchdog to pull reset, so we restart the
		 * watchdog on reboot with no heartbeat
		 */
		timeout = 1;
		wdt_start(0);
	}
	return NOTIFY_DONE;
}


/*
 *	The WDT needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off.
 */
static struct notifier_block wdt_notifier = {
	.notifier_call = wdt_notify_sys,
};








#ifdef CONFIG_OF
static const struct of_device_id atcwdt200_dog_match[] = {
	{ .compatible = "andestech,atcwdt200" },
	{},
};
MODULE_DEVICE_TABLE(of, atcwdt200_dog_match);
#endif

static int atcwdt200_dog_probe(struct platform_device *pdev){
	int ret;
	struct resource *res;
	struct device_node *node;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	wdt_base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(wdt_base))
		return PTR_ERR(wdt_base);

	if((idrev & ID_MSK)!=ID)
		return -ENOENT;

	node = of_find_matching_node(NULL, atcwdt200_dog_match);
	if (of_property_read_u32(node, "clock-frequency", &wdt_freq))
		panic("Can't read clock-frequency");

	ret = misc_register(&atcwdt200_dog_miscdev);
	if( ret)
		return ret;
	register_NMI_handler(wdt_irq);
	ret = register_reboot_notifier(&wdt_notifier);
	if (ret) {
		printk("cannot register reboot notifier (err=%d)\n", ret);
		return ret;
	}


	DEBUG("ATCWDT200 watchdog timer: timer timeout %d sec\n", timeout);
	return 0;
}

static int atcwdt200_dog_remove(struct platform_device *pdev){
	misc_deregister(&atcwdt200_dog_miscdev);
	unregister_reboot_notifier(&wdt_notifier);
	unregister_NMI_handler();
	return 0;
}

static struct platform_driver atcwdt200_dog_driver = {
	.probe		= atcwdt200_dog_probe,
	.remove		= atcwdt200_dog_remove,
	.driver		= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(atcwdt200_dog_match),
	},
};

module_platform_driver(atcwdt200_dog_driver);
MODULE_AUTHOR("Andes Corp.");
MODULE_LICENSE("GPL");
