/*
 *  Watchdog driver for the FTWDT010 Watch Dog Driver
 *
 *      (c) Copyright 2004 Faraday Technology Corp. (www.faraday-tech.com)
 *      Based on sa1100_wdt.c by Oleg Drokin <green@crimea.edu>
 *      Based on SoftDog driver by Alan Cox <alan@redhat.com>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *      27/11/2004 Initial release
 */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <linux/clk.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/reboot.h>

#define DRIVER_NAME	"atfwdt010"
#define DEBUG( str, ...)			\
	do{					\
		if( debug)			\
		printk( str, ##__VA_ARGS__);	\
	} while(0)

#define wdcounter	(*( volatile unsigned long *)(wdt_base + 0x00))
#define wdload		(*( volatile unsigned long *)(wdt_base + 0x04))
#define wdrestart	(*( volatile unsigned long *)(wdt_base + 0x08))
#define wdcr		(*( volatile unsigned long *)(wdt_base + 0x0C))
#define wdstatus	(*( volatile unsigned long *)(wdt_base + 0x10))
#define wdclear		(*( volatile unsigned long *)(wdt_base + 0x14))
#define wdintrcter	(*( volatile unsigned long *)(wdt_base + 0x18))

#define TIMER_MARGIN	60	/* (secs) Default is 1 minute */
#define RESTART_MAGIC	0x5AB9

static int expect_close = 0;
static int debug = 0;
static int timeout = TIMER_MARGIN;  /* in seconds */
static bool nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, bool, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started "
	"(default=" __MODULE_STRING(WATCHDOG_NOWAYOUT) ")");
module_param(debug, int, 0);
module_param(timeout, int, 0);

static u32 wdt_freq;
static void __iomem *wdt_base;

static void wdt_start(int rst)
{
	wdcr = 0;
	wdload = wdt_freq * rst;
	wdrestart = RESTART_MAGIC;	/* Magic number */
	wdcr = 0x03;			/* Enable WDT */
}

static int atfwdt010_dog_open(struct inode *inode, struct file *file){
	wdt_start(timeout);
	DEBUG("Activating WDT..\n");
	return 0;
}

static int atfwdt010_dog_release(struct inode *inode, struct file *file){
#ifndef CONFIG_WATCHDOG_NOWAYOUT
	/*
	 *  Shut off the timer.
	 *  Lock it in if it's a module and we defined ...NOWAYOUT
	 */
	if(expect_close){
		wdcr = 0;
		printk("disable wdt\n");
		DEBUG( "Deactivating WDT..\n");
	}
#endif
	expect_close = 0;
	return 0;
}

static ssize_t atfwdt010_dog_write(struct file *file, const char *data, size_t len, loff_t *ppos){
	if (!len)
		return 0;

	if (!nowayout){
		size_t i;
		for (i = 0; i < len; i++){
			char c;
			if (get_user(c, data + i))
				return -EFAULT;
			if (c == 'V'){
				expect_close = 42;
				break;
			}
		}
	}
	wdrestart = RESTART_MAGIC;
	return len;
}

static long atfwdt010_dog_ioctl(struct file *file, unsigned int cmd, unsigned long arg){
	static struct watchdog_info ident = {
		.identity = "ATFWDT010 watchdog",
	};

	int time = 0;
	int new_options, retval = -EINVAL;
	void __user *argp = (void __user *)arg;
	int __user *p = argp;

	switch( cmd){

		case WDIOC_GETSUPPORT:
			return copy_to_user(argp, &ident, sizeof( ident));

		case WDIOC_GETSTATUS:
			return put_user(0, p);

		case WDIOC_GETBOOTSTATUS:
			return put_user(wdstatus ? WDIOF_CARDRESET : 0, p);

		case WDIOC_SETTIMEOUT:
			if( get_user( time, p))
				return 0;

			if( time <= 0 || time > 255)

				return -EINVAL;

			timeout = time;
			wdload = wdt_freq * timeout;
			wdrestart = RESTART_MAGIC;
			return put_user(timeout, p);

		case WDIOC_SETOPTIONS:
			if (get_user(new_options, p))
				return -EFAULT;

			if (new_options & WDIOS_DISABLECARD){
				wdcr = 0;
				retval = 0;
			}
			if (new_options & WDIOS_ENABLECARD){
				wdt_start(timeout);
				retval = 0;
			}
			return retval;

		case WDIOC_GETTIMEOUT:
			return put_user(timeout, p);

		case WDIOC_KEEPALIVE:
			wdrestart = RESTART_MAGIC;
			return 0;

		default:
			return -ENOIOCTLCMD;

	}
}

static int wdt_notify_sys(struct notifier_block *this,
					unsigned long code, void *unused)
{
	if (code == SYS_RESTART) {
		/*
		 * Cobalt devices have no way of rebooting themselves other
		 * than getting the watchdog to pull reset, so we restart the
		 * watchdog on reboot with no heartbeat
		 */
		timeout = 1;
		wdt_start(timeout);
	}
	return NOTIFY_DONE;
}

static struct notifier_block wdt_notifier = {
	.notifier_call = wdt_notify_sys,
};


static struct file_operations atfwdt010_dog_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= atfwdt010_dog_write,
	.unlocked_ioctl	= atfwdt010_dog_ioctl,
	.open		= atfwdt010_dog_open,
	.release	= atfwdt010_dog_release,
};

static struct miscdevice atfwdt010_dog_miscdev = {
	.minor = WATCHDOG_MINOR,
	.name = "ATFWDT010 watchdog",
	.fops = &atfwdt010_dog_fops
};

#ifdef CONFIG_OF
static const struct of_device_id atfwdt010_dog_match[] = {
	{ .compatible = "andestech,atfwdt010" },
	{},
};
MODULE_DEVICE_TABLE(of, atfwdt010_dog_match);
#endif

static int atfwdt010_dog_probe(struct platform_device *pdev){
	int ret;
	struct resource *res;
	struct device_node *node;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	wdt_base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(wdt_base))
		return PTR_ERR(wdt_base);

	node = of_find_matching_node(NULL, atfwdt010_dog_match);
	if (of_property_read_u32(node, "clock-frequency", &wdt_freq))
		panic("Can't read clock-frequency");

	ret = misc_register(&atfwdt010_dog_miscdev);
	if( ret)
		return ret;

	ret = register_reboot_notifier(&wdt_notifier);
	if (ret) {
		printk("cannot register reboot notifier (err=%d)\n", ret);
		return ret;
	}


	DEBUG("ATFWDT010 watchdog timer: timer timeout %d sec\n", timeout);
	return 0;
}

static int atfwdt010_dog_remove(struct platform_device *pdev){
	misc_deregister(&atfwdt010_dog_miscdev);
	unregister_reboot_notifier(&wdt_notifier);
	return 0;
}

static struct platform_driver atfwdt010_dog_driver = {
	.probe		= atfwdt010_dog_probe,
	.remove		= atfwdt010_dog_remove,
	.driver		= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(atfwdt010_dog_match),
	},
};

module_platform_driver(atfwdt010_dog_driver);
MODULE_AUTHOR("Andes Corp.");
MODULE_LICENSE("GPL");
